/*        File: ifcns.c
     Copyright: (C) 1998, 1999 by Elise de Doncker, Ajay Gupta, Alan Genz and Rodger Zanny. 
                       All rights reserved. See Copyright.txt in top-level directory.
  ParInt Versn: All versions
   Description: Source for integration functions
       Version: 03/04/98  rrz  Original
		07/23/98  rrz  Add fcn18 and fcn19 (good for extrapolation)
		08/13/98  rrz  Add fcn22
		08/30/98  rrz  Added fcn23
		05/13/99  rrz  Change default return value to be 0
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
		11/17/99  rrz  Add fcn4x functions; add fcn31b
		03/19/00  rrz  Add fcn24 (sample CERN function)
*/


#include <math.h>
#include <parint.h>
#include <pi-version.h>
#include <math-misc.h>

#include <timer.h>

/* fcn1(x0,x1,x2,x3,x4,x5) = x0*x1^2*sin(x2)*(4+x3+x4+x5)^-1
 */
int fcn1(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = 4.0 + x[3] + x[4] + x[5];
    funvls[0] = (z != 0.0
		 ? (x[0] * x[1] * x[1] * pi_sin(x[2])) / z
		 : 0.0);
    return 0;
} /* fcn1() */


/* fcn2(x) = x2^2 * x3 * exp(x2*x3) / ( x0 + x1 + 1 )^2)
 */
int fcn2(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = x[2] * x[3];
    pi_base_t w = x[0] + x[1] + 1.0;
    w *= w;
    funvls[0] = (w != 0.0
		 ? x[2] * z * pi_exp(z) / w
		 : 0.0);
    return 0;
} /* fcn2() */


/* fcn3(x0,x1,x2) = 8/(1 + 2(x0+x1+x2))
 */
int fcn3(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = 1.0 + 2.0 * (x[0]+x[1]+x[2]);
    funvls[0] = (z != 0.0
		 ? 8.0 / z
		 : 0.0);
    return 0;
} /* fcn3() */


/* fcn4(x0,x1,x2,x3,x4) = cos(x0+x1+x2+x3+x4)
 */
int fcn4(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    funvls[0] = pi_cos(x[0] + x[1] + x[2] + x[3] + x[4]);
    /* nap(0.001); */
    return 0;
} /* fcn4() */


/* fcn5(x0,x1,x2,x3) = sin(10x1)
 */
int fcn5(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    funvls[0] = pi_sin(10.0 * x[1]);
    return 0;
} /* fcn5() */


/* fcn6(x0,x1) = cos(x0 + x1)
 */
int fcn6(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    funvls[0] = pi_cos(x[0] + x[1]);
    return 0;
} /* fcn6() */


/* fcn7(x0,x1,x2) = (x0+x1+x2)^-2
 */
int fcn7(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = x[0] + x[1] + x[2];
    z *= z;
    funvls[0] = (z != 0.0
		 ? 1.0 / z
		 : 0.0);
    /* nap(.001); */
    return 0;
} /* fcn7() */


/* fcn8(x0,x1) = 605*x1/((1+120(1-x1))((1+120(1-x1))^2+25x0^2*x1^2) 
 */
int fcn8(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = 1.0 + 120.0 * (1.0 - x[1]);
    pi_base_t w = (z * (z*z + 25.0*x[0]*x[0]*x[1]*x[1]));
    funvls[0] = (w != 0.0
		 ? 605.0 * x[1] / w
		 : 0.0);
    return 0;
} /* fcn8() */


/* fcn9(x0,x1) = 1/((x0^2 + .0001)*((x1+.25)^2 + .0001)); 
 */
int fcn9(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    pi_base_t z = (x[1] + 0.25L);
    pi_base_t w = (x[0]*x[0] + 0.0001L) * (z*z + 0.0001L);
#else
    pi_base_t z = (x[1] + 0.25);
    pi_base_t w = (x[0]*x[0] + 0.0001) * (z*z + 0.0001);
#endif
    funvls[0] = (w != 0.0
		 ? 1.0 / w
		 : 0.0);
    return 0;
} /* fcn9() */


/* fcn10(x0,x1) = exp(|x0 + x1 - 1|) 
 */
int fcn10(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = x[0] + x[1] - 1.0;
    funvls[0] = (z >= 0.0
		 ? pi_exp(z)
		 : pi_exp(-z));
    return 0;
} /* fcn10() */


int fcn14(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    /* (Note: using pow(x[0], -0.5) instead of 1/sqrt(x[0]) was tried, but proved 
       to run slower) */
    if (x[0] <= 1.0)
	funvls[0] = ((x[0] != 0.0) ? 1.0 / pi_sqrt(x[0]) : 0.0);
    else
	funvls[0] = ((x[0] != 2.0) ? 1.0 / pi_sqrt(2.0-x[0]) : 0.0);
    return 0;
} /* fcn14() */


int fcn15(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    funvls[0] = x[0] * x[0] + 5.0 * x[0] + 10.0;
    return 0;
} /* fcn15() */


int fcn16(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    funvls[0] = pi_exp(x[0] + x[1]);
    return 0;
} /* fcn15() */


int fcn17(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = x[0];
    pi_base_t y = x[0];
    int i;

    for (i = 0; i < 10; ++i)
    {
	funvls[i] = z + y + i;
	z *= x[0];
	y += x[0];
    }
    
    return 0;
} /* fcn17() */



int fcn18(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    funvls[0] = pi_pow(x[0], -0.2L) * pi_pow(x[1], -0.2L) * pi_pow(x[0] + x[1], -02.L);
#else
    funvls[0] = pi_pow(x[0], -0.2) * pi_pow(x[1], -0.2) * pi_pow(x[0] + x[1], -0.2);
#endif
    return 0;
} /* fcn18() */


int fcn19(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    funvls[0] = (  pi_pow(x[0], -0.20L) 
		 * pi_pow(x[1], -0.20L)
		 * pi_pow(x[2], -0.20L)
		 * pi_pow(x[3], -0.20L)
		 * pi_pow(x[0] + x[1] + x[2] + x[3], -0.20L));
#else    
    funvls[0] = (  pi_pow(x[0], -0.20) 
		 * pi_pow(x[1], -0.20)
		 * pi_pow(x[2], -0.20)
		 * pi_pow(x[3], -0.20)
		 * pi_pow(x[0] + x[1] + x[2] + x[3], -0.20));
#endif
    return 0;
} /* fcn19() */


int fcn21(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    funvls[0] = (  pi_pow(x[0], -0.20L) 
		 * pi_pow(x[1], -0.20L)
		 * pi_pow(x[2], -0.20L)
		 * pi_pow(x[0] + x[1] + x[2], -0.20L));
#else    
    funvls[0] = (  pi_pow(x[0], -0.20) 
		 * pi_pow(x[1], -0.20)
		 * pi_pow(x[2], -0.20)
		 * pi_pow(x[0] + x[1] + x[2], -0.20));
#endif
    return 0;
} /* fcn21() */


int fcn22(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = x[0] * x[1];
    funvls[0] = ((z > 0.0) ? 1.0 / pi_sqrt(z) : 0.0);
    return 0;
} /* fcn22() */


int fcn23(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    funvls[0] = ((x[0] > 0.0) ? pi_pow(x[0], -0.9L) : 0.0);
#else
    funvls[0] = ((x[0] > 0.0) ? pi_pow(x[0], -0.9) : 0.0);
#endif
    return 0;
} /* fcn23() */


/* Sample integral from CERN.  The following table lays out the results for varying eps:
 *  eps 1.e-1 2.4807
 *  eps 1.e-2 3.0295
 *  eps 1.e-3 3.1258
 *  eps 1.e-4 3.1396
 *  eps 1.e-5 3.1413
 *  eps 1.e-6 3.1416
 */
#define FCN24_EPS 1.0e-5
int fcn24(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t res = x[0] + x[1] - 1.0;
    funvls[0] = 2.0 * FCN24_EPS * x[1] / (res * res + FCN24_EPS * FCN24_EPS);
    return 0;
} /* fcn24() */


#define FCN25_ALPHA   0.9
#define FCN25_BETA    0.7
int fcn25(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = 1.0 - x[0];
    pi_base_t z1 = 1.0 - x[1];
    pi_base_t z = (pi_pow(x[0]*x[0] + x[1]*x[1] + x[2]*x[2] + x[3]*x[3], FCN25_ALPHA)
		   * pi_pow(z0*z0 + z1*z1 + x[2]*x[2] + x[3]*x[3], FCN25_BETA));
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn25() */



/* fcn31(x0) = (x0)^-0.9
 */
int fcn31(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    pi_base_t z = pow(x[0], 0.9L);
#else
    pi_base_t z = pow(x[0], 0.9);
#endif
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn31() */


/* fcn31b(x0) = (x0)^-0.99
 */
int fcn31b(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
#ifdef USING_LONG_DOUBLES
    pi_base_t z = pow(x[0], 0.99L);
#else
    pi_base_t z = pow(x[0], 0.99);
#endif
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn31b() */


/* fcn32(x0) = (x0 + x1)^-1
 */
int fcn32(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = (x[0] + x[1]);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn32() */


/* fcn34(x0) = (x0 + x1 + x2 + x3)^-3
 */
int fcn34(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = pow(x[0] + x[1] + x[2] + x[3], 3.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn34() */

/* fcn35(x0) = (x0 + x1 + x2 + x3)^-3
 */
int fcn35(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z = pow(x[0] + x[1] + x[2] + x[3] + x[4] + x[5] + x[6] + x[7] + x[8] + x[9], 
		      2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn35() */


/******************************************************************************
 * A new set of radial point singularity functions, of the form:
 * f(x0, x1, ..., x_(d-1) ) = 1.0 / ( (x-alpha0)^2 + (x-alpha1)^2 ... )^(beta/2)
 * where each 0.0 <= alpha_i <= 1.0 and -d < beta < 0.
 * d is the # of dimensions; the point defined by 
 *     (alpha_0, alpha_1, ..., alpha_(d-1))
 * is the coordinate of the singularity, and the smaller beta is, the 
 * steeper the singularity is.  The function is written as 
 * $r^\beta_{a_0, a_1, \ldots, a_{d-1}}$
 *
 * The constants below are merely for clarity; the beta constants
 * are written as positive numbers, and are effectively negated within 
 * the functions. Each function number has a different number of dimensions,
 * within each number the functions vary in the location and difficulty of the
 * singularity.
 *
 * At the point singularity the function value is a zero.
 */

#define FCN41_BETA        0.9
#define FCN41A_ALPHA_0    0.0
#define FCN41B_ALPHA_0    0.5

#define FCN42_BETA        1.9
#define FCN42A_ALPHA_0    0.1
#define FCN42A_ALPHA_1    0.1
#define FCN42B_ALPHA_0    0.1
#define FCN42B_ALPHA_1    0.5
#define FCN42C_ALPHA_0    0.5
#define FCN42C_ALPHA_1    0.5

#define FCN43_BETA         2.9
#define FCN43A_ALPHA_0     0.1
#define FCN43A_ALPHA_1     0.1
#define FCN43A_ALPHA_2     0.1
#define FCN43B_ALPHA_0     0.1
#define FCN43B_ALPHA_1     0.1
#define FCN43B_ALPHA_2     0.5
#define FCN43C_ALPHA_0     0.1
#define FCN43C_ALPHA_1     0.5
#define FCN43C_ALPHA_2     0.5
#define FCN43D_ALPHA_0     0.5
#define FCN43D_ALPHA_1     0.5
#define FCN43D_ALPHA_2     0.5

#define FCN44_BETA          2.0
#define FCN44A_ALPHA_0      0.1
#define FCN44A_ALPHA_1      0.1
#define FCN44A_ALPHA_2      0.1
#define FCN44A_ALPHA_3      0.1
#define FCN44A_ALPHA_4      0.1
#define FCN44A_ALPHA_5      0.1

#define FCN44B_ALPHA_0      0.1
#define FCN44B_ALPHA_1      0.1
#define FCN44B_ALPHA_2      0.1
#define FCN44B_ALPHA_3      0.1
#define FCN44B_ALPHA_4      0.1
#define FCN44B_ALPHA_5      0.5

#define FCN44C_ALPHA_0      0.1
#define FCN44C_ALPHA_1      0.1
#define FCN44C_ALPHA_2      0.1
#define FCN44C_ALPHA_3      0.1
#define FCN44C_ALPHA_4      0.5
#define FCN44C_ALPHA_5      0.5

#define FCN44D_ALPHA_0      0.1
#define FCN44D_ALPHA_1      0.1
#define FCN44D_ALPHA_2      0.1
#define FCN44D_ALPHA_3      0.5
#define FCN44D_ALPHA_4      0.5
#define FCN44D_ALPHA_5      0.5

#define FCN44E_ALPHA_0      0.1
#define FCN44E_ALPHA_1      0.1
#define FCN44E_ALPHA_2      0.5
#define FCN44E_ALPHA_3      0.5
#define FCN44E_ALPHA_4      0.5
#define FCN44E_ALPHA_5      0.5

#define FCN44F_ALPHA_0      0.1
#define FCN44F_ALPHA_1      0.5
#define FCN44F_ALPHA_2      0.5
#define FCN44F_ALPHA_3      0.5
#define FCN44F_ALPHA_4      0.5
#define FCN44F_ALPHA_5      0.5

#define FCN44G_ALPHA_0      0.5
#define FCN44G_ALPHA_1      0.5
#define FCN44G_ALPHA_2      0.5
#define FCN44G_ALPHA_3      0.5
#define FCN44G_ALPHA_4      0.5
#define FCN44G_ALPHA_5      0.5


int fcn41a(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    /* (Note: The ^2 of the x-alpha, and the /2 in the beta, cancel out): */
    pi_base_t z = pi_pow(absol(x[0] - FCN41A_ALPHA_0), FCN41_BETA);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn41a() */

int fcn41b(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    /* (Note: The ^2 of the x-alpha, and the /2 in the beta, cancel out): */
    pi_base_t z = pi_pow(absol(x[0] - FCN41B_ALPHA_0), FCN41_BETA);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn41b() */


int fcn42a(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN42A_ALPHA_0;
    pi_base_t z1 = x[1] - FCN42A_ALPHA_1;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1, FCN42_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn42a() */

int fcn42b(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN42B_ALPHA_0;
    pi_base_t z1 = x[1] - FCN42B_ALPHA_1;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1, FCN42_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn42b() */

int fcn42c(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN42C_ALPHA_0;
    pi_base_t z1 = x[1] - FCN42C_ALPHA_1;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1, FCN42_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn42c() */


int fcn43a(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN43A_ALPHA_0;
    pi_base_t z1 = x[1] - FCN43A_ALPHA_1;
    pi_base_t z2 = x[2] - FCN43A_ALPHA_2;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1 + z2 * z2, FCN43_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn43a() */

int fcn43b(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN43B_ALPHA_0;
    pi_base_t z1 = x[1] - FCN43B_ALPHA_1;
    pi_base_t z2 = x[2] - FCN43B_ALPHA_2;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1 + z2 * z2, FCN43_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn43b() */

int fcn43c(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN43C_ALPHA_0;
    pi_base_t z1 = x[1] - FCN43C_ALPHA_1;
    pi_base_t z2 = x[2] - FCN43C_ALPHA_2;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1 + z2 * z2, FCN43_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn43c() */

int fcn43d(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN43D_ALPHA_0;
    pi_base_t z1 = x[1] - FCN43D_ALPHA_1;
    pi_base_t z2 = x[2] - FCN43D_ALPHA_2;
    pi_base_t z = pi_pow(z0 * z0 + z1 * z1 + z2 * z2, FCN43_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn43d() */


int fcn44a(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44A_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44A_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44A_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44A_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44A_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44A_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44a() */


int fcn44b(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44B_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44B_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44B_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44B_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44B_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44B_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44b() */


int fcn44c(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44C_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44C_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44C_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44C_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44C_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44C_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44c() */


int fcn44d(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44D_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44D_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44D_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44D_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44D_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44D_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44d() */


int fcn44e(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44E_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44E_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44E_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44E_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44E_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44E_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44e() */


int fcn44f(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44F_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44F_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44F_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44F_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44F_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44F_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44f() */


int fcn44g(int *ndims, pi_base_t *x, int *nfcns, pi_base_t *funvls)
{
    pi_base_t z0 = x[0] - FCN44G_ALPHA_0;
    pi_base_t z1 = x[1] - FCN44G_ALPHA_1;
    pi_base_t z2 = x[2] - FCN44G_ALPHA_2;
    pi_base_t z3 = x[3] - FCN44G_ALPHA_3;
    pi_base_t z4 = x[4] - FCN44G_ALPHA_4;
    pi_base_t z5 = x[5] - FCN44G_ALPHA_5;
    pi_base_t z = pi_pow(z0*z0 + z1*z1 + z2*z2 + z3*z3 + z4*z4 + z5*z5, FCN44_BETA / 2.0);
    funvls[0] = ((z != 0.0) ? 1.0 / z : 0.0);
    return 0;
} /* fcn44g() */

