/*        File: pi-defs.h
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: All versions
   Description: Basic definitions common to all versions of ParInt (but not for user changing)
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/include/pi-defs.h,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:31 $

                $Log: pi-defs.h,v $
                Revision 1.1.1.1  2000/06/06 03:13:31  parint
                Initial checkin


       Version: 01/13/98  rrz  Original
		05/12/98  rrz  Code freeze for ParInt1.0
		06/18/98  rrz  Added fields to update_t for visualization
		05/01/99  rrz  Move pi_assert() macro and proto for
				pi_terminate here
		05/25/99  rrz  Add proto for pi_sys_terminate()
		07/13/99  rrz  Add fields to update_t and region_t, and add
				typedef's, for logging
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

/* This header file contains items which are common across all versions of ParInt (though
 * these items may depend upon version-specific items).  No MPI or PVM specific items
 * should therefore be here.
 */

#ifndef PI_DEFS_H
#define PI_DEFS_H

#include <stdlib.h>          /* for free() */
#include <string.h>          /* for string fcns */
#include <sys/types.h>       /* for ulong & ushort */

#include <pi-version.h>
#include <parint.h>

/*****************************************************************************************/
/*** The first section defines things that are specific to ParInt:                     ***/

/* These must be provided by all versions, but is version-specific (it used to be in messaging.h,
 * but that was incorrect, b/c the SEQ version does not have this file): */
extern int pi_terminate(int pi_err_num, const char *err_str);
extern void pi_sys_terminate(const char *fmt, /* arg1, arg2, */ ...);

/* This macro is not version-specific, while the function _failed_assertion() is, and
 * is provided by messaging.c (or sparint.c for the SEQ version)
 */
#ifdef USE_MY_ASSERTIONS
extern void _failed_assertion(char *str, char *file, int line);
#define pi_assert(expr, str) do                                       \
    { if (!(expr)) { _failed_assertion(str, __FILE__, __LINE__); }    \
    } while (0)
#else
#define pi_assert(expr, str)
#endif

#ifdef USE_LOGGING
typedef ulong         pl_nnum_t;
typedef unsigned char pl_rank_t;
#endif

typedef struct {
    pi_base_t center[PI_MAX_NDIM];
    pi_base_t h_width[PI_MAX_NDIM];
    pi_base_t rgnvol;
    pi_base_t rgnval[PI_MAX_NFCN];     /* vector of integral answer estimates       */
    pi_base_t rgnerr_v[PI_MAX_NFCN];   /* vector of integral error estimates        */
    pi_base_t rgnerr;                  /* infNorm(rgnerr_v, numfun) (always >= 0.0) */
    int       divaxn;                  /* dim. to split next (1-indexed!)           */
#ifdef USE_LOGGING
    pl_nnum_t rgn_num;                 /* Unique (for rgn_creator_id) identifier for this rgn */
    pl_rank_t rgn_creator_id;          /* Worker that created this region */
#endif
} region_t;

/* (Note: region_t could become version-specific later on; note that PI_MAX_NFCN and
 * PI_MAX_NDIM currently *are* version specific. */


typedef struct {
    pi_base_t res[PI_MAX_NFCN];   
    pi_base_t est[PI_MAX_NFCN];
    int       l_done;           /* either TRUE or FALSE                                       */
    int       cnt;              /* # of function evaluations that went into this update       */
#ifdef USE_LOGGING
    pl_nnum_t update_num;       /* Unique (for a worker) identifier for this update */
#endif
    double    l_estabs;         /* local infNorm(estabs) at time of update; for visualization */
    long      num_region;       /* jli!! total # of regions on this woker; for visulization*/
    double    l_time;           /* local elapsed time at time of update; for visualization    */
} update_t;

/* jli!! */
#define HISTOSIZE 20     /* the width of the histogram */

/* structure for histogram */
typedef struct {
   int worker_id; /*the worker id */
   long num_region[HISTOSIZE]; /* the histogram array, # of regions */
   double low_exp; 
   double high_exp;
   double time; /* current elapsed time */
} hist_t;

/* update_t may also become version specific later on (thinking of extrap!) */

#define PI_MAX_IFCN_NAME    15            /* max length of an ifcn name */
#define PI_MAX_IFCN_DESC    60            /* max length of an ifcn description */


/* This structure holds all of the parameters of the integral problem, as opposed to how
 * the problem is solved. This is here, as opposed to in pi-vspec.h.  This is a tricky and 
 * deliberate decision. The idea is that no matter what versions of ParInt we end up with, they 
 * will all have the same input parameters to the basic integration problem.  Note that this
 * structure will need to be changed when we add simplex regions.  Note also that the
 * pi_rparm_t structure (which holds run-time paramters) *is* in pi-vspec.h, and is
 * very version specific: */
typedef struct  { 
    char      ifcn_name[PI_MAX_IFCN_NAME + 1]; /* Only if using ifcn from the library */
    int       ndims;                           /* Dimension of the integration region */
    int       nfcns;                           /* Number of integrands                */
    int       intg_rule;                       /* integration rule (enum value)       */
    pi_base_t eps_a;                           /* Requested absolute error            */
    pi_base_t eps_r;                           /* Requested relative error            */
    int       fcn_eval_limit;                  /* Limit of # of function calls        */
    pi_base_t a[PI_MAX_NDIM];                  /* Vector of lower bounds              */
    pi_base_t b[PI_MAX_NDIM];                  /* Vector of upper bounds              */
} pi_iparm_t;


/*****************************************************************************************/
/*** The next section defines things that are not specific to ParInt:                  ***/

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/*** A type for values that are boolean, i.e., only ever set or compared to 0 or 1: */
typedef char bool;

/*** For clarity: */
#define	FOREVER		for (;;)
#define until(cond)	while (!(cond))
#define do_nothing()

/*** A better version of free(): */
#define FREE(p) (free(p), (p) = NULL)

/*** I prefer these macro versions.  But then, my middle name is macro: */
#define STRCMP(s1, op, s2)       (strcmp((s1), (s2)) op 0)
#define STRNCMP(s1, op, s2, n)   (strncmp((s1), (s2), (n)) op 0)
#define MEMCMP(s1, op, s2, n)    (memcmp((s1), (s2), (n)) op 0)
#define STRCASECMP(s1, op, s2)   (strcasecmp((s1), (s2)) op 0)

/* Converts a flag value to a 'Y' or a 'N' (for printing): */
#define B2YN(flag)   ((char) ((flag) ? 'Y' : 'N'))

 /* A 'safe' cast-to-bool: Converts any non-zero integral value into a 1 as
 * a bool, converts any zero integral value into a 0 as a bool (safer than
 * a cast to bool when storing a value in a boolean variable or passing 
 * into a boolean parameter and you're dealing with a value that may be 
 * greater than can fit in a char): */
#define tobool(expr) (bool)((expr) && 1)

/* See pi-version.h for more info on this macro.  Used to control debugging output: */
#if ((defined(DBG_ALL_FILES) || defined(DBG_THIS_FILE)) && !defined(DBG_NO_FILES))
#define D(x)   x
#else
#define D(x)
#endif


#endif /* PI_DEFS_H */
