/*        File: math-rgn.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: All versions (but is region specific)
   Description: Source for functions to manipulate regions
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/math/math-rgn.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:29 $

                $Log: math-rgn.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:29  parint
                Initial checkin


       Version: 01/13/98  rrz  Original (from existing code)
		05/12/98  rrz  Code freeze for ParInt1.0
		10/28/98  rrz  Add assertion to split_region() re bad dax value
		11/17/98  rrz  Modify get_initial_region() to allow arbitrary
				numbers of workers
		12/18/98  rrz  Modify get_initial_region() to allow for forcing
				of simple rgn calc
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

#include <math.h>

#include <pi-version.h>
#include <pi-defs.h>
#include <math-misc.h>
#include <math-rgn.h>
#include <math-intg.h>


/* Returns the volume of the region at rgn_p, *and*, as a side effect,
 * sets the rgnvol field inside rgn_p to that value
 */
extern pi_base_t volume(int ndims, region_t *rgn_p)
{
    pi_base_t vol = 1.0;
    register int i;
    for (i = 0; i < ndims; i++)
	vol *= rgn_p->h_width[i];
    return (rgn_p->rgnvol = vol * POW2(ndims));
} /* volume() */


/* Note that nworkers will include the IC, if she is working.  newrgn is the initial region
 * for the process that called this function (in which center[], h_width[], and rgnvol are set).
 * The local volume is also returned in local_vol and the total regions volume is returned
 * in full_vol.
 */
extern void get_initial_region(int ndims, int nworkers, const pi_base_t *a, const pi_base_t *b,
			       int id, region_t *newrgn_p, pi_base_t *full_vol,
			       pi_base_t *local_vol, pi_rparm_t *rparm_p)
{
    /* This variable represents the 0-indexed rank of this PE among the workers in this group.  If
       the IC is not working, then it is just this worker's group rank minus 1 (e.g.: the IW with
       rank of 1 has an id1 of zero). If the IC is working, then the id1 is equal to the group
       rank (since the IC takes the 0th position): */
    int id1 = (rparm_p->ic_is_worker ? id : id - 1);

    pi_base_t full_intv[PI_MAX_NDIM];
    region_t totrgn;

    pi_assert(nworkers > 0, "get_initial_region(): Bad nworkers value");
    pi_assert(id != 0 || rparm_p->ic_is_worker, "get_init_region(): id=0, but IC not working");
    
    /* Calc the total region and the full_intv[] array: */
    {
	pi_base_t half_intv;
	int j;
	for (j = 0; j < ndims; j++)
	{
	    full_intv[j] = b[j] - a[j];
	    half_intv = 0.5 * full_intv[j];
	    totrgn.center[j] = a[j] + half_intv;
	    totrgn.h_width[j] = half_intv;
	}
    }

    /* (both calculation methods can now use id1, totrgn, and full_intv[]) */

    /* If we are forcing simple init rgn calc, then do so, else, only if we have 2^k workers
       should we do the fancy region splitting technique: */
    if (!rparm_p->simple_rgns && (clog2(nworkers) == flog2(nworkers)))
    {
	/* Divide the ndims dimensional region into 2^dim subregions.  If the IC is not working,
	   then it implies that nPE is nworkers+1 (i.e., nPE is 3, 5, 9, ...).

	   nd[j] is the # of divisions in the jth direction;  nd[j] = 2^m where m = dim/N + delta, 
	   delta = 1 for 0 <= j < dim mod N and delta = 0 for j >= dim mod N. */ 

	int dim = clog2(nworkers);
	int nd_1 = POW2(dim / ndims);
	int nd_2 = 2 * nd_1;

	pi_base_t sd[PI_MAX_NDIM];
	int nd[PI_MAX_NDIM];
	int j, d;

	for (j = 0; j < dim % ndims; j++)
	    nd[j] = nd_2;
	for (j = dim % ndims; j < ndims; j++)
	    nd[j] = nd_1;
	for (j = 0; j < ndims; j++)
	    sd[j] = full_intv[j] / ((pi_base_t) nd[j]);

	/* Calculate this processor's subregion's bounds: */
	for (d = 1, j = 0; j < ndims; j++)
	{
	    newrgn_p->center[j] = a[j] + ((pi_base_t) ((id1/d) % nd[j]) + 0.5) * sd[j];
	    newrgn_p->h_width[j] = 0.5 * sd[j];
	    d *= nd[j];
	}
    }
    else
    {
	int max_dim, j;

	/* Find the widest dimension: */
	max_dim = 0;
	{
	    pi_base_t max_dim_width = b[0] - a[0];
	    for (j = 1; j < ndims; ++j)
	    {
		if (full_intv[j] > max_dim_width)
		{
		    max_dim_width = full_intv[j];
		    max_dim = j;
		}
	    }
	}
	
	/* First assign to this process the entire region: */
	for (j = 0; j < ndims; ++j)
	{
	    newrgn_p->center[j] = totrgn.center[j];
	    newrgn_p->h_width[j] = totrgn.h_width[j];
	}
	
	/* The max_dim^th dimension (0-indexed) is the widest dimension.  Slice it like a stick 
	   of butter and give this process the piece corresponding to the rank (id1): */
	newrgn_p->h_width[max_dim] /= nworkers;
	newrgn_p->center[max_dim] = (newrgn_p->h_width[max_dim] * 2 * id1 
				     + newrgn_p->h_width[max_dim]);
    }
    
    *full_vol = volume(ndims, &totrgn);      /* volume() also sets rgnvol field */
    *local_vol = volume(ndims, newrgn_p);

#if 0 /* For thorough testing of creating initial region: */
    {
	int i;
	printf("Init region for process %d (worker rank %d)\n", id, id1);
	for (i = 0; i < ndims; ++i)
	    printf("   a[%d]: %.15G; b[%d]: %.15G\n", 
		   i, newrgn_p->center[i] - newrgn_p->h_width[i], 
		   i, newrgn_p->center[i] + newrgn_p->h_width[i]);
	printf("   full volume: %G; local volume %G; size_ratio: %f\n", 
	       *full_vol, *local_vol, *local_vol / *full_vol);
	fflush(NULL);
    }
#endif    

} /* get_initial_region() */


extern void split_region(region_t new[], const region_t *oldrgn_p)
{
    int dax = oldrgn_p->divaxn - 1;
    pi_base_t new_h_width = 0.5 * oldrgn_p->h_width[dax];
    pi_base_t new_volume  = 0.5 * oldrgn_p->rgnvol;
    int i;

    pi_assert(dax >= 0 && dax <= PI_MAX_NDIM, "Bad dax in split_region()");
    
    for (i = 0; i < REGION_SPLITS; i++)
    {
	new[i] = *oldrgn_p;
	new[i].h_width[dax] = new_h_width;
	new[i].center[dax] = oldrgn_p->center[dax] + (2*i - 1) * new_h_width;
	new[i].rgnvol = new_volume;
    }
    
} /* split_region() */


/* Returns TRUE if the region has zero volume, i.e., if the region has one dimension i
 * where a[i] == b[i].  Note the region is specified by a[] and b[] vectors
 */
extern bool region_has_zero_volume(const pi_base_t a[], pi_base_t const b[], int ndims)
{
    register int i;
    for (i = 0; i < ndims; ++i)
    {
	if (a[i] == b[i])
	    return TRUE;
    }
    return FALSE;
} /* region_has_zero_volume() */
