/*        File: heap.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: All
   Description: Source for functions to manipulate a heap of regions
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/misc/heap.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:30 $

                $Log: heap.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:30  parint
                Initial checkin


       Version: ??/??/??  ???  Original
		05/??/97  mg   Initial MPI Conversion
		10/??/97  akg  (and edd)
		01/13/98  rrz  Prepped for Release 1.0
		05/12/98  rrz  Code freeze for ParInt1.0
		07/18/98  rrz  BUG FIX: Allow call to add_heap() on an empty
				heap
		07/18/98  rrz  Add: new routines heap_merge() and
				heap_create_empty()
		09/16/98  rrz  Remove some extrap debugging code
		05/25/99  rrz  Use new routine pi_sys_terminate()
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

#include <stdio.h>
#include <stdlib.h>

#include <pi-defs.h>
#include <parint.h>
#include <heap.h>

#define SZ_HNODE sizeof(hnode)

/*** First functions are called only within this source file: */

/******************************************************************************/
/*                                                                            */
/* get_new_avail_parent                                                       */
/*                                                                            */
/* Pre-condition: hptr->availptr points to the node to which the last two     */
/*    leaves have been added (as its right and left children).                */
/* Post-condition: hptr->availptr points to the node to which next leaf will  */
/*    be added (as a left child) in order to maintain a complete, balanced    */
/*    binary tree.                                                            */

static void get_new_avail_parent( heap_t *hptr )
{
    hnode  *child, *parent;

    /* If the current available parent is the heap's root, then make the       */
    /* available parent the root's left child                                  */
    if ((hptr->availptr)->p_nptr == NULL)
    {
	hptr->availptr = (hptr->availptr)->lc_nptr;
	return;
    }

    child = hptr->availptr;
    parent = child->p_nptr;

    /* Go up the tree until a change from child is right child to child is     */
    /*   left child or until root node.                                        */
    while ( (parent->p_nptr != NULL) && (child == parent->rc_nptr))
    {
	child = parent;
	parent = child->p_nptr;
    }

    /* Now go back down the tree to the left from **appropriate** parent to    */
    /*   find available parent node.                                           */

    if ( (parent->p_nptr != NULL) || (child != parent->rc_nptr) )
	/* Have not reached root from the rght: change parent for descent       */
	parent = parent->rc_nptr;
    /* else at root node and arrived from right: don't change parent           */

    while ( parent->lc_nptr != NULL )
	parent = parent->lc_nptr;
    hptr->availptr = parent;

} /*  End of function `get_new_avail_parent'   */
 


/******************************************************************************/
/*                                                                            */
/* delete_hnode                                                               */
/*                                                                            */
/* Frees all memory allocation for the binary tree with root p using post-    */
/*    order traversal of the tree                                             */

extern void delete_hnode(hnode *p)
{
    if (p != NULL)
    {
	delete_hnode(p->lc_nptr);
	delete_hnode(p->rc_nptr);
	free(p);
    }
} /* delete_hnode() */


/*** The rest of the functions can be called by outside source: */

/******************************************************************************/
/*                                                                            */
/* create_heap                                                                */
/*                                                                            */
/* Creates the heap *hptr, setting the root's region data to newrgn and set-  */
/*   ting availptr to rootptr.  availptr always points to the node to which   */
/*   a new leaf can be added to the heap so that the resulting tree will      */
/*   still be a complete, balanced binary tree.  lastptr is also set rootptr  */
/*   since lastptr always must point to that most recently added node to the  */
/*   heap.                                                                    */

void create_heap(heap_t *hptr, region_t newrgn)
{
    /* Allocate space for the root node */
    hptr->rootptr = malloc( SZ_HNODE );
    if ( !(hptr->rootptr) )
    {
	pi_sys_terminate("Memory allocation error in create_heap()");
	return;
    }

    /* Set fields of *rootptr */
    (hptr->rootptr)->rgn = newrgn;
    (hptr->rootptr)->p_nptr = NULL;
    (hptr->rootptr)->lc_nptr = NULL;
    (hptr->rootptr)->rc_nptr = NULL;

    hptr->lastptr = hptr->availptr  =  hptr->rootptr;
} /* create_heap() */


/******************************************************************************/
/*                                                                            */
/* replace_root                                                               */
/*                                                                            */
/* Given the heap with root pointed to by rootptr, this function sets the     */
/* root's rgn (region) data to newrgn and then readjusts the complete,        */
/* balanced tree so that it again satisfies the "heap condition".  So, the    */
/* algorithm starts at the `top' of the tree and shifts data `down' as neces- */
/* sary until the newly added region data is in the correct position to give  */
/* give a heap.                                                               */

void replace_root( heap_t *hptr, region_t newrgn )
{
    hnode *l_child, *r_child, *child, *parent; 
   
    /* Heapify from top down */
    pi_base_t abserr = newrgn.rgnerr;
    parent = hptr->rootptr;
    l_child = (hptr->rootptr)->lc_nptr;
    r_child = (hptr->rootptr)->rc_nptr;

    while (!( l_child==NULL && r_child==NULL ))
    {
	if ( r_child == NULL )
	    child = l_child;
	else if (l_child->rgn.rgnerr > r_child->rgn.rgnerr) 
	    child = l_child;
	else
	    child = r_child;

	if (child->rgn.rgnerr > abserr)
	{
	    parent->rgn = child->rgn;
	    parent = child;
	    l_child = parent->lc_nptr;
	    r_child = parent->rc_nptr;
	}
	else
	{
	    break;
	}
    }
  
    parent->rgn = newrgn;
} /* replace_root() */



/******************************************************************************/
/*                                                                            */
/* add_heap                                                                   */
/*                                                                            */
/* Attaches a new leaf to the heap *hptr at the position required in order    */
/* to maintain a complete, balanced binary tree and then re-heapifies the     */
/* tree.  "Re-heapifies" means that the algorithm starts at the new leaf and  */
/* shifts data `up' the tree as necessary until the newly added region data   */
/* is in the correct position to give a heap.                                 */
/*                                                                            */
/* If heap not empty, then */
/* availptr points to the node to which the new leaf should be attached.  If  */
/* the new leaf is added as a right-child of the node, then a new available   */
/* parent node must be found (to be used in the next call to add_heap) by a   */
/* call to get_new_avail_parent.                                              */
/*                                                                            */
/* Sets lastptr to point to the new leaf.                                     */
/* Note: hptr may now be empty when this is called */
void add_heap( heap_t *hptr, region_t newrgn )
{
    hnode *p, *add_to, *parent;
    pi_base_t abserr;

    if (heapEmpty(*hptr))
    {
	create_heap(hptr, newrgn);
	return;
    }

    /* -- implicit else -- */

    /* Allocate space for the new leaf */

    if ( !(p=malloc( SZ_HNODE )) )
    { 
	pi_sys_terminate("Memory allocation error in add_heap()");
	return;
    }

    hptr->lastptr = p;
    add_to = hptr->availptr;

    /* Set the new node's links */
    p->p_nptr = add_to;
    p->lc_nptr = p->rc_nptr = NULL;

    /* Attach new node as correct child of *availptr */
    if ( add_to->lc_nptr == NULL )
    {
	add_to->lc_nptr = p;
    }
    else
    {
	add_to->rc_nptr = p;
	get_new_avail_parent(hptr);
    }

    /* Now re-heapify working bottom to top */
    abserr = newrgn.rgnerr;
    parent = add_to;
    while ( (parent != NULL) && (abserr > parent->rgn.rgnerr))
    {
	p->rgn = parent->rgn;
	p = parent;
	parent = p->p_nptr;
    } /* End of while another level up needs processing */

    p->rgn = newrgn;
} /* add_heap() */


/******************************************************************************/
/*                                                                            */
/* heapDeleteMax                                                              */
/*                                                                            */
/* Removes the region at the root of the heap; moves the data from the node   */
/* pointer to by the heap's lastptr; and then re-adjusts the tree so that it  */
/* again satisfies the "heap condition".                                      */
/*                                                                            */
/* Calls heapRemoveLast.  This function changes lastptr and may change        */
/* availptr.                                                                  */
/*                                                                            */
/* Calls replace_root to do the actual top-down heapification                 */

void heapDeleteMax( heap_t *hptr, region_t *maxptr )         
{
    region_t tmprgn;

    pi_assert(hptr->rootptr != NULL, "heapDeleteMax() called when heap empty\n");
   
    *maxptr = hptr->rootptr->rgn; /* The max region to be returned           */  
    
    /* Save region from last node; remove this last node, finding new last     */
    /* node and, possibly, adjusting available parent pointer.  Finally, if    */
    /* the heap is not empty, use the last node's region in replace_root which */
    /* re-heapifies from top down.                                             */ 
    tmprgn = hptr->lastptr->rgn;                                                   

    heapRemoveLast( hptr );
    /* if root node was not the last node, find place for last node's rgn      */
    if ( hptr->rootptr != NULL )    
	replace_root( hptr, tmprgn );
    /* else we have just removed the last node of the heap                     */

} /* heapDeleteMax() */


/******************************************************************************/
/*                                                                            */
/* heapRemoveLast                                                             */
/*                                                                            */
/* Removes node *lastptr from the heap.  Updates heap's lastptr and, if       */
/* necessary, updates the heap's availptr.                                    */

void heapRemoveLast( heap_t *hptr )
{
    hnode *p, *child, *parent;

    p = child = hptr->lastptr;
    parent = child->p_nptr;

    /* Take care of root node separately */
    if ( parent == NULL )
    {
	hptr->rootptr = NULL;
	free(p);
	return;
    }

    /* -- implicit else -- */

    /* Remove general leaf, resetting lastptr and, possibly, availptr.         */
    if ( hptr->availptr != parent )   
    { 
	/* *lastptr is a right child node, so, availptr should pt to the 
	   parent of *lastptr: */ 
	hptr->availptr = parent; 
	hptr->availptr->rc_nptr = NULL;
	hptr->lastptr = hptr->availptr->lc_nptr;
    }
    else
    {
	/* availptr is unchanged but a new lastptr must be calculated */
	while ( (child == parent->lc_nptr) && (parent->p_nptr != NULL) )
	{
	    child = parent;
	    parent = child->p_nptr;
	}
	if (child == parent->rc_nptr)
	    parent = parent->lc_nptr ; 
	/* else parent is the root & child is left child: no change of parent*/
  
	/* Now that the pivot is done, go the right */
	child = parent->rc_nptr;
	/* Go to the right most leaf */
	while (child != NULL )
	{
	    parent = child;
	    child = parent->rc_nptr;
	}
	hptr->lastptr = parent;
	hptr->availptr->lc_nptr = NULL; /* Detach old *lastptr            */
    }
    free(p);
    
} /*  End of function `heapRemoveLast'  */



/******************************************************************************/
/*                                                                            */
/* heapEmpty                                                                  */
/*                                                                            */
/* Returns TRUE if heap is empty; FALSE, otherwise                            */

extern bool heapEmpty(heap_t hp) 
{
    return (hp.rootptr == NULL);
} /*  End of function `heapEmpty'  */



/* Actually deletes the heap pointed to by hp using a delete_hnode() call
 */
extern void delete_heap(heap_t *hp)
{
    if (hp != NULL)
    {
	delete_hnode(hp->rootptr);
	hp->rootptr  = NULL;
	hp->availptr = NULL;
	hp->lastptr  = NULL;
    }
} /* delete_heap() */


/* Initializes a heap structure to be an empty heap
 */
extern void create_heap_empty(heap_t *heap_p)
{
    heap_p->rootptr = heap_p->availptr = heap_p->lastptr = NULL;
} /* create_heap_empty() */


/* This is a crude heap merge routine.  It removes every region in the hp_src,
 * and adds them to hp_dst.
 */
extern void heap_merge(heap_t *hp_dst, heap_t *hp_src)
{
    region_t rgn;
    
    while (!heapEmpty(*hp_src))
    {
	heapDeleteMax(hp_src, &rgn);
	add_heap(hp_dst, rgn);
    }
} /* heap_merge() */

