/*        File: heapdeap.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: All
   Description: Source for functions that manipulate the heap/deap data structure
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/misc/heapdeap.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:31 $

                $Log: heapdeap.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:31  parint
                Initial checkin


       Version: 02/18/00  rrz  Created from deap.c and heap.c
*/

#include <stdio.h>
#include <stdlib.h>

#define USE_MY_ASSERTIONS

#include <pi-defs.h>
#include <parint.h>
#include <heapdeap.h>

#define SZ_HNODE (sizeof(hnode))

/************************************************************************************************/
/*** Internal routines for heaps and deaps                                                    ***/
/************************************************************************************************/

/* Recursive; used to free memory used by a heap. p may be NULL.
 * Works on heaps or deaps
 */
static void delete_hnode(hnode *p)
{
    if (p != NULL)
    {
	delete_hnode(p->lc_nptr);
	delete_hnode(p->rc_nptr);
	free(p);
    }
} /* delete_hnode() */


/* Recursive routine; used for debugging to print out max heaps, and either
 * the min or max version of the deap.
 */
static void print_heap_ll(hnode *hnode_p, int id, int depth, bool print_extra_deap_info)
{
    printf("[HP %d]:%*chnode: %s", id, 2 * depth, ' ', print_region(&(hnode_p->rgn)));
    if (print_extra_deap_info)
    {
	if (hnode_p->partner)
	    printf("; partner rgn %G\n", (double) hnode_p->partner->rgn.rgnerr);
	else
	    printf("; partner NULL\n");
    }
    else
    {
	printf("\n");
    }
    if (hnode_p->lc_nptr)
        print_heap_ll(hnode_p->lc_nptr, id, depth + 1, print_extra_deap_info);
    if (hnode_p->rc_nptr)
        print_heap_ll(hnode_p->rc_nptr, id, depth + 1, print_extra_deap_info);
} /* print_heap_ll() */



/************************************************************************************************/
/*** Internal routines for heaps only                                                         ***/
/************************************************************************************************/


/* Removes node *lastptr from the heap.  Updates heap's lastptr and, if
 * necessary, updates the heap's availptr.
 */
static void heap_remove_last(heapdeap_t *hptr)
{
    hnode *p, *child, *parent;

    pi_assert(!(hptr->is_a_deap), "heap_remove_last() called on a deap");

    p = child = hptr->lastptr;
    parent = child->p_nptr;

    /* Take care of root node separately */
    if ( parent == NULL )
    {
	hptr->maxrtptr = NULL;
	free(p);
	return;
    }

    /* -- implicit else -- */

    /* Remove general leaf, resetting lastptr and, possibly, availptr.         */
    if ( hptr->availptr != parent )   
    { 
	/* *lastptr is a right child node, so, availptr should pt to the 
	   parent of *lastptr: */ 
	hptr->availptr = parent; 
	hptr->availptr->rc_nptr = NULL;
	hptr->lastptr = hptr->availptr->lc_nptr;
    }
    else
    {
	/* availptr is unchanged but a new lastptr must be calculated */
	while ( (child == parent->lc_nptr) && (parent->p_nptr != NULL) )
	{
	    child = parent;
	    parent = child->p_nptr;
	}
	if (child == parent->rc_nptr)
	    parent = parent->lc_nptr ; 
	/* else parent is the root & child is left child: no change of parent*/
  
	/* Now that the pivot is done, go the right */
	child = parent->rc_nptr;
	/* Go to the right most leaf */
	while (child != NULL )
	{
	    parent = child;
	    child = parent->rc_nptr;
	}
	hptr->lastptr = parent;
	hptr->availptr->lc_nptr = NULL; /* Detach old *lastptr            */
    }
    free(p);
    
}  /* heap_remove_last() */


/* Given the heap with root pointed to by maxrtptr, this function sets the
 * root's rgn (region) data to newrgn and then readjusts the complete,
 * balanced tree so that it again satisfies the "heap condition".  So, the
 * algorithm starts at the `top' of the tree and shifts data `down' as neces-
 * sary until the newly added region data is in the correct position to give
 * give a heap.
 */
static void ll_heap_replace_root(heapdeap_t *hptr, const region_t *newrgn_p)
{
    hnode *l_child, *r_child, *child, *parent; 
   
    /* Heapify from top down */
    pi_base_t abserr = newrgn_p->rgnerr;

    pi_assert(!(hptr->is_a_deap), "ll_heap_replace_root() called on a deap");

    parent = hptr->maxrtptr;
    l_child = (hptr->maxrtptr)->lc_nptr;
    r_child = (hptr->maxrtptr)->rc_nptr;

    while (!( l_child==NULL && r_child==NULL ))
    {
	if ( r_child == NULL )
	    child = l_child;
	else if (l_child->rgn.rgnerr > r_child->rgn.rgnerr) 
	    child = l_child;
	else
	    child = r_child;

	if (child->rgn.rgnerr > abserr)
	{
	    parent->rgn = child->rgn;
	    parent = child;
	    l_child = parent->lc_nptr;
	    r_child = parent->rc_nptr;
	}
	else
	{
	    break;
	}
    }
  
    parent->rgn = *newrgn_p;
} /* ll_heap_replace_root() */


/* Removes the region at the root of the heap; moves the data from the node
 * pointer to by the heap's lastptr; and then re-adjusts the tree so that it
 * again satisfies the "heap condition".
 * Calls heap_remove_last.  This function changes lastptr and may change availptr.
 * Calls replace_root to do the actual top-down heapification
 */
static void ll_heap_delete_max(heapdeap_t *hptr, region_t *maxptr)
{
    region_t tmprgn;

    pi_assert(hptr->maxrtptr != NULL, "ll_heap_delete_max() called when heap empty\n");
    pi_assert(!(hptr->is_a_deap), "ll_heap_delete_max() called on a deap");
    
    *maxptr = hptr->maxrtptr->rgn; /* The max region to be returned           */  
    
    /* Save region from last node; remove this last node, finding new last     */
    /* node and, possibly, adjusting available parent pointer.  Finally, if    */
    /* the heap is not empty, use the last node's region in replace_root which */
    /* re-heapifies from top down.                                             */ 
    tmprgn = hptr->lastptr->rgn;                                                   

    heap_remove_last(hptr);
    /* if root node was not the last node, find place for last node's rgn      */
    if (hptr->maxrtptr != NULL)
	ll_heap_replace_root(hptr, &tmprgn);
    /* else we have just removed the last node of the heap                     */

} /* ll_heap_delete_max() */



/* Pre-condition: hptr->availptr points to the node to which the last two     */
/*    leaves have been added (as its right and left children).                */
/* Post-condition: hptr->availptr points to the node to which next leaf will  */
/*    be added (as a left child) in order to maintain a complete, balanced    */
/*    binary tree.                                                            */
static void heap_get_new_avail_parent(heapdeap_t *hptr)
{
    hnode *child, *parent;
    
    pi_assert(!(hptr->is_a_deap), "heap_get_new_avail_parent() called on a deap");

    /* If the current available parent is the heap's root, then make the       */
    /* available parent the root's left child                                  */
    if ((hptr->availptr)->p_nptr == NULL)
    {
	hptr->availptr = (hptr->availptr)->lc_nptr;
	return;
    }

    child = hptr->availptr;
    parent = child->p_nptr;

    /* Go up the tree until a change from child is right child to child is     */
    /*   left child or until root node.                                        */
    while ( (parent->p_nptr != NULL) && (child == parent->rc_nptr))
    {
	child = parent;
	parent = child->p_nptr;
    }

    /* Now go back down the tree to the left from **appropriate** parent to    */
    /*   find available parent node.                                           */
    if ( (parent->p_nptr != NULL) || (child != parent->rc_nptr) )
    {
	/* Have not reached root from the rght: change parent for descent       */
	parent = parent->rc_nptr;
    }
    /* else at root node and arrived from right: don't change parent           */

    while ( parent->lc_nptr != NULL )
	parent = parent->lc_nptr;
    hptr->availptr = parent;

}  /* heap_get_new_avail_parent() */


/* Low-level routine to add a node to a heap.
 * Works on heaps only
 */
static void ll_heap_add_node(heapdeap_t *hptr, const region_t *newrgn_p)
{
    hnode *p, *add_to, *parent;
    pi_base_t abserr;

    pi_assert(!(hptr->is_a_deap), "ll_heap_add_node() called on a deap");

    if (heap_empty(hptr))
    {
	heap_create(hptr, FALSE);

	/* Manually add the first region to the heap; Allocate space for the root node: */
	hptr->maxrtptr = malloc(SZ_HNODE);
	if (!(hptr->maxrtptr))
	{
	    pi_sys_terminate("Memory allocation error in create_heap()");
	    return;
	}

	/* Set fields of *maxrtptr */
	(hptr->maxrtptr)->rgn = *newrgn_p;
	(hptr->maxrtptr)->p_nptr = NULL;
	(hptr->maxrtptr)->lc_nptr = NULL;
	(hptr->maxrtptr)->rc_nptr = NULL;
	hptr->lastptr = hptr->availptr = hptr->maxrtptr;
	return;
    }

    /* -- implicit else -- */

    /* Allocate space for the new leaf */
    if ( !(p=malloc( SZ_HNODE )) )
    { 
	pi_sys_terminate("Memory allocation error in add_heap()");
	return;
    }

    hptr->lastptr = p;
    add_to = hptr->availptr;

    /* Set the new node's links */
    p->p_nptr = add_to;
    p->lc_nptr = p->rc_nptr = NULL;

    /* Attach new node as correct child of *availptr */
    if ( add_to->lc_nptr == NULL )
    {
	add_to->lc_nptr = p;
    }
    else
    {
	add_to->rc_nptr = p;
	heap_get_new_avail_parent(hptr);
    }

    /* Now re-heapify working bottom to top */
    abserr = newrgn_p->rgnerr;
    parent = add_to;
    while ( (parent != NULL) && (abserr > parent->rgn.rgnerr))
    {
	p->rgn = parent->rgn;
	p = parent;
	parent = p->p_nptr;
    } /* End of while another level up needs processing */

    p->rgn = *newrgn_p;
} /* ll_heap_add_node() */


/************************************************************************************************/
/*** Internal routines for deaps only                                                         ***/
/************************************************************************************************/


/* Pre-condition: availptr points to the node to which the last two leaves    */
/*    have been added (as right and left children); maxhp = 1 iff availptr    */
/*    points to a node in the max-heap.                                       */
/* Post-condition: availptr points to the node to which next leaf will be     */
/*    added (as a left child) in order to maintain a complete balanced        */
/*    binary tree; maxhp = 1 iff new value of availptr pts to a max-heap      */   
/*    node.                                                                   */
static void deap_get_new_avail_parent(heapdeap_t *d)
{
    hnode  *child, *parent;

    pi_assert(d->is_a_deap, "deap_get_new_avail_parent() called on a heap");

    /* If the current available parent is the min-heap's root, then make the   */
    /* available parent the max-heap's root.                                   */
    if (d->availptr == d->minrtptr)
    {
	d->availptr = d->maxrtptr;
	d->maxhp = 1;
	return;
    }
    /* If the current available parent is the max-heap's root                  */
    if (d->availptr == d->maxrtptr)
    {
	d->availptr = d->minrtptr->lc_nptr;
	d->maxhp = 0;
	return;
    }
    child = d->availptr;
    parent = child->p_nptr;

    /* Go up the tree until a change from child is right child to child is     */
    /*   left child or until root node.                                        */
    while ( (parent->p_nptr != NULL) && (child == parent->rc_nptr))
    {
	child = parent;
	parent = child->p_nptr;
    }

    /* Now go back down the tree to the left from **appropriate** parent to    */
    /*   find available parent node.                                           */
    if ( (parent->p_nptr != NULL) || (child != parent->rc_nptr) )
    {
	/* Have not reached a root from the rght: change parent for descent     */
	parent = parent->rc_nptr;
    }
    else /* at a root node and arrived from right: change min-max subtrees     */
    {
	d->maxhp = 1 - d->maxhp;
	if ( parent == d->minrtptr ) parent = d->maxrtptr->lc_nptr;
	else parent = d->minrtptr->lc_nptr;
    }

    while (parent->lc_nptr != NULL)
	parent = parent->lc_nptr;
    d->availptr = parent;

} /* deap_get_new_avail_parent() */


/* Returns a 1 if *availptr is a node in the max-heap; returns 0, otherwise.
 */
static int deap_which_heap(const heapdeap_t *deap_p)
{
    hnode *q;
    pi_assert(!heap_empty(deap_p), "deap_which_heap() called with empty deap");
    pi_assert(deap_p->is_a_deap, "deap_which_heap() called on a heap");
    q = deap_p->availptr;
    while ( q->p_nptr != NULL )
	q = q->p_nptr;
    return ( q == deap_p->maxrtptr ) ? 1 : 0 ;
} /* deap_which_heap() */


/* Given the region r and the empty leaf node in the min-heap pointed to
 * by p, MinInsert shifts data down along the path to the root until the
 * "correct" place (so that the deap's right subtree will satisfy the   
 * "min-heap condition") for r is found.                                
 * For deaps only.
 */
static void deap_min_insert(hnode *p, region_t r)
{
    pi_base_t abserr;
    hnode     *parent;

    /* Re-heapify working bottom to top */
    abserr = r.rgnerr;
    parent = p->p_nptr;
    while ( (parent != NULL) && (abserr < parent->rgn.rgnerr) )
    {
	p->rgn = parent->rgn;
	p = parent;
	parent = p->p_nptr;
    }

    p->rgn = r;

} /* deap_min_insert() */


/* Given the region r and the empty leaf node in the max-heap pointed to
 * by p, MaxInsert shifts data down along the path to the root until the
 * "correct" place (so that the deap's left subtree will satisfy the    
 * "max-heap condition") for r is found.
 * For deaps only.
*/
static void deap_max_insert(hnode *p, region_t r)
{
    pi_base_t abserr;
    hnode     *parent; 

    /* Re-heapify working bottom to top */
    abserr = r.rgnerr;
    parent = p->p_nptr;
    while ( (parent != NULL) && (abserr > parent->rgn.rgnerr) )
    {
	p->rgn = parent->rgn;
	p = parent;
	parent = p->p_nptr;
    }

    p->rgn = r;
} /* deap_max_insert() */


static void ll_deap_add_node(heapdeap_t *d, region_t r)
{
    hnode *p, *attach_to, *ppp;
    int max;

    pi_assert(d->is_a_deap, "ll_deap_add_node() called on a heap");

    /* Allocate space for new node */
    if (!(p = malloc(SZ_HNODE)))
    {
	pi_sys_terminate("Memory allocation error in ll_deap_add_node()");
	return;
    }

    /* Set tree structure components of new node and deap *d */
    p->lc_nptr = NULL;
    p->rc_nptr = NULL;
    d->lastptr = p;

    /* Handle the beginning cases */
    if (d->minrtptr == NULL)
    {
	p->rgn = r;                   /* Note: availptr not set in this case    */
	p->p_nptr = p->partner = NULL;
	d->minrtptr = p;      
	return;
    }
    if (d->maxrtptr == NULL)
    {
	p->p_nptr = NULL;
	p->partner = d->availptr = d->minrtptr;
	d->minrtptr->partner = d->maxrtptr = p;
	d->maxhp = 0;                         /* availptr pts to min-heap       */
	/*  Assure that min/max-heap relationship maintained */
	if ( d->minrtptr->rgn.rgnerr > r.rgnerr )
	{
	    p->rgn = d->minrtptr->rgn;
	    d->minrtptr->rgn = r;
	}
	else
	{
	    p->rgn = r;
	}
	return;
    }

    /* Handle regular insert into the deap */
    max = d->maxhp;                          /* Save current value of maxhp    */
    p->p_nptr = attach_to = d->availptr;     /* Attach *p to the deap & save   */
    if ( attach_to->lc_nptr == NULL )        /*   current value of availptr    */
    {
	attach_to->lc_nptr = p;
    }
    else
    {
	attach_to->rc_nptr =p;
	deap_get_new_avail_parent(d);         /* Updates availptr & possibly maxhp    */
    }
    /* Set new node's partner and enter region r at correct node */
    if ( max )
    {   /* New node attached to max-heap */
	if ( attach_to->rc_nptr == NULL )
	    ppp = p->partner = attach_to->partner->lc_nptr;
	else 
	    ppp = p->partner = attach_to->partner->rc_nptr; 
	/* Update p's partner's partner */
	ppp->partner = p; 
	if ( r.rgnerr < ppp->rgn.rgnerr )
	{
	    /* r does not belong in the max-heap */
	    p->rgn = ppp->rgn;
	    deap_min_insert( ppp, r );
	}
	else   /* r belongs someplace within the max-heap */
	{
	    deap_max_insert( p, r );
	}
    }   /* End of: new node attached to max-heap portion of the deap */
    else
    {   /* New node attached to min-heap */
	ppp = p->partner = attach_to->partner;
	if ( r.rgnerr > ppp->rgn.rgnerr )
	{
	    /* r does not belong in the min-heap */
	    p->rgn = ppp->rgn;
	    deap_max_insert( ppp, r );
	}
	else   /* r belongs someplace within the min-heap */
	{
	    deap_min_insert( p, r );
	}
    }   /* End of: new node attached to min-heap portion of the deap */

} /* ll_deap_add_node() */


/* Returns 1 if the max-heap is empty; 0, otherwise
 */
static int deap_max_heap_empty(const heapdeap_t *deap_p)
{
    pi_assert(deap_p->is_a_deap, "deap_max_heap_empty() called on a heap");
    if (deap_p->maxrtptr == NULL)
	return 1; 
    else
	return 0;
}  /* deap_max_heap_empty() */


/* Removes node *lastptr from the deap.  Updates deap's lastptr and, if
 * necessary, updates the deap's availptr and partner's partner-pointer.
 */
static void deap_remove_last_node(heapdeap_t *d)
{
    hnode *p, *child, *parent;

    pi_assert(d->is_a_deap, "deap_remove_last_node() called on a heap");

    p = child = d->lastptr;
    parent = child->p_nptr;

    /* Take care of root nodes separately */
    if ( parent == NULL )
    {
	if ( p == d->minrtptr )
	{
	    d->minrtptr = NULL;
	}
	else
	{
	    d->maxrtptr = NULL;
	    d->lastptr = d->minrtptr;
	    d->minrtptr->partner = NULL;
	    d->maxhp = 0;
	}
	free(p);
	return;
    }
    /* Remove general leaf, resetting lastptr and, possibly, availptr.         */
    /* Adjust *lastptr's partner's partner (sometimes unnecessarily).          */
    p->partner->partner = parent;

    if ( d->availptr != parent )
    {      /* *lastptr is a right child node    */
	d->availptr = parent;              /* So, availptr should pt to the     */
	d->availptr->rc_nptr = NULL;       /*   parent of *lastptr              */ 
	d->lastptr = d->availptr->lc_nptr;
	d->maxhp = deap_which_heap(d);
    }
    else
    {  /* availptr is unchanged but a new lastptr must be calculated    */
	/* Find new *lastptr */
	while ( (child == parent->lc_nptr) && (parent->p_nptr != NULL) )
	{
	    child = parent;
	    parent = child->p_nptr;
	}
	if (child == parent->rc_nptr)
	    parent = parent->lc_nptr ; 
	else                        /* parent is a root and child is left child */
	    parent = d->maxhp  ?  d->minrtptr   :   d->maxrtptr  ;
	/* Now that the pivot is done, go the right */
	child = parent->rc_nptr;
	/* Go to the right most leaf */
	while (child != NULL )
	{
	    parent = child;
	    child = parent->rc_nptr;
	}
	d->lastptr = parent;
	d->availptr->lc_nptr = NULL;                     /* Detach old *lastptr */
    }
    free(p);

} /* deap_remove_last_node() */


/* Removes the region at the root of the max-heap.                            */
/*                                                                            */
/* Shifts data up the max-heap along the correct path to a leaf node, leav-   */
/* ing an empty leaf node.  Then, deletes the node pointed to by lastptr,     */
/* finding the correct place for this deleted node's region by a strategy     */
/* similar to the one used in inserting a new region into the deap.           */
/*                                                                            */
/* Calls either MinInsert or MaxInsert depending on the relationship between  */
/* the rgnerr from the deleted node and the value of rgnerr in the empty      */
/* node's partner node.                                                       */
/*                                                                            */
/* Calls RemoveLastNode.  This function changes lastptr and may change        */
/* availptr and maxhp.                                                        */
static void ll_deap_delete_max(heapdeap_t *d, region_t *maxptr)         
{
    hnode *parent, *child, *l_child, *r_child, *empty, *inverse;
    region_t tmprgn;

    pi_assert(!heap_empty(d), "DeapDeleteMax() called with empty deap");
    pi_assert(d->is_a_deap, "ll_deap_delete_max() called on a heap");
    
    if ( deap_max_heap_empty(d) )
    {
	// printf("special case: max heap empty\n");
	*maxptr = d->minrtptr->rgn;
	free(d->minrtptr);
	d->minrtptr = NULL;
	return;
    }
    else if (d->lastptr == d->maxrtptr)
    {
	/* Only a single node in the max heap, so remove the max heap: */
	// printf("special case: maxrtptr is lastptr\n");
	*maxptr = d->maxrtptr->rgn;
	free(d->maxrtptr);
	d->maxrtptr = NULL;
	d->minrtptr->partner = NULL;
	d->availptr = NULL;
	d->lastptr = d->minrtptr;
	return;
    }

    *maxptr = d->maxrtptr->rgn;        
   
    /* General case of deleting a node:                                        */
    /* First, "shift up" along correct path from root to leaf within max-heap, */
    /* creating an empty leaf                                                  */
    parent = d->maxrtptr;
    l_child = parent->lc_nptr;
    r_child = parent->rc_nptr;

    while (!( l_child==NULL && r_child==NULL ))
    {
	/* Decide the direction down */
	if ( r_child == NULL )
	    child = l_child;
	else if (l_child->rgn.rgnerr > r_child->rgn.rgnerr) 
	    child = l_child;
	else
	    child = r_child;
	/* Shift regions up */
	parent->rgn = child->rgn;
	parent = child;
	l_child = parent->lc_nptr;
	r_child = parent->rc_nptr;
    }  /*  end of while still more levels to process  */

    /* Second, delete node *lastptr and insert its region, using just created  */ 
    /* empty node as starting point for the insert.                            */
    empty = parent;
    tmprgn = d->lastptr->rgn;                                                      

    /* Special case when the empty node is the right-most leaf.  In this case, empty
       will end up being deleted, so set it to be its parent: */
    if (empty == d->lastptr)
    {
	pi_assert(empty->p_nptr, "uh oh");
	inverse = empty->p_nptr->partner;
	empty = empty->p_nptr;
	// printf("Special case #1\n");
    }
    else
    {
	inverse = empty->partner;
    }
    
    deap_remove_last_node(d);

    l_child = inverse->lc_nptr;
    r_child = inverse->rc_nptr;
    if ( ! (l_child==NULL && r_child==NULL) )
    {
	if ( r_child==NULL )
	    inverse = l_child;
	else if (l_child->rgn.rgnerr > r_child->rgn.rgnerr )
	    inverse = l_child;
	else
	    inverse = r_child;
    }

    if (tmprgn.rgnerr < inverse->rgn.rgnerr)
    {
	/* Empty node's partner's largest child's region belongs in empty node  */
	/* and tmprgn belongs in min-heap along path starting at that child's   */
	/* node.                                                                */
	// printf("case 2a\n");
	empty->rgn = inverse->rgn;
	deap_min_insert(inverse, tmprgn);
    }
    else
    {
	/* tmprgn belongs in the max-heap along the path starting at empty node */
	/* Not sure about the empty_parent_p pointer being used here */
	// printf("case 2b\n");
	deap_max_insert(empty, tmprgn);
    }

} /* ll_deap_delete_max() */



/************************************************************************************************/
/*** Heap only external routines                                                              ***/
/************************************************************************************************/

extern void heap_print(heapdeap_t *hpdp_p, int id)
{
    pi_assert(!(hpdp_p->is_a_deap), "Not supported yet!");
    if (heap_empty(hpdp_p))
        printf("[HP %d]: heap is empty\n", id);
    else if (hpdp_p->is_a_deap && hpdp_p->maxrtptr == NULL)
        printf("[HP %d]: heap/deap is sort-of empty\n", id);
    else
        print_heap_ll(hpdp_p->maxrtptr, id, 1, FALSE);
} /* heap_print() */


/************************************************************************************************/
/*** Heap or deap external routines                                                           ***/
/************************************************************************************************/

/* Initializes a heap structure to be an empty heap.
 * Works on heaps or deaps
 */
extern void heap_create(heapdeap_t *hpdp_p, bool is_a_deap)
{
    hpdp_p->maxrtptr = hpdp_p->availptr = hpdp_p->lastptr = NULL;
    hpdp_p->minrtptr = NULL;
    hpdp_p->maxhp = -1;
    hpdp_p->is_a_deap = is_a_deap;
    hpdp_p->num_nodes = 0;
} /* heap_create() */


/* Returns TRUE if heap is empty, FALSE otw.  The check of minrtptr when hpdp_p is a deap is
 * *exactly* what the old deap code did; I'm not sure why they did not just use maxrtptr, 
 * but I don't want to change it if it ain't broke!
 * Works on heaps or deaps
 */
extern bool heap_empty(const heapdeap_t *hpdp_p) 
{
    return (hpdp_p->is_a_deap 
	    ? (hpdp_p->minrtptr == NULL)
	    : (hpdp_p->maxrtptr == NULL));
} /*  heap_empty() */


/* Returns the key of the max region on the heap
 * Works on heaps or deaps
 */
extern pi_base_t heap_max_key(const heapdeap_t *hpdp_p)
{
    pi_assert(!heap_empty(hpdp_p), "Called heap_max_key() with empty heap!");
    return (hpdp_p->is_a_deap
	    ? ((hpdp_p->maxrtptr != NULL)
	       ? (hpdp_p->maxrtptr->rgn).rgnerr
	       : (hpdp_p->minrtptr->rgn).rgnerr)
	    : (hpdp_p->maxrtptr->rgn).rgnerr);
} /*  heapdeap_t() */


/* This is a crude heap merge routine.  It removes every region in the hp_src,
 * and adds them to hp_dst.
 * Works on heaps or deaps
 */
extern void heap_merge(heapdeap_t *hp_dst, heapdeap_t *hp_src)
{
    region_t rgn;
    
    while (!heap_empty(hp_src))
    {
	heap_delete_max(hp_src, &rgn);
	heap_add_node(hp_dst, &rgn);
    }
} /* heap_merge() */


/* Free()'s all the memory used by the heap; calls delete_hnode() recursively.
 * Works on heaps or deaps
 */
extern void heap_free(heapdeap_t *hpdp_p)
{
    if (hpdp_p != NULL)
    {
	delete_hnode(hpdp_p->maxrtptr);
	hpdp_p->maxrtptr  = NULL;
	hpdp_p->availptr = NULL;
	hpdp_p->lastptr  = NULL;
	if (hpdp_p->is_a_deap)
	{
	    hpdp_p->minrtptr = NULL;
	    hpdp_p->is_a_deap = FALSE;
	    hpdp_p->maxhp = -1;
	    hpdp_p->num_nodes = 0;
	}
    }
} /* heap_free() */


/* Old comments:
 * Attaches a new leaf to the heap *hptr at the position required in order to maintain a complete,
 * balanced binary tree and then re-heapifies the tree.  "Re-heapifies" means that the algorithm
 * starts at the new leaf and shifts data `up' the tree as necessary until the newly added region
 * data is in the correct position to give a heap. If heap not empty, then availptr points to 
 * the node to which the new leaf should be attached. If the new leaf is added as a right-child
 * of the node, then a new available parent node must be found (to be used in the next call to
 * heap_add_node() by a call to get_new_avail_parent. Sets lastptr to point to the new leaf.
 * RRZ Note: hptr may now be empty when this is called.
 * Works on heaps or deaps
 */
extern void heap_add_node(heapdeap_t *hpdp_p, region_t *newrgn_p)
{
    if (hpdp_p->is_a_deap)
	ll_deap_add_node(hpdp_p, *newrgn_p);
    else
	ll_heap_add_node(hpdp_p, newrgn_p);
    ++(hpdp_p->num_nodes);
} /* heap_add_node() */

/* Note: no change to hpdp_p->num_nodes on this call */
extern void heap_replace_root(heapdeap_t *hpdp_p, const region_t *newrgn_p)
{
    pi_assert(!heap_empty(hpdp_p), "heap_replace_root() called on empty heap");
    if (hpdp_p->is_a_deap)
    {
	region_t rgn;
	ll_deap_delete_max(hpdp_p, &rgn);
	ll_deap_add_node(hpdp_p, *newrgn_p);
    }
    else
    {
	ll_heap_replace_root(hpdp_p, newrgn_p);
    }
} /* heap_replace_root() */


extern void heap_delete_max(heapdeap_t *hpdp_p, region_t *maxrgn_p)
{
    pi_assert(!heap_empty(hpdp_p), "heap_delete_max() called on empty heap");
    if (hpdp_p->is_a_deap)
	ll_deap_delete_max(hpdp_p, maxrgn_p);
    else
	ll_heap_delete_max(hpdp_p, maxrgn_p);
    --(hpdp_p->num_nodes);
} /* heap_delete_max() */


extern char *print_region(region_t *rgn_p)
{
    static char result[100];
    sprintf(result, "<RGN %G-%G %G-%G; %1d;%G>",
            (double) (rgn_p->center[0] - rgn_p->h_width[0]), 
	    (double) (rgn_p->center[0] + rgn_p->h_width[0]),
            (double) (rgn_p->center[1] - rgn_p->h_width[1]), 
	    (double) (rgn_p->center[1] + rgn_p->h_width[1]),
            rgn_p->divaxn, (double) rgn_p->rgnerr);
    return result;
} /* print_region() */


extern region_t *heap_peek_max(heapdeap_t *hpdp_p)
{
    pi_assert(!heap_empty(hpdp_p), "heap_peek_max() called with empty heap");
    return ((hpdp_p->is_a_deap)
	    ? ((hpdp_p->maxrtptr == NULL) ? &(hpdp_p->minrtptr->rgn) : &(hpdp_p->maxrtptr->rgn))
	    : &(hpdp_p->maxrtptr->rgn));
} /* heap_peek_max() */


extern int heap_size(heapdeap_t *hpdp_p)
{
    pi_assert(hpdp_p->num_nodes >= 0, "Invalid num_nodes in heap_size()");
    return hpdp_p->num_nodes;
} /* heap_size() */



/************************************************************************************************/
/*** Deap only external routines                                                              ***/
/************************************************************************************************/

/* Returns the key of the min region on the deap
 * Works on deaps only
 */
extern pi_base_t deap_min_key(const heapdeap_t *hpdp_p)
{
    pi_assert(hpdp_p->is_a_deap, "Calling deap routine on a heap!");
    pi_assert(!heap_empty(hpdp_p), "Called deap_min_key() with empty deap!");
    pi_assert(hpdp_p->minrtptr, "heap_min_key() called with minrtptr NULL");
    return (hpdp_p->minrtptr->rgn).rgnerr;
} /* deap_min_key() */


extern void deap_print(heapdeap_t *hpdp_p, int id)
{
    bool print_extra_deap_info = FALSE;
    pi_assert(hpdp_p->is_a_deap, "deap_print() called on a heap");
    if (heap_empty(hpdp_p))
    {
        printf("[DP %d]: deap is empty\n", id);
    }
    else if (hpdp_p->is_a_deap && hpdp_p->maxrtptr == NULL)
    {
        printf("[DP %d]: heap/deap has single node\n", id);
    }
    else
    {
	printf("[DP %d]: ", id);
	if (hpdp_p->availptr)
	    printf("availptr rgn %G; ", (double) hpdp_p->availptr->rgn.rgnerr);
	else
	    printf("availptr NULL; ");
	if (hpdp_p->lastptr)
	    printf("lastptr rgn %G\n", (double) hpdp_p->lastptr->rgn.rgnerr);
	else
	    printf("lastptr NULL\n");
	printf("[DP %d] --Min Heap:\n", id);
        print_heap_ll(hpdp_p->minrtptr, id, 1, print_extra_deap_info);
	printf("[DP %d] --Max Heap:\n", id);
        print_heap_ll(hpdp_p->maxrtptr, id, 1, print_extra_deap_info);
    }
} /* deap_print() */


/* Removes the region at the root of the min-heap.                            */
/*                                                                            */
/* Shifts data up the min-heap along the correct path to a leaf node,         */
/* leaving an empty leaf node.  Then, deletes the node pointed to by          */
/* lastptr, finding the correct place for this deleted node's region by       */
/* a strategy similar to the one used in inserting a new region into the      */
/* deap.                                                                      */
/*                                                                            */
/* Calls either MinInsert or MaxInsert depending on the relationship          */
/* between the rgnerr from the deleted node and the value of rgnerr in        */
/* the empty node's partner node.                                             */
/*                                                                            */
/* Calls RemoveLastNode.  This function changes lastptr and may change        */
/* availptr and maxhp.                                                        */
/*                                                                            */

extern void deap_delete_min(heapdeap_t *d, region_t *minptr)         
{  
    hnode *parent, *child, *l_child, *r_child, *empty;
    region_t tmprgn;

    pi_assert(!heap_empty(d), "DeapDeleteMin() called with empty deap");
    pi_assert(d->is_a_deap, "deap_delete_min() called on a heap");
    
    *minptr = d->minrtptr->rgn;        
    --(d->num_nodes);
    
    /* Delete from a single node deap */
    if ( d->maxrtptr == NULL )
    {
	free(d->minrtptr);
	d->minrtptr = NULL;
	return;
    }

    /* Delete from a double node deap */
    if ( d->minrtptr->lc_nptr == NULL )
    {
	d->minrtptr->rgn = d->maxrtptr->rgn;
	free( d->maxrtptr );    
	d->maxrtptr = NULL;
	d->minrtptr->partner = NULL;
	d->maxhp = 0;
	return;
    }

    /* General case of deleting a node:                                        */
    /* First, "shift up" along correct path from root to leaf within           */
    /* min-heap, creating an empty leaf                                        */
    parent = d->minrtptr;
    l_child = parent->lc_nptr;
    r_child = parent->rc_nptr;

    while (!( l_child==NULL && r_child==NULL ))   {
	/* Decide the direction down */
	if ( r_child == NULL ) child = l_child;
	else if (l_child->rgn.rgnerr < r_child->rgn.rgnerr) 
	    child = l_child;
	else child = r_child;
	/* Shift regions up */
	parent->rgn = child->rgn;
	parent = child;
	l_child = parent->lc_nptr;
	r_child = parent->rc_nptr;
    }  /*  end of while still more levels to process  */

    /* Second, delete node *lastptr and insert its region, using just          */ 
    /* created empty node as starting point for the insert.                    */
    empty = parent;
    tmprgn = d->lastptr->rgn;                                                      

    if ( empty == d->lastptr)
    {  /* Empty node same as node to be removed    */
	deap_remove_last_node(d);
	return;
    }
    deap_remove_last_node(d);
    if ( tmprgn.rgnerr > empty->partner->rgn.rgnerr )
    {
	/* Empty node's partner's region belongs in empty node and tmprgn       */
	/* belongs in max-heap along path starting at partner node.             */
	empty->rgn = empty->partner->rgn;
	deap_max_insert( empty->partner, tmprgn );
    }
    else {
	/* tmprgn belongs in the min-heap along the path starting at empty      */
	/* node                                                                 */
	deap_min_insert( empty, tmprgn );
    }
} /* deap_delete_min() */

