/*        File: intg-ctrl.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI
   Description: Source to implement the integral controller (which can also act as an IW)
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/intg-ctrl.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:32 $

                $Log: intg-ctrl.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:32  parint
                Initial checkin


       Version: 01/13/98  rrz  Original (from existing code)
		05/12/98  rrz  Code freeze for ParInt1.0
		06/28/98  rrz  Modified call to caw_init_intg() for different
				proto (for visualization)
		7/3/98	  jli  added code for visulization
		09/17/98  rrz  Remove unused ms_type variable
		11/15/98  rrz  Add round-robin selection of next worker to
				receive work in LB (as in extrap version)
		02/17/99  rrz  Add SPECIAL_GRAPH features for plotting current
				error vs #regions (for my thesis!)
		05/06/99  rrz  Duplicate the 03/15/99 ParInt1.0 bug fix in this
				(nextrel) code
		05/27/99  rrz  Use err_ratio() and eps_ratio macros
		07/13/99  rrz  Add logging calls
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

/* #define DBG_THIS_FILE */

/* #define SPECIAL_GRAPH */

/* #define EXTRA_RR_TESTING */

#include <stdio.h>
#include <mpi.h>

#include <parint.h>
#include <math-adapt.h>
#include <math-intg.h>     /* for neval() */
#include <pi-defs.h>
#include <messaging.h>
#include <pi-vspec.h>
#include <measures.h>
#include <heap.h>
#include <visual.h> /* jli!! */
#include <logging.h>

#ifdef SPECIAL_GRAPH
#include <math-intg.h>
int global_nevals;
#endif

/* For a short while, these will be compile-time constants.  When we have cemented their
 * behavior, I will elevate them to be run-time parameters: */

/* eps will be sent down to workers when the ratio between it and the last sent
 * eps is greater than or equal to the THRESHOLD value, or, if there have been
 * FREQUENCY iterations since the last sent eps: */
#define SEND_EPS_THRESHOLD         0.01
#define BASE_SEND_EPS_FREQUENCY      10
#define MAX_EPS_BETWEEN_UPDATES       5   /* this is a total kludge */

/* A global, which breaks tradiation, but oh well.  This contains the id of the last worker
 * to receive a region via load balancing.  Selection of the next worker will begin with
 * the worker after this one: */
static int last_lb_worker_id = 0;
static int previous_last_lb_worker_id = 0;
static int nworkers_global = 0;
static int eps_sent_since_last_update = 0;

/* This design is a little ugly, but is based on the fact that I have to send the caw 
 * helper everytime, b/c I (the IC) do not know if the caw is idle or not, b/c it
 * checks its local estabs against the eps on *every* call (since it is able to).
 * But, I do not want to pass by an idle worker in the round-robin location if I
 * do not have to.  The design of how this is done is not the cleanest, and is due to 
 * the fact that all other workers communicate with the IC via messages, while the CAW
 * communicates with the IC via function parameters.  If we change things to use a single-pass
 * message loop, then it will be cleaner.
 */
#define reset_last_used_helper()   { last_lb_worker_id = previous_last_lb_worker_id; }

#define time_to_send_eps(eps, last_sent_eps, loops_since_eps_last_sent, send_eps_frequency) \
   (   (eps_ratio(eps, last_sent_eps) >= SEND_EPS_THRESHOLD)                                \
    || ((loops_since_eps_last_sent) > send_eps_frequency))


/* This fn cleans up any extra messages for the controller.  The idea is that the controller
 * should be aware of specific types of messages remaining in its receive buffer, and should
 * clean them out.  It should not clean out other messages (measurement messages, etc...)
 * Also, while the extra update messages should already have been received and used, this will
 * also clean them out, in case any were delayed significantly
 */
static void ic_clean_msgs(MPI_Datatype update_mpi_t, MPI_Comm comm)
{
    MPI_Status status;

    /* Extra worker updates (see note above): */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
    {
	update_t wrkr_update;
	pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, comm, &status);
    }

    /* Extra worker-is-idle messages: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_LB_WRKR_IDLE, comm, &status))
	pi_recv_dummy(status.MPI_SOURCE, TAG_LB_WRKR_IDLE, comm, &status);

    /* Hmmm. What about the other tags.  The ic-as-iw code takes care of SEND_RGN messages.
       I assume there are no extra INIT_UPDATE messages (though a lengthy assertion could
       eliminate that assumption, I think its a pretty safe one).  So that's it! */
} /* ic_clean_msgs() */


static bool all_workers_idle(bool idle[])
{
    register int i;
    for (i = 0; i < nworkers_global; ++i)
    {
	if (!idle[i])
	    return FALSE;
    }
    return TRUE;
} /* all_workers_idle() */


/* This function returns the value of prev_worker, plus one, but: if the value wraps
 * around the last worker, then the id of the first worker is returned, but care is
 * taken not to return the id of the IC if they are not working
 */
static int increment_last_lb_worker(int prev_worker, bool use_controller)
{
    if (++prev_worker >= nworkers_global)
	return (use_controller ? 0 : 1);
    else
	return prev_worker;
} /* increment_last_lb_worker() */


/* Checks for an idle worker in the idle[] array.  If one is found, then its id is returned
 * in helper_id, and TRUE is returned, else FALSE is returned.  The flag use_controller controls
 * whether or not the controller's id is allowed to be sent along as a helper's id
 */
static bool idle_worker(bool idle[], int num_non_ic_workers, bool use_controller, int *helper_id)
{
    bool found_helper = FALSE;
    int wrk_id, i;

    pi_assert(last_lb_worker_id >= 0 && last_lb_worker_id < nworkers_global,
	      "Invalid last_lb_worker_id in idle_worker()");

    *helper_id = -1;

#ifdef EXTRA_RR_TESTING
    /* This and the section below for big-time testing of round-robin: */
    {
	int j;
	printf("[IC]: TESTRR: last_lb_worker: %d; idle[]:", last_lb_worker_id);
	for (j = 0; j < nworkers_global; ++j)
	    printf("%1d", idle[j]);
    }
#endif
    
    /* Control the loop based on the # of workers, but the loop variable we really want 
       is based on last_lb_worker_id.  Aye Carumba!: */
    for (i = 0, wrk_id = increment_last_lb_worker(last_lb_worker_id, use_controller);
	 i < (use_controller ? num_non_ic_workers : nworkers_global);
	 ++i, wrk_id = increment_last_lb_worker(wrk_id, use_controller))
    {
	if (idle[wrk_id])
	{
	    found_helper = TRUE;
	    previous_last_lb_worker_id = last_lb_worker_id;
	    last_lb_worker_id = *helper_id = wrk_id;
	    break;
	}
    }

#ifdef EXTRA_RR_TESTING
    printf("; found? %c; helper: %d\n", B2YN(found_helper), *helper_id);
#endif

    return found_helper;

} /* idle_worker() */


static void send_to_each_worker(void *msg, int count, MPI_Datatype mtype, 
				int tag, int num_non_ic_workers, MPI_Comm comm)
{
    int i;
    /* The first non-ic-worker always has a rank of 1: */
    for (i = 1; i <= num_non_ic_workers; ++i)
	pi_send(msg, count, mtype, i, tag, comm);
} /* send_to_each_worker() */


static void send_global_stop_message(int num_non_ic_workers, MPI_Comm comm)
{
    int i;
    /* The first non-ic-worker always has a rank of 1: */
    for (i = 1; i <= num_non_ic_workers; ++i)
	pi_send_dummy(i, TAG_INTG_STOP, comm);
} /* send_global_stop_message() */


/* Whether an IW, or the CAW, sends back an update, there are commonalities between
 * the actions taken by the IC as a result of that update.  This function performs
 * those common actions: checking for global termination, and handling load balancing.
 * The balance between function length and # of parameters is just barely on the side
 * of creating this function at all ;-)
 */
static void handle_update(int fcn_eval_limit, int num_non_ic_workers, pi_base_t eps_a, 
			  pi_base_t eps_r, pi_base_t result[], pi_base_t estabs[], 
			  int nfcns, update_t *wrkr_update_p, int last_source, int fcncount,
			  MPI_Comm comm, bool *globally_done, bool *fcn_limit_flag, pi_base_t *eps,
			  pi_base_t *last_sent_eps, int *loops_since_eps_last_sent, 
			  bool idle[], bool handle_idle_workers, int send_eps_frequency, 
			  bool ic_is_worker)
{
    idle[last_source] = wrkr_update_p->l_done;
    eps_sent_since_last_update = 0;
    pl_log_recv_upd(pl_upd_reg, wrkr_update_p, last_source, fcncount, 
		    calc_new_eps(eps_a, eps_r, result, nfcns), *last_sent_eps, 
		    result, estabs, idle);
    
    if (fcncount > fcn_eval_limit)
    {
	*globally_done = TRUE;
	*fcn_limit_flag = TRUE;
	D(printf("[IC]: Sending INTG_STOP to all workers (fcncount stop)\n"));
	send_global_stop_message(num_non_ic_workers, comm);
	pl_log_ic_stop(pl_stop_fcn);
    }
    else
    {
	*eps = calc_new_eps(eps_a, eps_r, result, nfcns);
	if (result_close_enough(estabs, *eps, nfcns))
	{
	    *globally_done = TRUE;
	    D(printf("[IC]: Sending INTG_STOP to all workers (eps stop)\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	    pl_log_ic_stop(pl_stop_eps);
	}
	else if (all_workers_idle(idle))
	{
	    /* (This test must be 3rd, else we will always terminate with this condition 
	       when nworkers == 1) */

	    *globally_done = TRUE;

	    /* If there is a single worker, and they happen to report done before the IC,
	       then we will 'believe' them, and not inform the user that anything unusual
	       happened: */
	    if (ic_is_worker ? num_non_ic_workers > 0 : num_non_ic_workers > 1)
		*fcn_limit_flag = TRUE;
	    /* I so rarely use all caps that you know this must be important! ;-): */ 
	    D(printf("[IC]: ALL WORKERS REPORT IDLE.  ROUND-OFF ERROR GLOBAL STOP #1!\n"));
	    D(printf("[IC]: Sending INTG_STOP to all workers\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	    pl_log_ic_stop(pl_stop_all_idle);
	}
	else 
	{
	    if (time_to_send_eps(*eps, *last_sent_eps, *loops_since_eps_last_sent, 
				 send_eps_frequency))
	    {
		D(printf("[IC]: Sending NEW_EPS to all workers (a); eps is %g\n", (double) *eps));
		send_to_each_worker(eps, 1, pi_base_mpi_t, TAG_NEW_EPS, num_non_ic_workers, comm);
		pl_log_send_eps(pl_eps_reg, *eps, *last_sent_eps);
		*last_sent_eps = *eps;
		*loops_since_eps_last_sent = 0;
	    }

	    if (handle_idle_workers)
	    {
		int helper_id;
		if (!wrkr_update_p->l_done
		    && idle_worker(idle, num_non_ic_workers, ic_is_worker, &helper_id))
		{
		    pi_send(&helper_id, 1, MPI_INT, last_source, TAG_LB_HELPER_ID, comm);
		    idle[helper_id] = FALSE;
		    pl_log_send_helper(last_source, helper_id, idle);
		    D(printf("[IC]: Sent HELPER_ID of %d to worker %d\n", helper_id, last_source));
		}
	    }
	}
    }
#ifdef SPECIAL_GRAPH
    printf("%10d %25.8f\n", fcncount / global_nevals,
	   (double) (infNorm(estabs, nfcns) / calc_new_eps(eps_a, eps_r, result, nfcns)));
#endif
    D({
	int j;
	printf("[IC]: Update from %d: ldone %c; fcncnt %d; (",
	       last_source, B2YN(wrkr_update_p->l_done), fcncount);
	for (j = 0; j < nworkers_global; ++j)
	    printf("%1d", idle[j]);
	printf("); g_err ratio %.8f\n",
	       (double) err_ratio(infNorm(estabs, nfcns), 
				  calc_new_eps(eps_a, eps_r, result, nfcns)));
    });
} /* handle_update() */


/* This fn contains all the IC functionality.
 * num_workers will include the IC, if the IC is a worker.  E.g., if the IC is a worker, then
 * num_workers will be 1, 2, 4, 8, ... (with worker ranks of 0, 0-1, 0-3, ...); if the IC is not a 
 * worker, then the num_workers will be 1, 3, 7, ...., (with worker ranks of 1, 1-3, 1-7, ...)
 * (Though now the nworkers can be any number, not just a power of 2)
 */
extern void integ_controller(pi_ifcn_t ifcn_ptr, int ndims, int nfcns, int irule, 
			     int fcn_eval_limit, pi_base_t eps_a, pi_base_t eps_r,
			     pi_base_t a[], pi_base_t b[], pi_base_t result[], pi_base_t estabs[], 
			     int id, int nworkers, MPI_Comm comm, pi_rparm_t *rparms_p, 
			     bool *fcn_limit_flag, int *fcn_eval_count, double *total_time,
			     char fcn_name[])
{
    bool globally_done = FALSE;

    int fcncount = 0;                   /* number of function evaluations      */
    int num_non_ic_workers = (rparms_p->ic_is_worker ? nworkers - 1 : nworkers);
    
    /* These control when eps updates are sent down: */
    int send_eps_frequency = max(num_non_ic_workers, BASE_SEND_EPS_FREQUENCY);
    int loops_since_eps_last_sent = 0;

    pi_base_t eps, last_sent_eps;

    /* Each ith entry is TRUE if the ith worker is idle.  This array is always used, as it is 
       used for lb, and to make sure that all workers are not idle: */
    bool idle[PI_MAX_WORKERS];

    MPI_Datatype update_mpi_t;
    MPI_Status status;
    update_t wrkr_update;
    double start_time;
    int i;

    /* Transfer run options from rparms to local variables: */
    bool do_lb           = rparms_p->do_lb;
    bool ic_is_worker    = rparms_p->ic_is_worker;
    int vis_type         = rparms_p->vis_type;/*!jli visualization */
    int portno		 =  rparms_p->vis_port; /* port no to listen to */
    int new_sockfd;             /* the socket used to send/recv data */
    /* end jli! */

#ifdef SPECIAL_GRAPH
    global_nevals = neval(ndims, irule);
#endif

    D(printf("[IC]: Beginning integ_controller(); do_lb: %d; ic_is_worker: %d\n", 
	     do_lb, ic_is_worker));

    pi_assert(nworkers <= PI_MAX_WORKERS, "Too many processes for ParInt!");
    last_lb_worker_id = 0;
    nworkers_global = nworkers;
    eps_sent_since_last_update = 0;
    
    /* Default values for returned parameters that are not always set: */
    *fcn_limit_flag = FALSE;

    /* This inits all workers, even the IC, to be initially non-idle.  This is especially
       important if the IC is not an IW, since in that case we never want the IC id to be sent
       out as a helper id: */
    for (i = 0; i < PI_MAX_WORKERS + 1; ++i)
	idle[i] = FALSE;

    /*jli!!  initialize the socket connection */
    D(printf("vis_type: %d\n",vis_type));
    D(printf("portno:%d\n",portno));
    if (vis_type == VIS_ON)
	vis_init_connect(portno, num_non_ic_workers, &new_sockfd, comm);

    create_mpi_type_update(&update_mpi_t);
    clear_results(result, estabs, nfcns);
    if (ic_is_worker)
	caw_init_first(comm, id);     /* (this can be called more than once w/ no problems) */
    pl_ic_init(fcn_name, ndims, nfcns, irule, fcn_eval_limit, eps_a, eps_r, a, b, rparms_p, 
	       nworkers + (ic_is_worker ? 0 : 1), id);
    ms_start_clock();
    start_time = MPI_Wtime();         /* (maintain a single clock outside of ms stuff) */
    
    /* Note that we do not have to handle idle[] for the initial round of work, 
       since the IW's (including the caw) cannot report themselves idle at this point,
       since they do not yet have an eps. If they do report idle after they get this initial
       eps, they do so with a WRKR_IDLE message; at the receipt of these messages, I do 
       check l_done and set idle[] properly. */

    /* Allow the controller to do its first bit of work and update results, 
       estabs, and the fcncount: */
    if (ic_is_worker)
    {
	update_t *ctrl_update_p;
	caw_init_intg(ndims, a, b, ifcn_ptr, nfcns, irule, nworkers, rparms_p);
	caw_first_pass(&ctrl_update_p);
	if (vis_type == VIS_ON) /* !!jli*/
	    /* send error estimates and # of regions to Applet */	
	    vis_send_update(new_sockfd,0,*ctrl_update_p);
	update_results(ctrl_update_p, nfcns, result, estabs, &fcncount);
	pl_log_recv_upd(pl_upd_init, ctrl_update_p, CAW_ID, fcncount, 0.0, 0.0, NULL, NULL, NULL);
    }
       
    /* Gather the full, initial values of result and estabs from each worker.  Note that this
       first step must be separate from the main loop, for we must make sure that we receive
       at least a single update from each worker: */
    for (i = 0; i < num_non_ic_workers; ++i)
    {
	pi_recv(&wrkr_update, 1, update_mpi_t, MPI_ANY_SOURCE, TAG_INIT_UPDATE, comm, &status);
	if (vis_type == VIS_ON) /* !!jli*/
	    /* send error estimates and # of regions to Applet */	
	    vis_send_update(new_sockfd, status.MPI_SOURCE, wrkr_update);
	update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
	pl_log_recv_upd(pl_upd_init, &wrkr_update, status.MPI_SOURCE, fcncount, 0.0, 
			0.0, NULL, NULL, NULL);
    }
    D(printf("[IC]: Received INIT_UPDATE from all workers\n"));
    
    /* Check to see if we are globally done.  The quick fcncount stop is here on purpose: it
       may be useful in fudging a non-adaptive monte carlo run that stops before any adapting
       can be done: */
    if (fcncount >= fcn_eval_limit)
    {
	globally_done = TRUE;
	*fcn_limit_flag = TRUE;
	D(printf("[IC]: Sending INTG_STOP to all workers (quick fcn count stop)\n"));
	send_global_stop_message(num_non_ic_workers, comm);
	pl_log_ic_stop(pl_stop_quick_fcn);
    }
    else
    {
	eps = calc_new_eps(eps_a, eps_r, result, nfcns);
	if (result_close_enough(estabs, eps, nfcns))
	{
	    globally_done = TRUE;
	    D(printf("[IC]: Sending INTG_STOP to all workers (quick eps stop)\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	    pl_log_ic_stop(pl_stop_quick_eps);
	}	
	else
	{          
	    D(printf("[IC]: Sending INIT_EPS to workers. The eps is %g\n", (double) eps));
	    send_to_each_worker(&eps, 1, pi_base_mpi_t, TAG_INIT_EPS, num_non_ic_workers, comm);
	    pl_log_send_eps(pl_eps_init, eps, 0.0);
	    last_sent_eps = eps;
	}
    }

    D(printf("[IC]: Beginning regular while (!done) loop\n"));
    
    /* Begin receiving regular updates until fcncount too big or estabs is small enough: */
    while (!globally_done )
    {
	++loops_since_eps_last_sent;

	if (vis_type == VIS_ON) {/* !! jli */
	    /* read a request from the applet */	
	    D(printf("[IC_VIS]want to recv a request from applet\n"));
	    vis_read_request(new_sockfd,num_non_ic_workers,ic_is_worker,comm);
	}
				
	pi_assert(!all_workers_idle(idle), "All workers idle where they shouldn`t be");

	/* Check for idle status correction: */
	if (pi_iprobe(MPI_ANY_SOURCE, TAG_LB_WRKR_IDLE, comm, &status))
	{
	    pi_recv_dummy(status.MPI_SOURCE, TAG_LB_WRKR_IDLE, comm, &status);
	    idle[status.MPI_SOURCE] = TRUE;
	    pl_log_recv_idle(status.MPI_SOURCE, idle);
	    D(printf("[IC]: Received msg WRKR_IDLE from worker %d\n", status.MPI_SOURCE));
	    if (all_workers_idle(idle))
	    {
		globally_done = TRUE;
		/* See comment in handle_update() concerning this check: */
		if (nworkers > 1)
		    *fcn_limit_flag = TRUE;
		D(printf("[IC]: ALL WORKERS REPORT IDLE. ROUND-OFF ERROR GLOBAL STOP #2!\n"));
		D(printf("[IC]: Sending INTG_STOP to all workers\n"));
		send_global_stop_message(num_non_ic_workers, comm);
		pl_log_ic_stop(pl_stop_all_idle);
	    }
	}

	if (ic_is_worker)
	{
	    bool caw_was_locally_done = idle[IC_ID];
	    update_t *ctrl_update_p;
	    bool use_update;

	    D(printf("[IC]: Calling myself to do work, currently idle? %d\n", idle[id]));
	    
	    if (do_lb)
	    {
		bool used_helper;
		int helper_id;
		/* The FALSE hard-codes the fact that I (the IC) should not be chosen
		   as a helper!: */
		caw_one_pass(eps, 
			     (idle_worker(idle, num_non_ic_workers, FALSE, &helper_id)
			      ? helper_id
			      : -1), 
			     &use_update, &ctrl_update_p, &idle[id], &used_helper);

		if (helper_id != -1)
		{
		    if (used_helper)
		    {
			idle[helper_id] = FALSE;
			D(printf("[IC]: CAW used helper id of %d\n", helper_id));
			/* Only if the caw uses a helper do I log the "sent-helper-id" event: */
			pl_log_send_helper(CAW_ID, helper_id, idle);
		    }
		    else
		    {
			reset_last_used_helper();
		    }
		}
	    }
	    else
	    {
		caw_one_pass(eps, -1, &use_update, &ctrl_update_p, &idle[id], NULL);
	    }

	    if (use_update)
	    {
		/* We have an update to use. This code here is the same as when we get an update
		   from a worker, except when the source (the controller) is not idle, we do not
		   try to send a message 3, rather we wait till its next bit of work and just 
		   pass in the id of an idle worker (if there is one): */
		if (vis_type == VIS_ON) /* !!jli*/
		    /* send error estimates and # of regions to Applet */
		    vis_send_update(new_sockfd,0,*ctrl_update_p);

		update_results(ctrl_update_p, nfcns, result, estabs, &fcncount);

#if 0
		printf("estabs check: IC says %20G; CAW says %20G\n", 
		       estabs[0], ctrl_update_p->l_estabs);
#endif

		/* This if{} add in R05/06/99 when fixing the 03/15/99 bug fix in nextrel code; I'm
		   not sure it is necessary since num_working_workers removed from this code,
		   but better safe than sorry.  When we use the new algorithm, all will become
		   clear, like a freshly washed pane of glass on a sunny spring day: */
		/* Since the caw always gets a new global_eps, and checks it even if it is locally
		   done, it is possible for a locally idle caw to become un-idle, even without
		   any load balancing.  This checks for this possibility.  Note that the idle
		   array is already properly updated: */
		if (caw_was_locally_done && !ctrl_update_p->l_done)
		{
		    /* BUG FIX (M03/15/99 -- rrz): A change in caw code for a bug fix necessitated
		       this change here, to ensure that idle[0] is updated properly: */
		    idle[0] = FALSE;
		    D(printf("[IC]: My work result: work done; newly non-idle\n"));
		}

		/* This handles updating other local status variables and sending out
		   done messages and new global eps.  The FALSE parameter tells the fn
		   to *not* deal with any load balancing (as mentioned above): */
		handle_update(fcn_eval_limit, num_non_ic_workers, eps_a, eps_r, result, estabs, 
			      nfcns, ctrl_update_p, id, fcncount, comm, &globally_done, 
			      fcn_limit_flag, &eps, &last_sent_eps, &loops_since_eps_last_sent,
			      idle, FALSE, send_eps_frequency, ic_is_worker);
	    }
	    else
	    {
		/* If we can not use the update, then it means we are locally idle.  Since I
		   always send in a new global_eps, it is possible for the caw to declare them-
		   selves idle without doing a last round of work.  Therefore, this checks
		   to see if this occurs, and if so, updates idle[] appropriately and checks
		   to see if all workers are now done: */
		if (idle[IC_ID] && !caw_was_locally_done)
		{
		    D(printf("[IC]: My work result: no work done; newly idle\n"));
		    if (all_workers_idle(idle))
		    {
			globally_done = TRUE;
			/* See comment in handle_update() concerning this check: */
			if (nworkers > 1)
			    *fcn_limit_flag = TRUE;
			D(printf("[IC]: ALL WORKERS REPORT IDLE. ROUND-OFF "
				 "ERROR GLOBAL STOP #3!\n"));
			D(printf("[IC]: Sending INTG_STOP to all workers\n"));
			send_global_stop_message(num_non_ic_workers, comm);
			pl_log_ic_stop(pl_stop_all_idle);
		    }
		}
	    }
	}
	
	/* This reads results until there are no more.  I think that there are better
	   techniques for getting worker updates, but those ideas will wait for another day: */
	while (!globally_done && pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
	{
	    ++loops_since_eps_last_sent;
	    pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, 
		    comm, &status);
	    if (vis_type == VIS_ON) /* !!jli*/
		/* send error estimates and # of regions to Applet */	
		vis_send_update(new_sockfd,status.MPI_SOURCE,wrkr_update);
	    update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
	    handle_update(fcn_eval_limit, num_non_ic_workers, eps_a, eps_r, result, estabs, 
			  nfcns, &wrkr_update, status.MPI_SOURCE, fcncount, comm, &globally_done,
			  fcn_limit_flag, &eps, &last_sent_eps, &loops_since_eps_last_sent, idle,
			  do_lb, send_eps_frequency, ic_is_worker);
	}

	if (vis_type == VIS_ON) /* !! jli */
	{
	    /* Receive histogram information,then send to the Applet */
	    D(printf("[IC_VIS]:try to recv a hist from worker\n"));
	    vis_recv_hist(new_sockfd, comm);
	}
	/* I need to also have this here.  Timing problems can result in the eps never being
	   sent otw.  Things would work better (IMHO) if the loop was more structured: */
#if 0
	if (!globally_done
	    && time_to_send_eps(eps, last_sent_eps, loops_since_eps_last_sent,
				send_eps_frequency)
	    && (eps_sent_since_last_update >= MAX_EPS_BETWEEN_UPDATES))
	{
	    printf("@");
	}
#endif

	if (!globally_done
	    && time_to_send_eps(eps, last_sent_eps, loops_since_eps_last_sent, send_eps_frequency)
	    && (eps_sent_since_last_update < MAX_EPS_BETWEEN_UPDATES))
	{
	    D(printf("[IC]: Sending NEW_EPS to all workers (b); eps is %g\n", (double) eps));
	    send_to_each_worker(&eps, 1, pi_base_mpi_t, TAG_NEW_EPS, num_non_ic_workers, comm);
	    pl_log_send_eps(pl_eps_reg, eps, last_sent_eps);
	    last_sent_eps = eps;
	    loops_since_eps_last_sent = 0;
	    ++eps_sent_since_last_update;
	}
    }

    /* There may be some final update messages out there. Since their comm time has
       already been paid, I might as well collect them and add them onto the result, in order
       to increase our  accuracy slightly: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
    {
	pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, comm, &status);
	update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
	D(printf("[IC]: Received extra WRKR_UPDATE from worker %d (new fcncount %d)\n",
		 status.MPI_SOURCE, fcncount));
	pl_log_recv_upd(pl_upd_braking, &wrkr_update, status.MPI_SOURCE, fcncount, 0.0, 
			0.0, result, estabs, NULL);
    }

    /* Stop the clock, then do some clean-up things: */
    *total_time = MPI_Wtime() - start_time;
    ms_stop_clock();
    *fcn_eval_count = fcncount;

    /* jli !! visulization */
    if (vis_type == VIS_ON) {
	/*close the sockets */
	vis_close(new_sockfd);
    }

    if (ic_is_worker)
	caw_finalize_intg();
    pl_finalize();
    ic_clean_msgs(update_mpi_t, comm);
    MPI_Type_free(&update_mpi_t);
    D(printf("[IC]: Leaving integ_controller()\n"));

} /* integ_controller() */

