/*        File: intg-ctrl.c
     Copyright: (C) 1998, 1999 by Elise de Doncker, Ajay Gupta, Alan Genz and Rodger Zanny. 
                       All rights reserved. See Copyright.txt in top-level directory.
  ParInt Versn: MPI; local receiver-initiated load-balancing, or, no load-balancing
   Description: Source to implement the integral controller (which can also act as an IW)
       Version: 01/13/98  rrz  Original (from existing code)
		05/12/98  rrz  Code freeze for ParInt1.0
		06/28/98  rrz  Modified call to caw_init_intg() for different
				proto (for visualization)
		7/3/98	  jli  added code for visulization
		09/17/98  rrz  Remove unused ms_type variable
		11/15/98  rrz  Add round-robin selection of next worker to
				receive work in LB (as in extrap version)
		11/19/98  rrz  Add reject_helper parameter to caw_one_pass() to
				fix bug
*/
/*
#define DBG_THIS_FILE
*/

#include <stdio.h>
#include <mpi.h>

#include <parint.h>
#include <math-adapt.h>
#include <pi-defs.h>
#include <messaging.h>
#include <pi-vspec.h>
#include <measures.h>
#include <heap.h>
#include <visual.h> /* jli!! */

/* For a short while, these will be compile-time constants.  When we have cemented their
 * behavior, I will elevate them to be run-time parameters: */

/* eps will be sent down to workers when the ratio between it and the last sent
 * eps is greater than or equal to the THRESHOLD value, or, if there have been
 * FREQUENCY iterations since the last sent eps: */
#define SEND_EPS_THRESHOLD         0.01
#define BASE_SEND_EPS_FREQUENCY       5

/* A global, which breaks tradiation, but oh well.  This contains the id of the last worker
 * to receive a region via load balancing.  Selection of the next worker will begin with
 * the worker after this one: */
static int last_lb_worker_id = 0;
static int previous_last_lb_worker_id = 0;
static int nworkers_global = 0;

/* This design is a little ugly, but is based on the fact that I have to send the caw 
 * helper everytime, b/c I (the IC) do not know if the caw is idle or not, b/c it
 * checks its local estabs against the eps on *every* call (since it is able to).
 * But, I do not want to pass by an idle worker in the round-robin location if I
 * do not have to.  The design of how this is done is not the cleanest, and is due to 
 * the fact that all other workers communicate with the IC via messages, while the CAW
 * communicates with the IC via function parameters.  If we change things to use a single-pass
 * message loop, then it will be cleaner.
 */
#define reset_last_used_helper()   { last_lb_worker_id = previous_last_lb_worker_id; }

#define time_to_send_eps(eps, last_sent_eps, loops_since_eps_last_sent, send_eps_frequency) \
   (   (absol((eps) - (last_sent_eps)) / (last_sent_eps)) >= SEND_EPS_THRESHOLD             \
    || (loops_since_eps_last_sent > send_eps_frequency))


/* This fn cleans up any extra messages for the controller.  The idea is that the controller
 * should be aware of specific types of messages remaining in its receive buffer, and should
 * clean them out.  It should not clean out other messages (measurement messages, etc...)
 * Also, while the extra update messages should already have been received and used, this will
 * also clean them out, in case any were delayed significantly
 */
static void ic_clean_msgs(MPI_Datatype update_mpi_t, MPI_Comm comm)
{
    MPI_Status status;

    /* Extra worker updates (see note above): */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
    {
	update_t wrkr_update;
	pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, comm, &status);
    }

    /* Extra worker-is-idle messages: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_LB_WRKR_IDLE, comm, &status))
	pi_recv_dummy(status.MPI_SOURCE, TAG_LB_WRKR_IDLE, comm, &status);

    /* Hmmm. What about the other tags.  The ic-as-iw code takes care of SEND_RGN messages.
       I assume there are no extra INIT_UPDATE messages (though a lengthy assertion could
       eliminate that assumption, I think its a pretty safe one).  So that's it! */
} /* ic_clean_msgs() */


/* This function returns the value of prev_worker, plus one, but: if the value wraps
 * around the last worker, then the id of the first worker is returned, but care is
 * taken not to return the id of the IC if they are not working
 */
static int increment_last_lb_worker(int prev_worker, bool use_controller)
{
    if (++prev_worker >= nworkers_global)
	return (use_controller ? 0 : 1);
    else
	return prev_worker;
} /* increment_last_lb_worker() */


/* Checks for an idle worker in the idle[] array.  If one is found, then its id is returned
 * in helper_id, and TRUE is returned, else FALSE is returned.  The flag use_controller controls
 * whether or not the controller's id is allowed to be sent along as a helper's id
 */
static bool idle_worker(bool idle[], int num_non_ic_workers, bool use_controller, int *helper_id)
{
    bool found_helper = FALSE;
    int wrk_id, i;

    pi_assert(last_lb_worker_id >= 0 && last_lb_worker_id < nworkers_global,
	      "Invalid last_lb_worker_id in idle_worker()");

    *helper_id = -1;

#if 1
    /* This and the section below for big-time testing of round-robin: */
    {
	int j;
	printf("[IC]: TESTRR: last_lb_worker: %d; idle[]:", last_lb_worker_id);
	for (j = 0; j < nworkers_global; ++j)
	    printf("%1d", idle[j]);
    }
#endif
    
    /* Control the loop based on the # of workers, but the loop variable we really want 
       is based on last_lb_worker_id.  Aye Carumba!: */
    for (i = 0, wrk_id = increment_last_lb_worker(last_lb_worker_id, use_controller);
	 i < (use_controller ? num_non_ic_workers : nworkers_global);
	 ++i, wrk_id = increment_last_lb_worker(wrk_id, use_controller))
    {
	if (idle[wrk_id])
	{
	    found_helper = TRUE;
	    previous_last_lb_worker_id = last_lb_worker_id;
	    last_lb_worker_id = *helper_id = wrk_id;
	    break;
	}
    }

#if 1
    printf("; found? %c; helper: %d\n", B2YN(found_helper), *helper_id);
#endif

    return found_helper;

} /* idle_worker() */


static void send_to_each_worker(void *msg, int count, MPI_Datatype mtype, 
				int tag, int num_non_ic_workers, MPI_Comm comm)
{
    int i;
    /* The first non-ic-worker always has a rank of 1: */
    for (i = 1; i <= num_non_ic_workers; ++i)
	pi_send(msg, count, mtype, i, tag, comm);
} /* send_to_each_worker() */


static void send_global_stop_message(int num_non_ic_workers, MPI_Comm comm)
{
    int i;
    /* The first non-ic-worker always has a rank of 1: */
    for (i = 1; i <= num_non_ic_workers; ++i)
	pi_send_dummy(i, TAG_INTG_STOP, comm);
} /* send_global_stop_message() */


/* Whether an IW, or the CAW, sends back an update, there are commonalities between
 * the actions taken by the IC as a result of that update.  This function performs
 * those common actions: checking for global termination, and handling load balancing.
 * The balance between function length and # of parameters is just barely on the side
 * of creating this function at all ;-)
 */
static void handle_update(int fcn_eval_limit, int num_non_ic_workers, pi_base_t eps_a, 
			  pi_base_t eps_r, pi_base_t result[], pi_base_t estabs[], 
			  int nfcns, update_t *wrkr_update_p, int last_source, int fcncount,
			  MPI_Comm comm, bool *globally_done, bool *fcn_limit_flag, pi_base_t *eps,
			  pi_base_t *last_sent_eps, int *loops_since_eps_last_sent, 
			  bool idle[], bool handle_idle_workers, int send_eps_frequency, 
			  bool ic_is_worker, int *num_working_workers)
{
    /* Regardless of whether we are load-balancing, decrement the number of
       working workers if this worker reported idle: */
    if (last_source != INTG_CTRL_ID && wrkr_update_p->l_done)
	--(*num_working_workers);
    D(if (last_source && wrkr_update_p->l_done) 
      printf("[IC]: Decrementing nww to %d due to l_done by IW %d\n",
	     *num_working_workers, last_source));

    if (fcncount > fcn_eval_limit)
    {
	*globally_done = TRUE;
	*fcn_limit_flag = TRUE;
	D(printf("[IC]: Sending INTG_STOP to all workers (fcncount stop)\n"));
	send_global_stop_message(num_non_ic_workers, comm);
    }
    else
    {
	*eps = calc_new_eps(eps_a, eps_r, result, nfcns);
	if (result_close_enough(estabs, *eps, nfcns))
	{
	    *globally_done = TRUE;
	    D(printf("[IC]: Sending INTG_STOP to all workers (eps stop)\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	}
	else if (*num_working_workers <= 0)
	{
	    /* (This test must be 3rd, else we will always terminate with this condition 
	       when nworkers == 1) */

	    *globally_done = TRUE;

	    /* If there is a single worker, and they happen to report done before the IC,
	       then we will 'believe' them, and not inform the user that anything unusual
	       happened: */
	    if (ic_is_worker ? num_non_ic_workers > 0 : num_non_ic_workers > 1)
		*fcn_limit_flag = TRUE;
	    /* I so rarely use all caps that you know this must be important! ;-): */ 
	    D(printf("[IC]: ALL WORKERS REPORT IDLE.  ROUND-OFF ERROR GLOBAL STOP!\n"));
	    D(printf("[IC]: Sending INTG_STOP to all workers\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	}
	else 
	{
	    if (time_to_send_eps(*eps, *last_sent_eps, *loops_since_eps_last_sent, 
				 send_eps_frequency))
	    {
		D(printf("[IC]: Sending NEW_EPS to all workers (a); eps is %g\n", (double) *eps));
		send_to_each_worker(eps, 1, pi_base_mpi_t, TAG_NEW_EPS, num_non_ic_workers, comm);
		*last_sent_eps = *eps;
		*loops_since_eps_last_sent = 0;
	    }

	    if (handle_idle_workers)
	    {
		int helper_id;
		idle[last_source] = wrkr_update_p->l_done;
		if (   !wrkr_update_p->l_done
		       && idle_worker(idle, num_non_ic_workers, ic_is_worker, &helper_id))
		{
		    pi_send(&helper_id, 1, MPI_INT, last_source, TAG_LB_HELPER_ID, comm);
		    idle[helper_id] = FALSE;
		    ++(*num_working_workers);
		    D(printf("[IC]: Sent HELPER_ID of %d to worker %d (nww now %d)\n", 
			     helper_id, last_source, *num_working_workers));
		}
	    }
	}
    }

#if 1
    D(printf("[IC]: Update from %d: ldone %c; fcncnt %d; nww %d; "
	     "g_eps ratio %.8f\n", last_source, B2YN(wrkr_update_p->l_done), 
	     fcncount, *num_working_workers,
	     (double) (infNorm(estabs, nfcns) / calc_new_eps(eps_a, eps_r, result, nfcns))));
#else
    {
	int j;
	printf("[IC]: Update from %d: ldone %c; fcncnt %d; nww %d (",
	       last_source, B2YN(wrkr_update_p->l_done), fcncount, *num_working_workers);
	for (j = 0; j < nworkers_global; ++j)
	    printf("%1d", idle[j]);
	printf("); g_eps ratio %.8f\n",
	       (double) (infNorm(estabs, nfcns) / calc_new_eps(eps_a, eps_r, result, nfcns)));
    }
#endif

} /* handle_update() */


/* This fn contains all the IC functionality.
 * num_workers will include the IC, if the IC is a worker.  E.g., if the IC is a worker, then
 * num_workers will be 1, 2, 4, 8, ... (with worker ranks of 0, 0-1, 0-3, ...); if the IC is not a 
 * worker, then the num_workers will be 1, 3, 7, ...., (with worker ranks of 1, 1-3, 1-7, ...)
 */
extern void integ_controller(pi_ifcn_t ifcn_ptr, int ndims, int nfcns, int irule, 
			     int fcn_eval_limit, pi_base_t eps_a, pi_base_t eps_r,
			     pi_base_t a[], pi_base_t b[], pi_base_t result[], pi_base_t estabs[], 
			     int id, int nworkers, MPI_Comm comm, pi_rparm_t *rparms_p, 
			     bool *fcn_limit_flag, int *fcn_eval_count, double *total_time)
{
    bool globally_done = FALSE;

    int fcncount = 0;                   /* number of function evaluations      */
    int num_non_ic_workers = (rparms_p->ic_is_worker ? nworkers - 1 : nworkers);
    
    /* These control when eps updates are sent down: */
    int send_eps_frequency = max(num_non_ic_workers, BASE_SEND_EPS_FREQUENCY);
    int loops_since_eps_last_sent = 0;

    pi_base_t eps, last_sent_eps;

    /* This is used in the rare instance when all workers think that they are done, but the
       IC thinks we are not globally_done.  In that instance, we terminate: */
    int num_working_workers = nworkers;
    
    /* Each ith entry is TRUE if the ith worker is idle.  This array is only used if
       do_lb is TRUE: */
    bool idle[PI_MAX_WORKERS];

    MPI_Datatype update_mpi_t;
    MPI_Status status;
    update_t wrkr_update;
    double start_time;
    int i;

    /* Transfer run options from rparms to local variables: */
    bool do_lb           = rparms_p->do_lb;
    bool ic_is_worker    = rparms_p->ic_is_worker;
    int vis_type         = rparms_p->vis_type;/*!jli visualization */
    /* int ms_type          = rparms_p->ms_type;*/ /* measure type */
    char ms_file[PI_MAX_PATH +1];
    int portno		 =  rparms_p->vis_port; /* port no to listen to */
    int new_sockfd;             /* the socket used to send/recv data */

    strcpy(ms_file, rparms_p->ms_file);/* measure output filename */     
    /* end jli! */

    D(printf("[IC]: Beginning integ_controller(); do_lb: %d; ic_is_worker: %d\n", 
	     do_lb, ic_is_worker));

    pi_assert(nworkers <= PI_MAX_WORKERS, "Too many processes for ParInt!");
    last_lb_worker_id = 0;
    nworkers_global = nworkers;

    /* Default values for returned parameters that are not always set: */
    *fcn_limit_flag = FALSE;

    if (do_lb)
    {
	/* This inits all workers, even the IC, to be initially non-idle.  This is especially
	   important if the IC is not an IW, since in that case we never want the IC id to be sent
	   out as a helper id: */
	for (i = 0; i < PI_MAX_WORKERS + 1; ++i)
	    idle[i] = FALSE;
    }

    /*jli!!  initialize the socket connection */
    D(printf("vis_type: %d\n",vis_type));
    D(printf("portno:%d\n",portno));
    if (vis_type == VIS_ON)
	vis_init_connect(portno,num_non_ic_workers,&new_sockfd,comm);

    create_mpi_type_update(&update_mpi_t);
    clear_results(result, estabs, nfcns);
    if (ic_is_worker)
	caw_init_first(comm, id);     /* (this can be called more than once w/ no problems) */
    ms_start_clock();
    start_time = MPI_Wtime();         /* (maintain a single clock outside of ms stuff) */
    
    /* Note that we do not have to handle num_working_workers for the initial round of work, 
       since the IW's (including the caw) cannot report themselves idle at this point,
       since they do not yet have an eps. If they do report idle after they get this initial
       eps, they do so with a WRKR_IDLE message; at the receipt of these messages, I do handle
       num_working_workers properly. */

    /* Allow the controller to do its first bit of work and update results, 
       estabs, and the fcncount: */
    if (ic_is_worker)
    {
	update_t *ctrl_update_p;
	caw_init_intg(ndims, a, b, ifcn_ptr, nfcns, irule, nworkers, rparms_p);
	caw_first_pass(&ctrl_update_p);
	if (vis_type == VIS_ON) /* !!jli*/
	    /* send error estimates and # of regions to Applet */	
	    vis_send_update(new_sockfd,0,*ctrl_update_p);
	update_results(ctrl_update_p, nfcns, result, estabs, &fcncount);
    }
       
    /* Gather the full, initial values of result and estabs from each worker.  Note that this
       first step must be separate from the main loop, for we must make sure that we receive
       at least a single update from each worker: */
    for (i = 0; i < num_non_ic_workers; ++i)
    {
	pi_recv(&wrkr_update, 1, update_mpi_t, MPI_ANY_SOURCE, TAG_INIT_UPDATE, comm, &status);
	if (vis_type == VIS_ON) /* !!jli*/
	    /* send error estimates and # of regions to Applet */	
	    vis_send_update(new_sockfd,i,wrkr_update);
	update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
    }
    D(printf("[IC]: Received INIT_UPDATE from all workers\n"));
    
    /* Check to see if we are globally done.  The quick fcncount stop is here on purpose: it
       may be useful in fudging a non-adaptive monte carlo run that stops before any adapting
       can be done: */
    if (fcncount >= fcn_eval_limit)
    {
	globally_done = TRUE;
	*fcn_limit_flag = TRUE;
	D(printf("[IC]: Sending INTG_STOP to all workers (quick fcn count stop)\n"));
	send_global_stop_message(num_non_ic_workers, comm);
    }
    else
    {
	eps = calc_new_eps(eps_a, eps_r, result, nfcns);
	if (result_close_enough(estabs, eps, nfcns))
	{
	    globally_done = TRUE;
	    D(printf("[IC]: Sending INTG_STOP to all workers (quick eps stop)\n"));
	    send_global_stop_message(num_non_ic_workers, comm);
	}	
	else
	{          
	    D(printf("[IC]: Sending INIT_EPS to workers. The eps is %g\n", (double) eps));
	    send_to_each_worker(&eps, 1, pi_base_mpi_t, TAG_INIT_EPS, num_non_ic_workers, comm);
	    last_sent_eps = eps;
	}
    }

    D(printf("[IC]: Beginning regular while (!done) loop\n"));
    
    /* Begin receiving regular updates until fcncount too big or estabs is small enough: */
    while (!globally_done )
    {
	++loops_since_eps_last_sent;

	if (vis_type == VIS_ON) {/* !! jli */
	    /* read a request from the applet */	
	    D(printf("[IC_VIS]want to recv a request from applet\n"));
	    vis_read_request(new_sockfd,num_non_ic_workers,ic_is_worker,comm);
	}
				
#if 1
	{
	    int count_nww = 0;
	    int j; 
	    for (j = 0; j < nworkers; ++j)
		if (!idle[j])
		    ++count_nww;
	    if (count_nww != num_working_workers)
	    {
		printf("[IC]: PROBLEM1: count_nww %d; num_working_workers: %d; idle[]:", 
		       count_nww, num_working_workers);
		for (j = 0; j < nworkers_global; ++j)
		    printf("%1d", idle[j]);
		printf("\n");
	    }
	}
#endif
	pi_assert(num_working_workers >= 1 && num_working_workers <= nworkers, 
		  "Invalid num_working_workers value!\n");

	/* Check for idle status correction: */
	if (pi_iprobe(MPI_ANY_SOURCE, TAG_LB_WRKR_IDLE, comm, &status))
	{
	    pi_recv_dummy(status.MPI_SOURCE, TAG_LB_WRKR_IDLE, comm, &status);
	    idle[status.MPI_SOURCE] = TRUE;
	    --num_working_workers;
	    D(printf("[IC]: Received msg WRKR_IDLE from worker %d; nww now %d\n", 
		     status.MPI_SOURCE, num_working_workers));
	}

	if (ic_is_worker)
	{
	    bool caw_was_locally_done = idle[INTG_CTRL_ID];
	    update_t *ctrl_update_p;
	    bool use_update;

	    D(printf("[IC]: Calling myself to do work, currently idle? %d\n", idle[id]));
	    
	    if (do_lb)
	    {
		bool used_helper, rejected_help;
		int helper_id;
		/* The FALSE hard-codes the fact that I (the IC) should not be chosen
		   as a helper!: */
		caw_one_pass(eps, 
			     (idle_worker(idle, num_non_ic_workers, FALSE, &helper_id)
			      ? helper_id
			      : -1), 
			     &use_update, &ctrl_update_p, &idle[id], &used_helper, &rejected_help);
		/* If a non-ic worker has their help offer rejected, they send me a WRKR_IDLE
		   message to inform me, and, I decrement num_working_workers.  The CAW informs
		   me via the rejected_help flag: */
#if 0
		/* This section disabled for now: */
		if (rejected_help)
		    --num_working_workers;
#endif
		D(if (rejected_help) 
		  printf("[IC]: Decrementing nww to %d due to rejected help at caw\n", 
			 num_working_workers));
		
		if (helper_id != -1)
		{
		    if (used_helper)
		    {
			idle[helper_id] = FALSE;
			++num_working_workers;
			D(printf("[IC]: CAW used helper id of %d; nww is now %d\n",
				 helper_id, num_working_workers));
		    }
		    else
		    {
			reset_last_used_helper();
		    }
		}
	    }
	    else
	    {
		bool locally_done;
		caw_one_pass(eps, -1, &use_update, &ctrl_update_p, &locally_done, NULL, NULL);
	    }

	    if (use_update)
	    {
		/* We have an update to use. This code here is the same as when we get an update
		   from a worker, except when the source (the controller) is not idle, we do not
		   try to send a message 3, rather we wait till its next bit of work and just 
		   pass in the id of an idle worker (if there is one): */
		if (vis_type == VIS_ON) /* !!jli*/
		    /* send error estimates and # of regions to Applet */
		    vis_send_update(new_sockfd,0,*ctrl_update_p);

		update_results(ctrl_update_p, nfcns, result, estabs, &fcncount);

		/* Since the caw always gets a new global_eps, and checks it even if it is locally
		   done, it is possible for a locally idle caw to become un-idle, even without
		   any load balancing.  This checks for this possibility.  Note that the idle
		   array is already properly updated: */
		if (caw_was_locally_done && !ctrl_update_p->l_done)
		    ++num_working_workers;
		else if (!caw_was_locally_done && ctrl_update_p->l_done)
		    --num_working_workers;
		
		D(if (caw_was_locally_done && !ctrl_update_p->l_done)
		  printf("Incremementing nww to %d due to idle caw waking up\n", 
			 num_working_workers));
		D(if (!caw_was_locally_done && ctrl_update_p->l_done)
		  printf("Decrementing nww to %d due to caw going idle (but have update)\n", 
			 num_working_workers));

		/* This handles updating other local status variables and sending out
		   done messages and new global eps.  The FALSE parameter tells the fn
		   to *not* deal with any load balancing (as mentioned above): */
		handle_update(fcn_eval_limit, num_non_ic_workers, eps_a, eps_r, result, estabs, 
			      nfcns, ctrl_update_p, id, fcncount, comm, &globally_done, 
			      fcn_limit_flag, &eps, &last_sent_eps, &loops_since_eps_last_sent,
			      idle, FALSE, send_eps_frequency, ic_is_worker, &num_working_workers);
	    }
	    else
	    {
		/* If we can not use the update, then it means we are locally idle.  Since I
		   always send in a new global_eps, it is possible for the caw to declare them-
		   selves idle without doing a last round of work.  Therefore, this checks
		   to see if this occurs, and if so, updates idle[] and num_working_workers
		   appropriately: */
		if (idle[INTG_CTRL_ID] && !caw_was_locally_done)
		{
		    --num_working_workers;
		    D(printf("[IC]: My work result: no work done; newly idle (nww now %d)!\n",
			     num_working_workers));
		    if (num_working_workers <= 0)
		    {
			globally_done = TRUE;
			/* See comment in handle_update() concerning this check: */
			if (nworkers > 1)
			    *fcn_limit_flag = TRUE;
			D(printf("[IC]: ALL WORKERS REPORT IDLE. ROUND-OFF ERROR GLOBAL STOP!\n"));
			D(printf("[IC]: Sending INTG_STOP to all workers\n"));
			send_global_stop_message(num_non_ic_workers, comm);
		    }
		}
	    }
	}
	
	/* This reads results until there are no more.  I think that there are better
	   techniques for getting worker updates, but those ideas will wait for another day: */
	while (!globally_done && pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
	{
	    ++loops_since_eps_last_sent;
	    pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, 
		    comm, &status);
	    if (vis_type == VIS_ON) /* !!jli*/
		/* send error estimates and # of regions to Applet */	
		vis_send_update(new_sockfd,status.MPI_SOURCE,wrkr_update);
	    update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
	    handle_update(fcn_eval_limit, num_non_ic_workers, eps_a, eps_r, result, estabs, 
			  nfcns, &wrkr_update, status.MPI_SOURCE, fcncount, comm, 
			  &globally_done, fcn_limit_flag, &eps, &last_sent_eps, 
			  &loops_since_eps_last_sent, idle, do_lb, send_eps_frequency,
			  ic_is_worker, &num_working_workers);
	}

	if (vis_type == VIS_ON) {/* !! jli */
	    /* receive histogram information,then send to the Applet */
	    D(printf("[IC_VIS]:try to recv a hist from worker\n"));
	    vis_recv_hist(new_sockfd,comm);
	}
	/* I need to also have this here.  Timing problems can result in the eps never being
	   sent otw.  Things would work better (IMHO) if the loop was more structured: */
	if (   !globally_done
	       &&  time_to_send_eps(eps, last_sent_eps, loops_since_eps_last_sent, 
				    send_eps_frequency))
	{
	    D(printf("[IC]: Sending NEW_EPS to all workers (b); eps is %g\n", (double) eps));
	    send_to_each_worker(&eps, 1, pi_base_mpi_t, TAG_NEW_EPS, num_non_ic_workers, comm);
	    last_sent_eps = eps;
	    loops_since_eps_last_sent = 0;
	}
    }

    /* There may be some final update messages out there. Since their comm time has
       already been paid, I might as well collect them and add them onto the result, in order
       to increase our  accuracy slightly: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_WRKR_UPDATE, comm, &status))
    {
	pi_recv(&wrkr_update, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, comm, &status);
	update_results(&wrkr_update, nfcns, result, estabs, &fcncount);
	D(printf("[IC]: Received extra WRKR_UPDATE from worker %d (new fcncount %d)\n",
		 status.MPI_SOURCE, fcncount));
    }

    /* Stop the clock, then do some clean-up things: */
    *total_time = MPI_Wtime() - start_time;
    ms_stop_clock();
    *fcn_eval_count = fcncount;

    /* jli !! visulization */
    if (vis_type == VIS_ON) {
	/*close the sockets */
	vis_close(new_sockfd);
    }

    if (ic_is_worker)
	caw_finalize_intg();
    ic_clean_msgs(update_mpi_t, comm);
    MPI_Type_free(&update_mpi_t);
    D(printf("[IC]: Leaving integ_controller()\n"));

} /* integ_controller() */

