/*        File: intg-wrkr.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI
   Description: Source to implement the integral worker functionality
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/intg-wrkr.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:32 $

                $Log: intg-wrkr.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:32  parint
                Initial checkin


       Version: 01/14/98  rrz  Original (from existing code)
		05/12/98  rrz  Code freeze for ParInt1.0
		7/7/98	  jli  add code for visulization
		09/17/98  rrz  Remove unused ms_type variable
		10/22/98  rrz  For keeping heap: Added keep heap logic
		11/17/98  rrz  Modify call to get_initial_region()
		12/18/98  rrz  Modify for changed get_initial_region()
				prototype
		07/13/99  rrz  Add logging calls
		07/14/99  rrz  BUG FIX: Add explicit check after receiving a
				region to see if we are idle or not
		07/19/99  rrz  Modify logging of recv-helper and recv-rgn
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
		08/20/99  rrz  Add in: measuring idle time
		10/14/99  rrz  BUG FIX: Call the send update logging function
				before sending the update so that the update#
				gets set properly
		03/18/00  rrz  Add changes for using deap
		06/01/00  rrz  Eliminate several compiler warnings
*/

/* #define DBG_THIS_FILE */

#include <stdio.h>

#include <mpi.h>
#include <parint.h>
#include <pi-defs.h>
#include <pi-vspec.h>
#include <math-adapt.h>
#include <math-intg.h>
#include <math-rgn.h>
#include <measures.h>
#include <messaging.h>
#include <heapdeap.h>
#include <visual.h> /* !!jli */
#include <logging.h>

#define half 0.5

static MPI_Datatype hist_mpi_t; /* jli!! */

/* Analogous to its partner in the controller code.  This cleans up messages that the
 * worker expects to have to clean up.
 */
static void iw_clean_msgs(MPI_Datatype region_mpi_t, MPI_Comm comm)
{
    MPI_Status status;
    
    /* Clean new eps messages: */
    while (pi_iprobe(IC_ID, TAG_NEW_EPS, comm, &status))
    {
	pi_base_t eps;
	pi_recv(&eps, 1, pi_base_mpi_t, IC_ID, TAG_NEW_EPS, comm, &status);
    }

    /* Clean controller-sending-help messages: */
    while (pi_iprobe(IC_ID, TAG_LB_HELPER_ID, comm, &status))
    {
	int helper_id;
	pi_recv(&helper_id, 1, MPI_INT, IC_ID, TAG_LB_HELPER_ID, comm, &status);
    }

    /* Clean sent-region messages: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_LB_SEND_RGN, comm, &status))
    {
	region_t newrgn;
	pi_recv(&newrgn, 1, region_mpi_t, status.MPI_SOURCE, TAG_LB_SEND_RGN, comm, &status);
    }
    
    /* jli!! */
    /* Clean threshold messages: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_VIS_THRESHOLD, comm, &status))
    {
	double threshold;;
	pi_recv(&threshold, 1, MPI_DOUBLE, status.MPI_SOURCE, TAG_VIS_THRESHOLD, comm, &status);
    }

    /* Clean send-histogram messages: */
    while (pi_iprobe(MPI_ANY_SOURCE, TAG_HIST, comm, &status))
    {
        hist_t hist; 
        pi_recv(&hist, 1, hist_mpi_t, status.MPI_SOURCE, TAG_HIST, comm, &status);
    }

} /* iw_clean_msgs() */


/* Note that nworkers will include the IC, if the IC is also working
 */
extern void integ_worker(pi_ifcn_t ifcn_ptr, int ndims, int nfcns, int irule,
			 pi_base_t a[], pi_base_t b[], int nworkers, int id, MPI_Comm comm, 
			 pi_rparm_t *rparms_p)
{
    int fcns_per_basrul = neval(ndims, irule);
    bool globally_done = FALSE;
    bool locally_done  = FALSE;
    bool rec78 = FALSE; 
    bool updateNotsent = TRUE;
    
    pi_base_t full_vol;            /* Volume of original full region (defined by a and b) */
    pi_base_t local_vol;           /* Volume of workers regions in the heap               */
    pi_base_t size_ratio;          /* Ratio local_vol / full_vol                          */
    pi_base_t l_eps;               /* Local error estimate                                */
    pi_base_t g_eps;               /* Global error estimate                               */
    pi_base_t estabs[PI_MAX_NFCN]; /* Local ESTABS                                        */

    MPI_Datatype region_mpi_t;
    MPI_Datatype update_mpi_t;
    heapdeap_t hpdp;
    
    /* This is the structure sent to the IC.  Note that the worker tells the IC it is done
       via the l_done fields in wrkr_update. I [rrz-01/17/98] split this out from the locally_done
       variable since the lifetime of the wrkr_update variable, as a whole, is different from
       the locally_done variable: */
    update_t wrkr_update;
    
    /* These are used temporarily where needed: */
    MPI_Status status;
    region_t rgn;
    
    /* Transfer run parameters from rparms to local variables: */
    bool do_lb           = rparms_p->do_lb;
    double lb_help_ratio = rparms_p->lb_help_ratio;
    int ns               = rparms_p->ns;
    bool using_deap      = rparms_p->max_heap_size > 0;
    int vis_type         = rparms_p->vis_type;/*!jli visualization */
    double start_time = 0.0;
    double current_time;
    double vis_threshold;
    /* end jli! */

    int max_heap_size = using_deap ? rparms_p->max_heap_size : -1;

    D(printf("[IW-%d]: Beginning integ_worker(); fcns_per_basrul is %d; lb is %d; ns is %d\n",
	     id, fcns_per_basrul, do_lb, ns));

    create_mpi_type_region(&region_mpi_t);
    create_mpi_type_update(&update_mpi_t);

    pl_iw_init(ndims, nfcns, irule, rparms_p, nworkers + (rparms_p->ic_is_worker ? 0 : 1), id);
    pl_log_iw_start();
    ms_start_clock();
    if (vis_type == VIS_ON) /* !!jli vis */
    {
	start_time = MPI_Wtime();
	create_mpi_type_hist(&hist_mpi_t);
	 /* Block receive a threshold message  */
    	pi_probe(IC_ID, MPI_ANY_TAG, comm, &status);
    	pi_assert((status.MPI_TAG == TAG_VIS_THRESHOLD), 
		  "Initial message received was not TAG_VIS_THRESHOLD");
	pi_recv(&vis_threshold ,1, MPI_DOUBLE,IC_ID,TAG_VIS_THRESHOLD,comm,&status);
    }      

    /* Each worker processor ( 1<=id<=nproc1 ) calculates its starting subregion; calculates
       its local RESULT and ESTABS, maintaining its own independent heap based on a region's
       error (rgnerr); each worker periodly reports its local RESULT and ESTABS to the
       controller process: */
    
    clear_update(&wrkr_update, nfcns);
    clear_estabs(estabs, nfcns);
    get_initial_region(ndims, nworkers, a, b, id, &rgn, &full_vol, &local_vol, rparms_p);
    size_ratio = local_vol / full_vol;
    D(printf("[IW-%d]: My region: local_vol: %g; full_vol: %g; ratio: %f\n", 
	     id, (double) local_vol, (double) full_vol, (double) size_ratio));
    
    /*  Initialize the heap with results from the initial region: */
    work_initialize(ndims, irule);
    work_do_round(ndims, nfcns, ifcn_ptr, &rgn);
    
    update_update(&wrkr_update, estabs, &rgn, nfcns, fcns_per_basrul);
    pl_log_work(pl_work_init, &rgn, NULL, estabs, wrkr_update.cnt, 0.0);
    heap_create(&hpdp, using_deap);
    heap_add_node(&hpdp, &rgn);
    
    /* !! jli visualization */
   if (vis_type == VIS_ON)
   {
       long num_region;
       current_time = MPI_Wtime();
       vis_init_min_err(rgn);
       /* update l_eps, #regions, etc. */
       num_region = (rgn.rgnerr > vis_threshold?1:0);
       vis_update_update(&wrkr_update, estabs, nfcns, current_time - start_time, num_region);
   }

    ms_update_fcncount(fcns_per_basrul);

    /* Send the initial result as an INIT_UPDATE ([BUG-FIX-R10/15/99-RRZ] Call pl_log_send_upd()
       first, so that it can set the update_num inside wrkr_update): */
    pl_log_send_upd(pl_upd_init, &wrkr_update);
    pi_send(&wrkr_update, 1, update_mpi_t, IC_ID, TAG_INIT_UPDATE, comm); 
    D(printf("[IW-%d]: Sent INIT_UPDATE to controller\n", id));
 
    /* Block receive either a stop message or an eps update: */ 
    pi_probe(IC_ID, MPI_ANY_TAG, comm, &status);
    pi_assert((status.MPI_TAG == TAG_INTG_STOP || status.MPI_TAG == TAG_INIT_EPS),
	      "Initial message received was not INTG_STOP or INIT_EPS");
    
    if (status.MPI_TAG == TAG_INTG_STOP)
    {
	pi_recv_dummy(IC_ID, TAG_INTG_STOP, comm, &status);
	globally_done = TRUE;
	l_eps = 0.0;
	D(printf("[IW-%d]: Received INTG_STOP (in place of INIT_EPS)\n", id));
    }
    else
    {
	pi_recv(&g_eps, 1, pi_base_mpi_t, IC_ID, TAG_INIT_EPS, comm, &status);
	l_eps = g_eps * size_ratio;
	rec78 = TRUE; 
	pl_log_recv_eps(pl_eps_init, g_eps, l_eps, err_ratio_2(estabs, nfcns, l_eps));
	D(printf("[IW-%d]: Received INIT_EPS (g_eps) of %g\n", id, (double) g_eps));
    }
    
    if (!globally_done) 
    {
	if (result_close_enough(estabs, l_eps, nfcns))
	{
	    locally_done = TRUE;
	    pi_send_dummy(IC_ID, TAG_LB_WRKR_IDLE, comm);
	    pl_log_send_idle(pl_idle_quick);
	    ms_update_idle(locally_done);
	    D(printf("[IW-%d]: Sending a quick WRKR_IDLE to IC\n", id));
	}
	else
	{
	    locally_done = FALSE;
	}
	
	clear_update(&wrkr_update, nfcns);
	updateNotsent = FALSE;  /* initial result was sent */
    }

    D(printf("[IW-%d]: Starting main loop\n", id));

    /* Enter the main loop; keep looping until we are globally done: */
    while (!globally_done)
    {
	/* Check if it is time to stop: globally or locally. Current results are sent if and
	   only if not globally done;  chk for message from controller of load balancing help;
	   if locally done, possibly receive region from another worker: */

	/* First check done message from the controller: */
	if (pi_iprobe(IC_ID, TAG_INTG_STOP, comm, &status))
	{
	    pi_recv_dummy(IC_ID, TAG_INTG_STOP, comm, &status);
	    globally_done = TRUE;
	    D(printf("[IW-%d]: Received message INTG_STOP from %d\n", id, status.MPI_SOURCE));
	}
	else /* not globally done */
	{
	    if (vis_type == VIS_ON) {/* !!jli */
		/* check to see if there's messages of threshold or histogram inquiry */
		/* Non Block receive either a threshold message or a histogram */
		int recv = 0;
    		recv = pi_iprobe(IC_ID, MPI_ANY_TAG, comm, &status);
		D(printf("[VIS_WK-%d]probe request from IC: %d\n",id,recv));

    		if (recv &&status.MPI_TAG == TAG_HIST)
    		{
		    /* get the histogram and send back */
		    hist_t hist;
		    pi_recv(&hist,1,hist_mpi_t,IC_ID, TAG_HIST, comm, &status);
		    D(printf("[VIS_WK-%d]recv hist from IC\n",id));
		    current_time = MPI_Wtime();
		    vis_get_hist(hpdp.maxrtptr, &hist, 1, current_time - start_time);
		    pi_send(&hist, 1, hist_mpi_t, IC_ID, TAG_HIST, comm);
		    D(printf("[VIS_WR-%d]send hist to IC\n",id));
    		}    
    		else if (recv && status.MPI_TAG == TAG_VIS_THRESHOLD)
    		{   
		    pi_recv(&vis_threshold, 1, MPI_DOUBLE, IC_ID, TAG_VIS_THRESHOLD, 
			    comm, &status);
    		}
	    }

	    /* If not globally done, then do some work: */
	    if (!locally_done)
	    {
		region_t oldrgn, new[REGION_SPLITS];
		int i, j, k;
		
		updateNotsent = TRUE;

		for (i = 0; i < ns; i++)
		{
		    /* Get a region with the largest error: */
		    oldrgn = *heap_peek_max(&hpdp);

		    for (k = 0; k < nfcns; k++)
		    {
			wrkr_update.res[k] -= oldrgn.rgnval[k];
			wrkr_update.est[k] -= oldrgn.rgnerr_v[k];
			estabs[k] -= oldrgn.rgnerr_v[k];
		    }

		    /* Split the retrieved region: */
		    split_region(new, &oldrgn);

		    /* Calc the integral over each of the new regions: */
		    for (j = 0; j < REGION_SPLITS; j++ )
		    {
			/* This does a complete round of work, including the updating
			   of rgnerr and divaxn in new[j]: */
			work_do_round(ndims, nfcns, ifcn_ptr, &new[j]);
 
			/* Update the worker's update and add the new region onto the heap: */
			update_update(&wrkr_update, estabs, &new[j], nfcns, fcns_per_basrul);
			pl_log_work((j == 0 ? pl_work_first : pl_work_second), 
				    &new[j], &oldrgn, estabs, wrkr_update.cnt, l_eps);
			if (j == 0)
			    heap_replace_root(&hpdp, &new[j]);
			else
			    heap_add_node(&hpdp, &new[j]);
			/* !! jli */
			if (vis_type == VIS_ON)
			{
			    long num_region;
			    vis_min_rgn_err(new[j]);
			    current_time = MPI_Wtime();
			    num_region = vis_get_num_region(hpdp.maxrtptr, vis_threshold);
			    /* update l_eps, #regions, etc. */
			    vis_update_update(&wrkr_update, estabs, nfcns, 
					      current_time - start_time, num_region);
   			}

			ms_update_fcncount(fcns_per_basrul);
		    }

		    if (using_deap && heap_size(&hpdp) > max_heap_size)
		    {
			region_t rgn_to_throw_out;
			deap_delete_min(&hpdp, &rgn_to_throw_out);
			D(printf("[IW-%d]: Threw out region %G\n", id, rgn_to_throw_out.rgnerr));
		    }
		}
		D(printf("[IW-%d]: IW after ns loop (fcns this upd %d; eps ratio %.2f; hpsz %d)\n",
			 id, wrkr_update.cnt, (double) (infNorm(estabs, nfcns) / l_eps),
			 heap_size(&hpdp)));
	    }

	    /* After doing a unit of work, check for various incoming messages.  First check
	       for latest eps update (if any) from controller.  We want to loop and receive
	       until there are no more to receive, in case several are waiting to be received: */
	    while (pi_iprobe(IC_ID, TAG_NEW_EPS, comm, &status))
	    {
		pi_recv(&g_eps, 1, pi_base_mpi_t, IC_ID, TAG_NEW_EPS, comm, &status);
		l_eps = g_eps * size_ratio;
		pl_log_recv_eps(pl_eps_reg, g_eps, l_eps, err_ratio_2(estabs, nfcns, l_eps));
		rec78 = TRUE;
		D(printf("[IW-%d]: Received NEW_EPS (g_eps) of %g\n", id, (double) g_eps));
	    }

	    /* Next, check to see if old result was close enough, and maybe send an
	       update to the controller: */
	    locally_done = result_close_enough(estabs, l_eps, nfcns);
	    ms_update_idle(locally_done);
	    
	    if (rec78 && updateNotsent)
	    {
		/* Send update to controller.  The l_done field allows us to piggyback a
		   locally-done msg on the update message: */
		wrkr_update.l_done = locally_done;
		pi_send(&wrkr_update, 1, update_mpi_t, IC_ID, TAG_WRKR_UPDATE, comm);
		pl_log_send_upd(pl_upd_reg, &wrkr_update);
		D(printf("[IW-%d]: Sent WRKR_UPDATE to controller\n", id));
		rec78 = FALSE; 
		updateNotsent = FALSE;
		clear_update(&wrkr_update, nfcns);
		D(printf("[IW-%d]: gdone %d l_done %d infNorm %g l_eps %g\n",
			 id, globally_done, locally_done, (double) infNorm(estabs, nfcns),
			 (double) l_eps));
	    }
	    
	    /* Only perform the rest of the loop body if we are load-balancing: */
	    if (do_lb)
	    {
		/* Now check for message from controller sending help information: */
		if (pi_iprobe(IC_ID, TAG_LB_HELPER_ID, comm, &status))
		{
		    int helper_id;
		    
		    pi_recv(&helper_id, 1, MPI_INT, IC_ID, TAG_LB_HELPER_ID, comm, &status);
		    
		    if (locally_done || (infNorm(estabs,nfcns) <= lb_help_ratio * l_eps))
		    {
			/* We are locally done, or, close enough to reject the help: */
			rgn.divaxn = -1;
			pl_log_recv_helper(pl_lb_nowork, err_ratio_2(estabs, nfcns, l_eps), NULL,
					   0.0, 0.0, NULL);
			D(printf("[IW-%d]: Received helper id of %d; rejecting offer of help "
				 "(why: l_done: %d; factor: %f)\n", id, helper_id, locally_done,
				 (double) (infNorm(estabs, nfcns) / l_eps)));
		    }
		    else
		    {
			/* Send a region.  Delete it from the heap, and update estabs, the local
			   volume, and the size_ratio: */
			int k;
#ifdef USE_LOGGING
			pi_base_t old_err_ratio = err_ratio_2(estabs, nfcns, l_eps);
#endif			
			D(printf("[IW-%d]: Received helper id of %d; accepting offer of help "
				 "(factor: %f)\n", id, helper_id, 
				 (double) (infNorm(estabs, nfcns) / l_eps)));
			heap_delete_max(&hpdp, &rgn);
			for (k = 0; k < nfcns; k++)
			    estabs[k] -= rgn.rgnerr_v[k]; 
			local_vol -= rgn.rgnvol;
			size_ratio = local_vol / full_vol;
			l_eps = g_eps * size_ratio;
			pl_log_recv_helper(pl_lb_sending, old_err_ratio, &rgn, size_ratio,
					   err_ratio_2(estabs, nfcns, l_eps), estabs);
		    }
		
		    pi_send(&rgn, 1, region_mpi_t, helper_id, TAG_LB_SEND_RGN, comm);
		}
	    
		/* If locally done, check for region from another worker: */
		if (locally_done && pi_iprobe(MPI_ANY_SOURCE, TAG_LB_SEND_RGN, comm, &status))
		{
		    pi_recv(&rgn, 1, region_mpi_t, status.MPI_SOURCE, TAG_LB_SEND_RGN,
			    comm, &status);
		
		    if (rgn.divaxn >= 0)
		    {
			/* Help accepted.  Add new region to heap, then update *all*, 
			   affected local variables: */
			int k;
			heap_add_node(&hpdp, &rgn);
			local_vol += rgn.rgnvol;
			for (k = 0; k < nfcns; k++)
			    estabs[k] += rgn.rgnerr_v[k];
			size_ratio = local_vol / full_vol;
			l_eps = g_eps * size_ratio;
			if (using_deap && heap_size(&hpdp) > max_heap_size)
			{
			    region_t rgn_to_throw_out;
			    deap_delete_min(&hpdp, &rgn_to_throw_out);
			    D(printf("[IW-%d]: Threw out region (after lb) %G\n", id, 
				     rgn_to_throw_out.rgnerr));
			}
			/* BUG FIX (RRZ-07/13/99): We used to assume that the sent region would
			   make us non-idle.  This was a bad asumption.  Now I make a check: */
			locally_done = result_close_enough(estabs, l_eps, nfcns);
			ms_update_idle(locally_done);
			pl_log_recv_rgn((locally_done ? pl_lb_fail : pl_lb_success), 
					status.MPI_SOURCE, &rgn, size_ratio, 
					l_eps, estabs, locally_done);
			if (locally_done)
			{
			    /* Furthermore, if we are locally done, then I must report this
			       to the IC, since he thinks we aren't, since he knew about this
			       instance of load balancing: */
			    pi_send_dummy(IC_ID, TAG_LB_WRKR_IDLE, comm);
			    pl_log_send_idle(pl_idle_lb_fail);
			}
			if (vis_type == VIS_ON)
			    vis_min_rgn_err(rgn);
			D(printf("[IW-%d]: Received SEND_RGN w/rgn from %d; new size_ratio "
				 "%G; new erratio %G%s\n", id, status.MPI_SOURCE, size_ratio,
				 err_ratio_2(estabs, nfcns, l_eps),
				 (locally_done ? "; STILL IDLE" : "")));
		    }
		    else
		    {
			/* Help rejected; inform controller still idle: */
			pi_send_dummy(IC_ID, TAG_LB_WRKR_IDLE, comm);
			pl_log_recv_rgn(pl_lb_nowork, status.MPI_SOURCE, NULL, 0.0, 0.0, NULL, 
					TRUE);
			pl_log_send_idle(pl_idle_lb_nowork);
			D(printf("[IW-%d]: Received SEND_RGN from %d, but help rejected "
				 "(sending WRKR_IDLE to IC)\n", id, status.MPI_SOURCE));
		    }
		}
	    }
	}
    }

    ms_update_idle(locally_done);
    ms_stop_clock();
    pl_log_iw_stop();
    heap_free(&hpdp);
    iw_clean_msgs(region_mpi_t, comm);
    MPI_Type_free(&update_mpi_t);
    MPI_Type_free(&region_mpi_t);
    pl_finalize();
    
} /* integ_worker() */

