/*        File: logging.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI (logging)
   Description: All source needed for logging (but not used by filter programs)
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/logging.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:33 $

                $Log: logging.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:33  parint
                Initial checkin


       Version: 05/21/99  rrz  Original
		07/14/99  rrz  Remove killheap/keepheap option
		07/19/99  rrz  Small change to event writing of recv-rgn
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

/* Is there a better way to do this?  I do not know it.  The #include of pi-version.h
 * must be first: */
#include <pi-version.h>
#ifdef USE_LOGGING

#define DBG_THIS_FILE

#include <stdio.h>      /* for sprintf() */
#include <sys/types.h>  /* for open(), close(), ... */
#include <sys/stat.h>   /* for open(), close(), ... */
#include <fcntl.h>      /* for open(), close(), ... */
#include <unistd.h>

#include <timer.h>
#include <math-misc.h>
#include <math-adapt.h>   /* for err_ratio(), eps_ratio() macros */
#include <math-intg.h>    /* for neval() */

#include <logging.h>
#include <log-filter.h>

/* Like memcpy(), but the destination pointer is incremented by the number of bytes copied.
 * variable_size should be in # of bytes, and buf_p should be a ptr to a single byte value
 * (so the pointer arithmetic works ok): */
#define lmemcpy(buf_p, variable, variable_size) {    \
    memcpy(buf_p, &(variable), variable_size);       \
    buf_p += variable_size; }

/* The same, but used when the source is an array: */
#define lmemcpy_array(buf_p, array, array_byte_size) {    \
    memcpy(buf_p, array, array_byte_size);                \
    buf_p += array_byte_size; }

/* The size of the internal buffer gathered before writing to disk: */
#define LOG_BUFFER_SIZE     8192

/* This is set (based on plfile) by pl_ic_init() or pl_iw_init(); if FALSE, then it means that
 * the logging is turned on at compile-time, but no plfile is specified, so no logging will
 * be done.  All externally callable functions check this variable, and if it is not set,
 * then they return immediately: */
static bool doing_logging = FALSE;

/* The next variables are global for this file; they store needed parameters about the integration
 * being solved.  They are initialized in pi_ic_init() and pl_iw_init(): */
static int ndims, nfcns, fcns_per_irule, nPE;
static pl_rank_t id;
static bool ic_is_working;
static bool logging_initialized = FALSE;

/* For the large buffer that holds multiple records in memory until we do an actual write().
 * It is a char* for ptr arithmetic: */
static char *log_buffer = NULL;

/* The 0-indexed byte position where the beginning of the *next* record should go in the
 * log_buffer.  When this hits LOG_BUFFER_SIZE, the buffer is full (i.e., the last usable
 * value is LOG_BUFFER_SIZE-1): */
static int log_buffer_pos = 0;

/* The full path/file name of the log file, and the corresponding file descrptor: */
static char log_file_name[PI_MAX_PATH + 1];
static int log_file_fd;

/* Holds a single log record before being transferred to the log_buffer: */
static char *logrec_buffer = NULL;

/* Next number values, for the IC and IW */
static int logrec_next_num, update_next_num, region_next_num;

/* Record sizes for the events.  The order within this array is based on the event's
 * enumerated order, with the 0th entry blank (since I start all enumerated values at 1): */
static int event_rec_sizes[pl_event_max];

/**********************************************************************************************/
/********** Internal routines                                                        **********/

/* Called for all events; this puts into the buffer the information common to all events
 */
static void prepare_log_buffer(pl_enum_t event_type, char **buffer_p)
{
    pl_time_t log_time = timer_retrieve();
    lmemcpy(*buffer_p, event_type, SZ_ENUM_T);
    lmemcpy(*buffer_p, logrec_next_num, SZ_NNUM_T);
    lmemcpy(*buffer_p, log_time, sizeof(pl_time_t));
    lmemcpy(*buffer_p, id, SZ_RANK_T);
    logrec_next_num += 1;
    D(printf("[PL]: In prepare_log_buffer(): Event %d; Rec# %ud; Time %f; ID %d\n",
	     event_type, logrec_next_num, log_time, id));
} /* prepare_log_buffer() */


static void initialize_logging(const pi_rparm_t *rparms_p, int id)
{
    sprintf(log_file_name, "%s/%s-%02d.log", PL_LOG_DIR, rparms_p->pl_file, id);

    if ((log_file_fd = open(log_file_name, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC, 
			    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) < 0)
    {
	pi_sys_terminate("Unable to open log file '%s'", log_file_name);
    }
    
    /* Allocate memory for the large buffer that is used before doing an actual write() call: */
    if (!(log_buffer = (char *) malloc(LOG_BUFFER_SIZE)))
	pi_sys_terminate("Error from malloc() of log_buffer");

    memset(log_buffer, '\0', LOG_BUFFER_SIZE);
    log_buffer_pos = 0;

    logging_initialized = TRUE;
} /* initialize_logging() */


static void buffer_flush(void)
{
    if (log_buffer_pos > 0)
    {
	if (write(log_file_fd, log_buffer, log_buffer_pos) != log_buffer_pos)
	    pi_sys_terminate("Error writing record to log file '%s'", log_file_name);
	log_buffer_pos = 0;
    }
} /* buffer_flush() */


static void buffer_log_record(int rec_len)
{
    pi_assert(logging_initialized, "Init problem in buffer_log_record()");
    pi_assert(log_buffer, "Init problem in buffer_log_record()");

    D(printf("In buffer_log_record(); rec_len is %d; buffer_pos is %d\n",
	     rec_len, log_buffer_pos));
    
    /* This seems simple, but is actually a tricky conditional, as we are mingling
       size and index values in these expressions: */
    if (log_buffer_pos + rec_len <= LOG_BUFFER_SIZE)
    {
	memcpy(log_buffer + log_buffer_pos, logrec_buffer, rec_len);
	log_buffer_pos += rec_len;
    }
    else            /* (time to flush the buffer) */
    {
	buffer_flush();
	memcpy(log_buffer, logrec_buffer, rec_len);
	log_buffer_pos = rec_len;
    }
} /* buffer_log_record() */


/***********************************************************************************************/
/********** Externally provided routines                                              **********/

/**** Note on the logging functions: */
/* I squeeze everything together in the logged record.  Therefore, I cannot simply
 * assign variables to positions in the log record (due to restrictions on where, e.g., a double
 * can be stored in memory).  Therefore I use memcpy() to copy all the necessary bytes into 
 * position in memory.  To make sure that I copy from the correct variable size, I temporarily
 * store the value to be copied in a variable of the correct type (I am not sure enough of
 * how casts work to rely on them in what seems to me to be an unsafe situation).  Then I memcpy()
 * it (via the lmemcpy() macro).  I do this also to store temporary values where actually needed.
 * Also, compiler warnings can result if you pass a constant integer into a one-byte parameter
 * (unless you cast), so this technique allows the calling code to pass parameters without
 * casting or using funny variable types.
 */

/***** The next few routines are the logging functions for the IC */

extern void _pl_log_ic_stop(int pl_stop_type)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
    
	pi_assert(logging_initialized, "Initialization problem in pl_log_ic_stop()");
	prepare_log_buffer(pl_event_ic_stop, &buffer_p);

	t_enum = pl_stop_type; lmemcpy(buffer_p, t_enum, SZ_ENUM_T);

	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_ic_stop],
		  "Bad memcpy()'s for pl_log_ic_stop()");
	buffer_log_record(event_rec_sizes[pl_event_ic_stop]);
    }
} /* pl_log_ic_stop() */


extern void _pl_log_recv_upd(int pl_upd_type, const update_t *update_p, int sender_id, 
			     int fcn_count, pi_base_t eps, pi_base_t last_sent_eps,
			     const pi_base_t result[], const pi_base_t estabs[], 
			     const bool idle_workers[])
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pl_rank_t t_rank;
	bool t_bool;
	ulong t_ulong;
	pl_base_t t_base;

	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_upd()");
	prepare_log_buffer(pl_event_recv_upd, &buffer_p);

	t_enum = pl_upd_type;               
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	t_rank = sender_id;
	lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	lmemcpy(buffer_p, update_p->update_num, SZ_NNUM_T);
	t_ulong = fcn_count / fcns_per_irule;
	lmemcpy(buffer_p, t_ulong, SZ_ULONG_T);
	if (pl_upd_type == pl_upd_init)
	{
	    int i;
	    t_bool = FALSE;
	    t_base = 0.0;
	    lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    for (i = 0; i < nfcns * 2; ++i)
		lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    for (i = 0; i < nPE; ++i)
		lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	}
	else if (pl_upd_type == pl_upd_braking)
	{
	    int i;
	    t_bool = update_p->l_done;
	    lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	    t_base = 0.0;
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_base = infNorm(estabs, nfcns);   
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_base = 0.0;
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy_array(buffer_p, result, SZ_BASE_T * nfcns);
	    lmemcpy_array(buffer_p, estabs, SZ_BASE_T * nfcns);
	    for (i = 0; i < nPE; ++i)
		lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	}
	else
	{
	    t_bool = update_p->l_done;
	    lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	    lmemcpy(buffer_p, eps, SZ_BASE_T);
	    t_base = eps_ratio(eps, last_sent_eps);  
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_base = infNorm(estabs, nfcns);   
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_base = err_ratio(t_base, eps);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy_array(buffer_p, result, SZ_BASE_T * nfcns);
	    lmemcpy_array(buffer_p, estabs, SZ_BASE_T * nfcns);
	    lmemcpy_array(buffer_p, idle_workers, SZ_BOOL_T * nPE);
	}
	
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_recv_upd],
		  "Bad memcpy()'s for pl_log_recv_upd()");
	buffer_log_record(event_rec_sizes[pl_event_recv_upd]);
    }
} /* pl_log_recv_upd() */


extern void _pl_log_recv_idle(int sender_id, const bool idle_workers[])
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_rank_t t_rank;

	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_idle()");
	prepare_log_buffer(pl_event_recv_idle, &buffer_p);
	
	t_rank = sender_id;    lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	lmemcpy_array(buffer_p, idle_workers, SZ_BOOL_T * nPE);

	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_recv_idle],
		  "Bad memcpy() for pl_log_recv_idle()");
	buffer_log_record(event_rec_sizes[pl_event_recv_idle]);
    }
} /* pl_log_recv_idle() */


extern void _pl_log_send_eps(int pl_eps_type, pi_base_t eps, pi_base_t last_sent_eps)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pl_base_t t_base;

	pi_assert(logging_initialized, "Initialization problem in pl_log_send_eps()");
	D(printf("[PL]: In pl_log_send_eps(); logging eps of %f\n", eps));
	prepare_log_buffer(pl_event_send_eps, &buffer_p);
    
	t_enum = pl_eps_type;  lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	lmemcpy(buffer_p, eps, SZ_BASE_T);
	t_base = eps_ratio(eps, last_sent_eps);  lmemcpy(buffer_p, t_base, SZ_BASE_T);
    
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_send_eps], 
		  "Bad memcpy()'s for pl_log_send_eps()");
	buffer_log_record(event_rec_sizes[pl_event_send_eps]);
    }
} /* pl_log_send_eps() */


extern void _pl_log_send_helper(int recipient_id, int helper_id, const bool idle_workers[])
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_rank_t t_rank;
    
	pi_assert(logging_initialized, "Initialization problem in pl_log_send_helper()");
	prepare_log_buffer(pl_event_send_helper, &buffer_p);
    
	t_rank = recipient_id; lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	t_rank = helper_id;    lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	lmemcpy_array(buffer_p, idle_workers, SZ_BOOL_T * nPE);

	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_send_helper], 
		  "Bad memcpy()'s for pl_log_send_helper()");
	buffer_log_record(event_rec_sizes[pl_event_send_helper]);
    }
} /* pl_log_send_helper() */


extern void _pl_log_iw_start(void)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pi_assert(logging_initialized, "Initialization problem in pl_log_iw_start()");
	prepare_log_buffer(pl_event_iw_start, &buffer_p);
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_iw_start], 
		  "Bad memcpy()'s for pl_log_iw_start()");
	buffer_log_record(event_rec_sizes[pl_event_iw_start]);
    }
} /* pl_log_iw_start() */


extern void _pl_log_send_upd(int pl_upd_type, update_t *update_p)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	ulong t_ulong;
	bool t_bool;
	
	pi_assert(logging_initialized, "Initialization problem in pl_log_send_upd()");
	prepare_log_buffer(pl_event_send_upd, &buffer_p);
	t_enum = pl_upd_type;                      lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	lmemcpy(buffer_p, update_next_num, SZ_NNUM_T);
	update_p->update_num = update_next_num++;
	t_bool = update_p->l_done;                 lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	t_ulong = update_p->cnt / fcns_per_irule;  lmemcpy(buffer_p, t_ulong, SZ_ULONG_T);
	lmemcpy_array(buffer_p, update_p->est, nfcns * SZ_BASE_T);
	lmemcpy_array(buffer_p, update_p->res, nfcns * SZ_BASE_T);
	  
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_send_upd],
		  "Bad memcpy()'s for pl_log_send_upd()");
	buffer_log_record(event_rec_sizes[pl_event_send_upd]);
    }
} /* pl_log_send_upd() */


extern void _pl_log_work(int pl_work_type, region_t *rgn_p, const region_t *parent_rgn_p,
			 const pl_base_t estabs[], int fcn_count, pl_base_t local_eps)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pl_base_t t_base;
	ushort t_ushort;
	pi_assert(logging_initialized, "Initialization problem in pl_log_work()");

	/* Store information in the region: */
	rgn_p->rgn_num = region_next_num++;
	rgn_p->rgn_creator_id = id;

	prepare_log_buffer(pl_event_work, &buffer_p);
	t_enum = pl_work_type;
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	lmemcpy(buffer_p, rgn_p->rgnvol, SZ_BASE_T);
	lmemcpy(buffer_p, rgn_p->rgnerr, SZ_BASE_T);
	lmemcpy_array(buffer_p, rgn_p->center, SZ_BASE_T * ndims);
	lmemcpy_array(buffer_p, rgn_p->h_width, SZ_BASE_T * ndims);
	lmemcpy_array(buffer_p, rgn_p->rgnval, SZ_BASE_T * nfcns);
	lmemcpy_array(buffer_p, rgn_p->rgnerr_v, SZ_BASE_T * nfcns);
	t_ushort = rgn_p->divaxn;
	lmemcpy(buffer_p, t_ushort, SZ_USHORT_T);
	if (parent_rgn_p)
	{
	    lmemcpy(buffer_p, parent_rgn_p->rgn_num, SZ_NNUM_T);
	    lmemcpy(buffer_p, parent_rgn_p->rgn_creator_id, SZ_RANK_T);
	}
	else
	{
	    pl_nnum_t t_nnum = 0;
	    pl_rank_t t_rank = 0;
	    /* Note: The 1st region has a parent rgn # of 0, and a parent rgn creator of 0: */
	    lmemcpy(buffer_p, t_nnum, SZ_NNUM_T);
	    lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	}
	lmemcpy(buffer_p, rgn_p->rgn_num, SZ_NNUM_T);
	lmemcpy(buffer_p, rgn_p->rgn_creator_id, SZ_RANK_T);
	lmemcpy_array(buffer_p, estabs, SZ_BASE_T * nfcns);
	/* Note: Usually I use a ulong for region counts, but this is only an incremental
	   region count, so I use a ushort: */
	t_ushort = fcn_count / fcns_per_irule;
	lmemcpy(buffer_p, t_ushort, SZ_USHORT_T);
	t_base = infNorm(estabs, nfcns);
	lmemcpy(buffer_p, t_base, SZ_BASE_T);
	t_base = (pl_work_type == pl_work_init ? 0.0 : err_ratio(t_base, local_eps));
	lmemcpy(buffer_p, t_base, SZ_BASE_T);
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_work],
		  "Bad memcpy()'s for pl_log_work()");
	buffer_log_record(event_rec_sizes[pl_event_work]);
    }
} /* pl_log_work() */



extern void _pl_log_recv_rgn(int pl_lb_type, int sender_id, const region_t *rgn_p,
			     pi_base_t size_ratio, pi_base_t local_eps, const pi_base_t estabs[],
			     bool locally_done)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pl_base_t t_base;
	pl_rank_t t_rank;
	bool t_bool;
	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_rgn()");
	prepare_log_buffer(pl_event_recv_rgn, &buffer_p);
	
	t_rank = sender_id;
	lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	t_enum = pl_lb_type;
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	if (pl_lb_type == pl_lb_success || pl_lb_type == pl_lb_fail)
	{
	    lmemcpy(buffer_p, rgn_p->rgn_num, SZ_NNUM_T);
	    lmemcpy(buffer_p, rgn_p->rgn_creator_id, SZ_RANK_T);
	    lmemcpy_array(buffer_p, estabs, SZ_BASE_T * nfcns);
	    lmemcpy(buffer_p, size_ratio, SZ_BASE_T);
	    lmemcpy(buffer_p, local_eps, SZ_BASE_T);
	    t_base = infNorm(estabs, nfcns);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_base = err_ratio(t_base, local_eps);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    t_bool = !locally_done;
	    lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	}
	else
	{
	    /* The worker got a dummy region; I need to write out as many bytes as when there
	       is a valid region: */
	    pl_nnum_t t_nnum = 0;
	    int i;
	    t_rank = 0;
	    t_base = 0.0;
	    t_bool = FALSE;
	    lmemcpy(buffer_p, t_nnum, SZ_NNUM_T);
	    lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	    for (i = 0; i < nfcns; ++i)
		lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_bool, SZ_BOOL_T);
	}
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_recv_rgn],
		  "Bad memcpy()'s for pl_log_recv_rgn()");
	buffer_log_record(event_rec_sizes[pl_event_recv_rgn]);
    }
} /* pl_log_recv_rgn() */


extern void _pl_log_iw_stop(void)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;

	pi_assert(logging_initialized, "Initialization problem in pl_log_iw_stop()");
	prepare_log_buffer(pl_event_iw_stop, &buffer_p);
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_iw_stop], 
		  "Bad memcpy()'s for pl_log_iw_stop()");
	buffer_log_record(event_rec_sizes[pl_event_iw_stop]);
    }
} /* pl_log_iw_stop() */


extern void _pl_log_recv_eps(int pl_eps_type, pi_base_t global_eps, pi_base_t local_eps,
			     pi_base_t error_ratio)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	
	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_eps()");
	prepare_log_buffer(pl_event_recv_eps, &buffer_p);
	t_enum = pl_eps_type;
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	lmemcpy(buffer_p, global_eps, SZ_BASE_T);
	lmemcpy(buffer_p, local_eps, SZ_BASE_T);
	lmemcpy(buffer_p, error_ratio, SZ_BASE_T);
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_recv_eps], 
		  "Bad memcpy()'s for pl_log_recv_eps()");
	buffer_log_record(event_rec_sizes[pl_event_recv_eps]);
    }
} /* pl_log_recv_eps() */



extern void _pl_log_send_idle(int pl_idle_type)
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_eps()");
	prepare_log_buffer(pl_event_send_idle, &buffer_p);
	t_enum = pl_idle_type;
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_send_idle], 
		  "Bad memcpy()'s for pl_log_send_idle()");
	buffer_log_record(event_rec_sizes[pl_event_send_idle]);
    }
} /* pl_log_send_idle() */


extern void _pl_log_recv_helper(int pl_lb_type, pi_base_t cur_err_ratio, const region_t *rgn_p,
				pi_base_t size_ratio, pi_base_t new_error_ratio,
				pi_base_t estabs[])
{
    if (!doing_logging)
    {
	return;
    }
    else
    {
	char *buffer_p = logrec_buffer;
	pl_enum_t t_enum;
	pl_base_t t_base;
	
	pi_assert(logging_initialized, "Initialization problem in pl_log_recv_helper()");
	prepare_log_buffer(pl_event_recv_helper, &buffer_p);

        t_enum = pl_lb_type;
	lmemcpy(buffer_p, t_enum, SZ_ENUM_T);
	lmemcpy(buffer_p, cur_err_ratio, SZ_BASE_T);
	if (pl_lb_type == pl_lb_sending)
	{
	    lmemcpy(buffer_p, rgn_p->rgn_num, SZ_NNUM_T);
	    lmemcpy(buffer_p, rgn_p->rgn_creator_id, SZ_RANK_T);
	    lmemcpy(buffer_p, size_ratio, SZ_BASE_T);
	    lmemcpy(buffer_p, new_error_ratio, SZ_BASE_T);
	    t_base = infNorm(estabs, nfcns);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy_array(buffer_p, estabs, nfcns * SZ_BASE_T);
	}
	else
	{
	    pl_nnum_t t_nnum = 0;
	    pl_rank_t t_rank = 0;
	    int i;
	    t_base = 0.0;
	    lmemcpy(buffer_p, t_nnum, SZ_NNUM_T);
	    lmemcpy(buffer_p, t_rank, SZ_RANK_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    lmemcpy(buffer_p, t_base, SZ_BASE_T);
	    for (i = 0; i < nfcns; ++i)
		lmemcpy(buffer_p, t_base, SZ_BASE_T);
	}

	pi_assert(buffer_p - logrec_buffer == event_rec_sizes[pl_event_recv_helper], 
		  "Bad memcpy()'s for pl_log_recv_helper()");
	buffer_log_record(event_rec_sizes[pl_event_recv_helper]);
    }
} /* pl_log_recv_helper() */


/* This should not be called by the caw */
extern void _pl_iw_init(int ndims_parm, int nfcns_parm, int irule, const pi_rparm_t *rparms_p,
			int nPE_parm, int id_parm)
{
    if (!(doing_logging = tobool(strlen(rparms_p->pl_file))))
	return;

    pi_assert(!logging_initialized, "Initialization problem in pl_iw_init()");
    pi_assert(id_parm != 0, "pl_iw_init() called by the rank==0 process");

    D(printf("[PL]: In pl_iw_init()\n"));

    /* Store global values: */
    nfcns = nfcns_parm;
    ndims = ndims_parm;
    nPE = nPE_parm;
    id = (pl_rank_t) id_parm;
    ic_is_working = rparms_p->ic_is_worker;
    fcns_per_irule = neval(ndims_parm, irule);

    initialize_logging(rparms_p, id);

    /* Initialize unique next-number values: */
    logrec_next_num = 1;
    update_next_num = 1;
    region_next_num = 1;

    /* Calc needed record sizes: */
    {
	int max_logrec_size = 0;
	calc_iw_record_sizes(&max_logrec_size, event_rec_sizes, nfcns, ndims);
	pi_assert(max_logrec_size <= LOG_BUFFER_SIZE, 
		  "LOG_BUFFER_SIZE too small for max_logrec_size");
	if (!(logrec_buffer = malloc(max_logrec_size)))
	    pi_sys_terminate("Error from malloc() of logrec_buffer");
    }

    /* (Here, the IC writes the header record; the IW's do not do this) */
    
    /* Start the clock: */
    timer_start();
} /* pl_iw_init() */


/* I write this as this project (ParInt Logging) draws to a close. It is late on a summer
 * Saturday; the air is cool, the music plays.  My cat Merlin wanted to 
 * commemorate this event (somehow?), but is fast asleep in a perfect upside-down loll,  
 * precariously on his perch next to the dark summer window, even as the pen he was playing
 * with moments ago is next to him, awaiting his waking for another mauling by his
 * Friday-the-13th, horror-movie-suitable claws. Good night!
 */

extern void _pl_ic_init(const char fcn_name[], int ndims_parm, int nfcns_parm, int irule, 
			int fcn_eval_limit, pi_base_t eps_a, pi_base_t eps_r,
			const pi_base_t a[], const pi_base_t b[], const pi_rparm_t *rparms_p,
			int nPE_parm, int id_parm)
{
    if (!(doing_logging = tobool(strlen(rparms_p->pl_file))))
	return;

    pi_assert(!logging_initialized, "Initialization problem in pl_ic_init()");
    pi_assert(id == 0, "pl_ic_init() called by a rank!=0 process");

    D(printf("[PL]: In pl_ic_init()\n"));
    
    /* Store global values: */
    nfcns = nfcns_parm;
    ndims = ndims_parm;
    nPE = nPE_parm;
    id = (pl_rank_t) id_parm;
    ic_is_working = rparms_p->ic_is_worker;
    fcns_per_irule = neval(ndims_parm, irule);
    
    initialize_logging(rparms_p, id);

    /* Initialize the next numbers for the IC/CAW: */
    logrec_next_num = 1;
    if (ic_is_working)
    {
	update_next_num = 1;
	region_next_num = 1;
    }

    /* Calc needed record sizes: */
    {
	int max_logrec_size = 0;
	calc_ic_record_sizes(&max_logrec_size, event_rec_sizes, nPE, nfcns, ndims);
	if (ic_is_working)
	    calc_iw_record_sizes(&max_logrec_size, event_rec_sizes, nfcns, ndims);
	pi_assert(max_logrec_size <= LOG_BUFFER_SIZE,
		  "LOG_BUFFER_SIZE too small for max_logrec_size");
	if (!(logrec_buffer = malloc(max_logrec_size)))
	    pi_sys_terminate("Error from malloc() of logrec_buffer");
    }

    /* Write the header record: */
    {
	pl_header_rec_t headrec;
	int arr_size = SZ_BASE_T * ndims;
	
	memset(&headrec, '\0', sizeof(headrec));
	memcpy(headrec.ifcn_name, fcn_name, sizeof(headrec.ifcn_name));
	headrec.ndims          = ndims;
	headrec.nfcns          = nfcns;
	headrec.intg_rule      = irule;
	headrec.eps_a          = eps_a;
	headrec.eps_r          = eps_r;
	headrec.fcn_eval_limit = fcn_eval_limit;
	headrec.do_lb          = rparms_p->do_lb;
	headrec.ic_is_worker   = rparms_p->ic_is_worker;
	headrec.simple_rgns    = rparms_p->simple_rgns;
	headrec.lb_help_ratio  = rparms_p->lb_help_ratio;
	headrec.ns             = rparms_p->ns;
	headrec.fcns_per_irule = fcns_per_irule;
	headrec.nPE            = nPE;
	if (write(log_file_fd, &headrec, sizeof(headrec)) != sizeof(headrec))
	    pi_sys_terminate("Error writing hdr record to log file '%s'", log_file_name);
	if (write(log_file_fd, a, arr_size) != arr_size)
	    pi_sys_terminate("Error writing hdr record a[] to log file '%s'", log_file_name);
	if (write(log_file_fd, b, arr_size) != arr_size)
	    pi_sys_terminate("Error writing hdr record b[] to log file '%s'", log_file_name);
    }

    /* Start the clock: */
    timer_start();
} /* pl_ic_init() */


extern void _pl_finalize(void)
{
    if (!doing_logging)
	return;
    
    buffer_flush();
    if (close(log_file_fd) < 0)
	pi_sys_terminate("Error closing log file '%s'", log_file_name);

    FREE(log_buffer);
    FREE(logrec_buffer);
    logging_initialized = FALSE;
} /* pl_finalize() */

#endif
