/*        File: measures.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI (logging)
   Description: Source for parint measurement functionality
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/measures.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:32 $

                $Log: measures.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:32  parint
                Initial checkin


       Version: 10/24/97  rrz  Original (for CS526 project)
		01/17/98  rrz  Modified for ParInt 1.0
		05/12/98  rrz  Code freeze for ParInt1.0
		05/12/99  rrz  Add the use of USE_COMM_MEASUREMENTS
		07/19/99  rrz  Allow for fcn count to be zero
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
		08/20/99  rrz  Add in: measuring idle time
*/

#include <stdio.h>
#include <math.h>
#include <mpi.h>

#include <pi-version.h>
#include <pi-vspec.h>
#include <measures.h>
#include <math-misc.h>

/* The entire source file is wrapped in a #ifdef so that it does not add to the compile-time,
 * nor the executable or .o size, if we are not using measurements.  I prefer this 
 * technique to monkeying with the Makefile ("Monkeying With a Makefile?  Holy aliteration,
 * Batman!": */

#ifdef USE_MEASUREMENTS

#define NUM_REAL_TAGS   (TAG_MAX - TAG_BASE)

/* This returns the number of regions evaluated by the worker: */
#define fcncount_2_rgns(fcns_per_basrul, fcncount)  ((fcncount) / (fcns_per_basrul))

static int sends_by_tag[NUM_REAL_TAGS];

/* For maintaining function counts: */
static int fcncount = 0;

/* For maintaining times: */
static double comm_time = 0.0;
static double total_time = 0.0;
static double start_time = 0.0;
static double idle_time = 0.0;
static double idle_start_time = 0.0;
static bool currently_idle = FALSE;

/* For storing all idle percentages, for getting a multi-run average: */
static double mrun_idle_percents[PI_MAX_NRUNS];

/* Filled in by ms_handle_results() and ms_mrun_print(); allows a caller to get
 * additional status information: */
static ms_morestatus_t morestatus;

/* About the ms results and ms multi-run results:
   There are some pieces of data (e.g., idle percentage) that are not
   by default recorded and reported to the user.  They are only
   recorded and reported if USE_MEASUREMENTS is on.  Most of the ms
   measurements are simply printed, but the idle% needs to be
   calculated as a multi-run value, and, reported (optionally) as a
   status value (so that I can graph it in bmark3).  To support this,
   I made the following changes:
    - multi_run_init() (in user-fcns.c) calls ms_multirun_init() to
       initialize things for all the runs
    - ms_handle_results() (in measures.c) updates the array(s) of
       multirun data (this is done whether or not there are multiple
       runs).  The structure ms_morestatus_t (stored in measures.c) is
       updated with this run's results. (Note that this means that
       multi_run_update() (in user-fcns.c) does not have to call the
       ms multi run update fn.)
    - multi_run_finish() calls ms_mrun_print() to print the table of
       multi-run data (as before), including the usual (time, tper, etc.), 
       and, things like idle%.  This will also update the data in the
       ms_morestatus_t structure to reflect all runs.
   - After a single run the function ms_get_morestatus() can be called
       to get the runs' extra status data.
   - After multiple runs that same function returns the averaged data
       over all runs.
*/


extern void ms_get_morestatus(ms_morestatus_t *morestatus_p)
{
    *morestatus_p = morestatus;
} /* ms_get_morestatus() */


extern void _ms_mrun_initialize(void)
{
    int i;
    for (i = 0; i < PI_MAX_NRUNS; ++i)
	mrun_idle_percents[i] = 0.0;
} /* _ms_mrun_initialize() */


extern void _ms_initialize_run(void)
{
    int i;
    for (i = 0; i < NUM_REAL_TAGS; ++i)
	sends_by_tag[i] = 0;

    fcncount        = 0;
    comm_time       = 0.0;
    total_time      = 0.0;
    start_time      = 0.0;
    idle_time       = 0.0;
    idle_start_time = 0.0;
    currently_idle  = FALSE;
} /* _ms_initialize_run() */


/* All PE's can call this; PE #0 will collect and print results in this
 * function, other PE's merely send results
 */
extern void _ms_handle_results(int fcns_per_basrul, int run)
{
    int id, nPE;
    
    MPI_Comm_rank(MPI_COMM_WORLD, &id);
    MPI_Comm_size(MPI_COMM_WORLD, &nPE);
    
    if (id == MS_COLLECTOR_ID)
    {
	/* We are the collector: collect results and print them: */
	double rtimes[3], this_idle_percent, total_idle_percent;
	int results[NUM_REAL_TAGS];
	int PEid, recv_fcncount;
	MPI_Status status;
	
	MPI_Comm_rank(MPI_COMM_WORLD, &PEid);
	/* The list of printed tags should be changed for each version to print the
	   version's most important tags: */
#ifdef USE_COMM_MEASUREMENTS
	printf("        %4s %4s %4s %4s %4s %7s %5s %5s %5s %5s %5s\n", 
	       "RUPD", "REPS", "IDLE", "HELP", "SRGN", "#FnEval", "#Rgns", "TotTm", 
	       "IdlTm", "Idle%", "ComTm");
	printf("PE: %2d  %4d %4d %4d %4d %4d %7d %5d %.3f %5s %5s %.3f\n",
	       PEid, sends_by_tag[3], sends_by_tag[4], sends_by_tag[5], sends_by_tag[6],
	       sends_by_tag[7], fcncount, fcncount_2_rgns(fcns_per_basrul, fcncount),
	       total_time, "     ", "     ", comm_time);
#else
	printf("        %4s %4s %4s %4s %4s %7s %5s %5s %5s %5s\n", 
	       "RUPD", "REPS", "IDLE", "HELP", "SRGN", "#FnEval", "#Rgns",
	       "TotTm", "IdlTm", "Idle%");
	printf("PE: %2d  %4d %4d %4d %4d %4d %7d %5d %.3f\n",
	       PEid, sends_by_tag[3], sends_by_tag[4], sends_by_tag[5], sends_by_tag[6],
	       sends_by_tag[7], fcncount, fcncount_2_rgns(fcns_per_basrul, fcncount), 
	       total_time);
#endif	

	total_idle_percent = 0.0;
	for (PEid = 1; PEid < nPE; ++PEid)   /* (Skip PE #0, that should be me) */
	{
	    MPI_Recv(results, NUM_REAL_TAGS, MPI_INT, PEid, TAG_MEASURES, MPI_COMM_WORLD, &status);
	    MPI_Recv(&recv_fcncount, 1, MPI_INT, PEid, TAG_MEASURES, MPI_COMM_WORLD, &status);
	    MPI_Recv(rtimes, 3, MPI_DOUBLE, PEid, TAG_MEASURES, MPI_COMM_WORLD, &status);
	    this_idle_percent = (rtimes[0] != 0.0 ? (rtimes[1] / rtimes[0]) * 100.0 : 0.0);
#ifdef USE_COMM_MEASUREMENTS
	    printf("PE: %2d  %4d %4d %4d %4d %4d %7d %5d %.3f %.3f %5.2f %.3f\n",
		   PEid, results[3], results[4], results[5], results[6], results[7],
		   recv_fcncount, fcncount_2_rgns(fcns_per_basrul, recv_fcncount), 
		   rtimes[0], rtimes[1], this_idle_percent, rtimes[2]);
#else
	    printf("PE: %2d  %4d %4d %4d %4d %4d %7d %5d %.3f %.3f %5.2f\n",
		   PEid, results[3], results[4], results[5], results[6], results[7],
		   recv_fcncount, fcncount_2_rgns(fcns_per_basrul, recv_fcncount), 
		   rtimes[0], rtimes[1], this_idle_percent);
	    total_idle_percent += this_idle_percent;
#endif
	}

	/* Print overall ms info about this run, and, store this run's data in the multi-run 
	   array(s) for later use by ms_mrun_print(): */
	if (nPE > 1)
	{
	    printf("Overall Idle Percentage: %5.2f\n", total_idle_percent / (nPE - 1));
	    mrun_idle_percents[run] = total_idle_percent / (nPE - 1);

	    /* Store additional status information: */
	    morestatus.idle_percent = total_idle_percent / (nPE - 1);
	}
	else
	{
	    printf("\n");
	    mrun_idle_percents[run] = 0.0;

	    /* Store additional status information: */
	    morestatus.idle_percent = 0.0;
	}
    }
    else
    {
	double rtimes[3] = { -1.0, -1.0, -1.0 };
	
	/* We are not the collector so just send our results to the collector: */
	MPI_Send(sends_by_tag, NUM_REAL_TAGS, MPI_INT, MS_COLLECTOR_ID, 
		 TAG_MEASURES, MPI_COMM_WORLD);
	MPI_Send(&fcncount, 1, MPI_INT, MS_COLLECTOR_ID, TAG_MEASURES, MPI_COMM_WORLD);
	rtimes[0] = total_time;
	rtimes[1] = idle_time;
	rtimes[2] = comm_time;  /* (this might be 0.0; that's ok) */
	MPI_Send(rtimes, 3, MPI_DOUBLE, MS_COLLECTOR_ID, TAG_MEASURES, MPI_COMM_WORLD);
    }
} /* _ms_handle_results() */


extern void _ms_start_clock(void)
{
    start_time = MPI_Wtime();
} /* _ms_start_clock() */


extern void _ms_stop_clock(void)
{
    total_time = MPI_Wtime() - start_time;
} /* _ms_stop_clock() */


extern void _ms_update_fcncount(int fcn_calls)
{
    fcncount += fcn_calls;
} /* _ms_update_fcncount() */


/* Need to add support for a user-configurable comm here */
extern void _ms_update_idle(bool now_idle)
{
    if (currently_idle)
    {
	if (!now_idle)
	{
	    int id;
	    MPI_Comm_rank(MPI_COMM_WORLD, &id);
	    pi_assert(idle_start_time != 0.0, "Bad idle_start_time");
	    D(printf("[IW-%d]: IDLE: Now not idle; This End Time %.4f; This idle time %.4f; "
		     "Tot Idle Time %.4f\n", id, MPI_Wtime(),
		     MPI_Wtime() - idle_start_time, idle_time + (MPI_Wtime() - idle_start_time)));
	    idle_time += MPI_Wtime() - idle_start_time;
	    idle_start_time = 0.0;
	}
    }
    else
    {
	if (now_idle)
	{
	    int id;
	    MPI_Comm_rank(MPI_COMM_WORLD, &id);
	    D(printf("[IW-%d]: IDLE: Now idle; Idle Start Time %.4f\n", id, MPI_Wtime()));
	    pi_assert(idle_start_time == 0.0, "Bad idle_start_time");
	    idle_start_time = MPI_Wtime();
	}
    }
    currently_idle = now_idle;
} /* _ms_update_idle() */


extern void _ms_update_comm_time(double comm_start, double comm_stop)
{
    comm_time += comm_stop - comm_start;
} /* _ms_update_comm_time() */


extern void _ms_update_sends_by_tag(int tag)
{
    ++sends_by_tag[tag - TAG_BASE];
} /* _ms_update_sends_by_tag() */


extern void _ms_mrun_print(double mrun_times[], int mrun_fcn_cnts[], int nruns, 
			   int fcn_eval_limit)
{
    double tot_time = 0.0, sqr_time = 0.0, min_time = 0.0;
    double tot_tper = 0.0, sqr_tper = 0.0, min_tper = 0.0;
    double tot_idlp = 0.0, sqr_idlp = 0.0, min_idlp = 0.0;
    int tot_fcnt = 0, min_fcnt = 0;
    double sqr_fcnt = 0.0;      /* (this is a very big number, hence the 'double' type) */

    double avg_time, avg_tper, avg_idlp;
    int avg_fcnt;
    int tot_flag = 0;

    bool flag_hit;
    int i, fcnt;
    double tper, time, idlp;
    
    printf("%6s  %6s %10s %9s %5s %5s\n", "  Run ", "Time",  "Fcn Evals",
	   "Time/1M", "Idle%", "Flag?");
    for (i = 0; i < nruns; ++i)
    {
	idlp = mrun_idle_percents[i];
	flag_hit = (mrun_fcn_cnts[i] >= fcn_eval_limit);
	time = mrun_times[i];
	fcnt = mrun_fcn_cnts[i];
	tper = (fcnt > 0 ? (time / fcnt) * 1000000 : 0.0);
	
	printf("%4d   %7.3f %10d %9.4f %5.2f %3c\n", i + 1, time, fcnt,
	       tper, idlp, B2YN(flag_hit));

	if (i == 0)
	{
	    min_time = time;
	    min_fcnt = fcnt;
	    min_tper = tper;
	    min_idlp = idlp;
	}
	else
	{
	    min_time = MIN(time, min_time);
	    min_fcnt = MIN(fcnt, min_fcnt);
	    min_tper = MIN(tper, min_tper);
	    min_idlp = MIN(idlp, min_idlp);
	}

	tot_time += time;
	sqr_time += time * time;
	tot_fcnt += fcnt;
	sqr_fcnt += fcnt * fcnt;
	tot_tper += tper;
	sqr_tper += tper * tper;
	tot_idlp += idlp;
	sqr_idlp += idlp * idlp;
	if (flag_hit)
	    ++tot_flag;
    }
    
    avg_time = tot_time / nruns;
    avg_fcnt = tot_fcnt / nruns;
    avg_tper = tot_tper / nruns;
    avg_idlp = tot_idlp / nruns;
    printf(" Avgs: %7.3f %10d %9.4f %5.2f\n", avg_time, avg_fcnt, avg_tper, avg_idlp);
    printf(" Sdev: %7.3f %10.0f %9.4f %5.2f\n", 
	   sqrt(absol((sqr_time / nruns) - (avg_time * avg_time))), 
	   sqrt(absol((sqr_fcnt / nruns) - (avg_fcnt * avg_fcnt))),
	   sqrt(absol((sqr_tper / nruns) - (avg_tper * avg_tper))),
	   sqrt(absol((sqr_idlp / nruns) - (avg_idlp * avg_idlp))));
    printf(" Mins: %7.3f %10d %9.4f %5.2f\n", min_time, min_fcnt, min_tper, min_idlp);
    printf("Hit flag %d times\n\n", tot_flag);

    /* Store additional status information: */
    morestatus.idle_percent = avg_idlp;

} /* _ms_mrun_print() */

#endif
