/*        File: messaging.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI
   Description: Source code for MPI send/recv/probe wrapper functions
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/messaging.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:33 $

                $Log: messaging.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:33  parint
                Initial checkin


       Version: 01/15/98  rrz  Original
		05/12/98  rrz  Code freeze for ParInt1.0
                09/17/98  jli  Changes for visualization
		05/12/99  rrz  Add the use of USE_COMM_MEASUREMENTS
		05/25/99  rrz  Add pi_sys_terminate()
		07/13/99  rrz  Add fields to MPI data types for logging
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
*/

/* Note: Some of the functions here are provided as wrappers to the usual MPI functions.  The
   MPI/PMPI techniques were not used, b/c: by using my own wrappers, I can design the 
   prototypes.  (Though, mostly the prototypes are the same, except for pi_iprobe().) Also, I
   did not want to mess up any other wrapper functionality. */


/* #define DBG_THIS_FILE */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <mpi.h>

#include <pi-version.h>
#include <pi-defs.h>
#include <pi-vspec.h>
#include <messaging.h>
#include <measures.h>

#include <logging.h>      /* for definitions of MPI types used for logging */

#define MAX_ERR_LINE   256

static int id = -1;

static void ensure_id(void)
{
    static bool got_rank = FALSE;
    if (!got_rank)
    {
	MPI_Comm_rank(MPI_COMM_WORLD, &id);
	got_rank = TRUE;
    }
} /* ensure_id() */


extern void _failed_assertion(char *str, char *file, int line)
{
    fflush(stdout);
    printf("FAILED ASSERTION: %s\n"
	   " ASSERT LOCATION: File %s, Line %d\n", str, file, line);
    fflush(stdout);
    MPI_Abort(MPI_COMM_WORLD, -1);
} /* _failed_assertion() */


extern int pi_terminate(int pi_err_num, const char *err_str)
{
    int mpi_initialized;
    
    printf("FATAL PARINT ERROR: %s\n", err_str);
    MPI_Initialized(&mpi_initialized);
    if (mpi_initialized)
    {
	/* I have found that for MPICH, MPI_Abort() does not exit very gracefully (SEGV signals
	   result), but I cannot think of a better way of aborting. Note that I abort the entire
	   WORLD, not any user-subgroup, since if this group dies, probably everything will have
	   to die.  The MPI problem is apparently a known problem that will hopefully be fixed in
	   a later release of MPICH: */
	MPI_Abort(MPI_COMM_WORLD, pi_err_num);
    }

    exit(pi_err_num);
    
} /* pi_terminate() */


/* Terminates a run of ParInt.  Call this in circumstances where some system call failed, and
 * where you want the standard message coressponding to errno to be printed out
 * Allows for formatted message strings! (fancy! ;-) )
 */
extern void pi_sys_terminate(const char *fmt, /* arg1, arg2, */ ...)
{
    int errno_save = errno;
    char buf[MAX_ERR_LINE];
    int mpi_initialized;
    va_list ap;
    
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    fflush(stdout);
    fprintf(stdout, "FATAL PARINT SYSTEM ERROR:\n");
    fprintf(stdout, "%s: %s\n", buf, strerror(errno_save));
    fflush(stdout);

    va_end(ap);

    MPI_Initialized(&mpi_initialized);
    if (mpi_initialized)
	MPI_Abort(MPI_COMM_WORLD, PI_ERR_INTERNAL);

    exit(PI_ERR_INTERNAL);
} /* pi_sys_terminate() */


/* This cleans out *all* possible messages
 */
extern void clean_message_buffers(MPI_Comm comm)
{
    MPI_Datatype region_mpi_t;
    MPI_Datatype update_mpi_t;
    MPI_Datatype hist_mpi_t; /* jli!! */
    MPI_Status status;
    update_t update_dummy;
    region_t rgn_dummy;
    hist_t   hist_dummy; /* jli!! dummy for histogram */
    pi_base_t pi_base_dummy;
    int int_dummy;

    create_mpi_type_region(&region_mpi_t);
    create_mpi_type_update(&update_mpi_t);
    create_mpi_type_hist(&hist_mpi_t); /* jli!! */
    D(ensure_id());
    
    while (pi_iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &status))
    {
	D(printf("[%d]: CLEANING: prior cleaning out msg of type %d from %d\n", 
		 id, status.MPI_TAG, status.MPI_SOURCE));
	
	switch (status.MPI_TAG)
	{
	case TAG_WRKR_UPDATE:
	    pi_recv(&update_dummy, 1, update_mpi_t, status.MPI_SOURCE, TAG_WRKR_UPDATE, 
		    comm, &status);
	    break;
	case TAG_NEW_EPS:
	    pi_recv(&pi_base_dummy, 1, pi_base_mpi_t, status.MPI_SOURCE, TAG_NEW_EPS,
		    comm, &status);
	    break;
	case TAG_LB_WRKR_IDLE:
	    pi_recv_dummy(status.MPI_SOURCE, TAG_LB_WRKR_IDLE, comm, &status);
	    break;
	case TAG_LB_HELPER_ID:
	    pi_recv(&int_dummy, 1, MPI_INT, status.MPI_SOURCE, TAG_LB_HELPER_ID, comm, &status);
	    break;
	case TAG_LB_SEND_RGN:
	    pi_recv(&rgn_dummy, 1, region_mpi_t, status.MPI_SOURCE, TAG_LB_SEND_RGN, 
		    comm, &status);
	    break;
	case TAG_HIST: /* jli!! */
	    pi_recv(&hist_dummy, 1, hist_mpi_t, status.MPI_SOURCE, TAG_HIST, 
		    comm, &status);
	    break;
	default:
	    /* Other messages do exist, but if they are still out there now, and we hit
	       this spot of code, then we should die, because things are really messed up: */
	    ensure_id();
	    printf("[%d]: Found tag of %d from %d in clean_message_buffers()\n", 
		   id, status.MPI_TAG, status.MPI_SOURCE);
	    pi_terminate(PI_ERR_INTERNAL, "Bad message tag type in clean_message_buffers()");
	    break;
	}
    }
    
    MPI_Type_free(&update_mpi_t);
    MPI_Type_free(&region_mpi_t);

} /* clean_message_buffers() */



extern void create_mpi_type_region(MPI_Datatype *region_mpi_t)
{

#ifdef USE_LOGGING
#define NUM_FLDS    9
#else
#define NUM_FLDS    7
#endif

    int base, i;
    region_t region; 
    MPI_Aint Rdisp[NUM_FLDS];

#ifdef USE_LOGGING
    MPI_Datatype RMPItype[NUM_FLDS] = {pi_base_mpi_t, pi_base_mpi_t, pi_base_mpi_t, pi_base_mpi_t, 
				       pi_base_mpi_t, pi_base_mpi_t, MPI_INT, MPI_PL_NNUM_T,
				       MPI_PL_RANK_T };
    int Rblocklen[NUM_FLDS] = { PI_MAX_NDIM, PI_MAX_NDIM, 1, PI_MAX_NFCN, PI_MAX_NFCN, 1, 1, 1, 1};
#else
    MPI_Datatype RMPItype[NUM_FLDS]={pi_base_mpi_t, pi_base_mpi_t, pi_base_mpi_t, pi_base_mpi_t, 
			      pi_base_mpi_t, pi_base_mpi_t, MPI_INT};
    int Rblocklen[NUM_FLDS] = { PI_MAX_NDIM, PI_MAX_NDIM, 1, PI_MAX_NFCN, PI_MAX_NFCN, 1, 1 };
#endif    

    i = 0;
    MPI_Address(region.center,    Rdisp + i++);
    MPI_Address(region.h_width,   Rdisp + i++);
    MPI_Address(&(region.rgnvol), Rdisp + i++);
    MPI_Address(region.rgnval,    Rdisp + i++);
    MPI_Address(region.rgnerr_v,  Rdisp + i++);
    MPI_Address(&(region.rgnerr), Rdisp + i++);
    MPI_Address(&(region.divaxn), Rdisp + i++);
#ifdef USE_LOGGING
    MPI_Address(&(region.rgn_num),        Rdisp + i++);
    MPI_Address(&(region.rgn_creator_id), Rdisp + i++);
#endif
    
    base = Rdisp[0]; 
    for (i = 0; i < NUM_FLDS; i++) 
	Rdisp[i] -= base;
    
    MPI_Type_struct(NUM_FLDS, Rblocklen, Rdisp, RMPItype, region_mpi_t);
    MPI_Type_commit(region_mpi_t); 
} /* create_mpi_type_region() */

/* jli!! */
extern void create_mpi_type_hist(MPI_Datatype *hist_mpi_t)
{
    int base, i;
    hist_t hist; 
    MPI_Datatype RMPItype[5]={MPI_INT, MPI_LONG, MPI_INT, MPI_INT, pi_base_mpi_t};
    MPI_Aint Rdisp[5];
    int Rblocklen[5] = { 1, HISTOSIZE, 1, 1, 1 };
    
    MPI_Address(&(hist.worker_id), Rdisp);
    MPI_Address(hist.num_region, Rdisp + 1);
    MPI_Address(&(hist.low_exp), Rdisp + 2);
    MPI_Address(&(hist.high_exp), Rdisp + 3);
    MPI_Address(&(hist.time), Rdisp + 4);
    
    base = Rdisp[0]; 
    for (i = 0; i < 5; i++) 
	Rdisp[i] -= base;
    
    MPI_Type_struct(5, Rblocklen, Rdisp, RMPItype, hist_mpi_t);
    MPI_Type_commit(hist_mpi_t); 
} /* create_mpi_type_hist() */


extern void create_mpi_type_update(MPI_Datatype *update_mpi_t)
{
    /* I would have preferred to use a const int for # of fields, but you cannot initialize
       an array in C with a variable dimension!: */
#ifdef USE_LOGGING
#define NUM_UPD_FLDS   8
#else
#define NUM_UPD_FLDS   7
#endif

    update_t update;
    int base, i;
    MPI_Aint disp[NUM_UPD_FLDS];
    
#ifdef USE_LOGGING
    MPI_Datatype MPItype[NUM_UPD_FLDS] = {pi_base_mpi_t, pi_base_mpi_t, MPI_INT, MPI_INT, 
					  MPI_PL_NNUM_T, MPI_DOUBLE, MPI_LONG, MPI_DOUBLE };
    int blocklen[NUM_UPD_FLDS] = { PI_MAX_NFCN, PI_MAX_NFCN, 1, 1, 1, 1, 1, 1 };
#else
    MPI_Datatype MPItype[NUM_UPD_FLDS] = {pi_base_mpi_t, pi_base_mpi_t, MPI_INT, MPI_INT, 
					  MPI_DOUBLE, MPI_LONG, MPI_DOUBLE };
    int blocklen[NUM_UPD_FLDS] = { PI_MAX_NFCN, PI_MAX_NFCN, 1, 1, 1, 1, 1 };
#endif

    i = 0;
    MPI_Address(update.res,           disp + i++);
    MPI_Address(update.est,           disp + i++);
    MPI_Address(&(update.l_done),     disp + i++);
    MPI_Address(&(update.cnt),        disp + i++);
#ifdef USE_LOGGING
    MPI_Address(&(update.update_num), disp + i++);
#endif
    MPI_Address(&(update.l_estabs),   disp + i++);
    MPI_Address(&(update.num_region), disp + i++);
    MPI_Address(&(update.l_time),     disp + i++);
    
    base = disp[0]; 
    for (i = 0; i < NUM_UPD_FLDS; i++) 
	disp[i] -= base;
    
    MPI_Type_struct(NUM_UPD_FLDS, blocklen, disp, MPItype, update_mpi_t);
    MPI_Type_commit(update_mpi_t); 

#undef NUM_UPD_FLDS

} /* create_mpi_type_update() */


extern int pi_send_dummy(int source, int tag, MPI_Comm comm)
{
    int dummy = 1;
    return pi_send(&dummy, 1, MPI_INT, source, tag, comm);
} /* pi_send_dummy() */


extern int pi_recv_dummy(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
    int dummy;
    return pi_recv(&dummy, 1, MPI_INT, source, tag, comm, status);
} /* pi_recv_dummy() */


extern bool pi_iprobe(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
    int messageflag;

#ifdef USE_COMM_MEASUREMENTS
    double comm_start = MPI_Wtime();
#endif

    pi_assert(valid_pi_tag(tag), "Invalid tag in pi_iprobe()");
    MPI_Iprobe(source, tag, comm, &messageflag, status);
    
#ifdef USE_COMM_MEASUREMENTS
    _ms_update_comm_time(comm_start, MPI_Wtime());
#endif

    return messageflag;
} /* pi_iprobe() */


/***********************************************************************************************/
/*** The next functions only need to be defined if either of these flags are defined.  They
 *** are defined with the extra '_' to match the standard for the actual ms function 
 *** definitions (as in measures.c):
 */

#if defined(USE_MEASUREMENTS) || defined(USE_MSG_DEBUG)

extern int _ms_pi_send(void *msg, int cnt, MPI_Datatype typ, int dest, int tag, MPI_Comm comm)
{
    int retval;

#ifdef USE_COMM_MEASUREMENTS
    double comm_start = MPI_Wtime();
#endif

    pi_assert(valid_pi_tag(tag), "Invalid tag in pi_send()");
    
#ifdef USE_MSG_DEBUG
    ensure_id();
    printf("MSG SEND: PE %2d is sending msg type %3d to %2d\n", id, tag, dest);
#endif

    retval = MPI_Send(msg, cnt, typ, dest, tag, comm);
    
#ifdef USE_COMM_MEASUREMENTS
    _ms_update_comm_time(comm_start, MPI_Wtime());
#endif

#ifdef USE_MEASUREMENTS
    _ms_update_sends_by_tag(tag);
#endif

    return retval;
} /* _ms_pi_send() */


extern int _ms_pi_recv(void *msg, int cnt, MPI_Datatype typ, int src, int tag, MPI_Comm comm, 
		       MPI_Status *status)
{
    int retval;
    
#ifdef USE_COMM_MEASUREMENTS
    double comm_start;
#endif

    pi_assert(valid_pi_tag(tag), "Invalid tag in pi_recv()");

#ifdef USE_MSG_DEBUG
    ensure_id();
    printf("MSG RECV%s: PE %2d is block-receiving msg type %3d from %2d\n", 
	   ((src == MPI_ANY_SOURCE || tag == MPI_ANY_TAG) ? " BEF" : ""), id, tag, src);
#endif

#ifdef USE_COMM_MEASUREMENTS
    comm_start = MPI_Wtime();
#endif

    retval = MPI_Recv(msg, cnt, typ, src, tag, comm, status);

#ifdef USE_COMM_MEASUREMENTS
    _ms_update_comm_time(comm_start, MPI_Wtime());
#endif

#ifdef USE_MSG_DEBUG
    if (src == MPI_ANY_SOURCE || tag == MPI_ANY_TAG)
	printf("MSG RECV AFT: ... and PE %2d received msg type %3d from %2d\n", 
	       id, status->MPI_TAG, status->MPI_SOURCE);
#endif

    return retval;
} /* _ms_pi_recv() */


extern int _ms_pi_probe(int src, int tag, MPI_Comm comm, MPI_Status *status)
{
    int retval;
    
#ifdef USE_COMM_MEASUREMENTS
    double comm_start;
#endif

    pi_assert(valid_pi_tag(tag), "Invalid tag in pi_probe()");

#ifdef USE_MSG_DEBUG
    ensure_id();
    printf("MSG PROBE BEF: PE %2d is block-probing for msg type %3d from %2d\n", id, tag, src);
#endif

#ifdef USE_COMM_MEASUREMENTS
    comm_start = MPI_Wtime();
#endif

    retval = MPI_Probe(src, tag, comm, status);

#ifdef USE_COMM_MEASUREMENTS
    _ms_update_comm_time(comm_start, MPI_Wtime());
#endif

#ifdef USE_MSG_DEBUG
    if (src == MPI_ANY_SOURCE || tag == MPI_ANY_TAG)
	printf("MSG PROBE AFT: ... and PE %2d found msg type %3d from %2d\n", 
	       id, status->MPI_TAG, status->MPI_SOURCE);
#endif

    return retval;
} /* _ms_pi_probe() */


#endif /* defined(USE_MEASUREMENTS) || defined(USE_MSG_DEBUG) */
/**********************************************************************************************/
