/*        File: user-fcns.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: MPI
   Description: Source for ParInt API functions
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/mpi/user-fcns.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:32 $

                $Log: user-fcns.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:32  parint
                Initial checkin


       Version: 01/13/98  rrz  Original
		05/12/98  rrz  Code freeze for ParInt1.0
		07/02/98  rrz  Added for visualization; added pi_setopt_ms()
		10/22/98  rrz  For keeping heap: Add defaulting of new rparm_t
				field
		10/22/98  rrz  For keeping heap: Added PI_OPT_KILLHEAP bit to
				pi_setopt()
		11/14/98  rrz  Change to allow eps_r to be 0.0 if eps_a is >
				0.0
		11/17/98  rrz  Change to allow #PE other than 2^k
		12/18/98  rrz  Add bit option to pi_setopt() for forcing simple
				initial region selection
		05/01/99  rrz  Move rparm_load_defaults() into cmd-line-args.c
		05/11/99  rrz  Add call to good_intg_rule_parameter() (to check
				chosen rule against other parameters)
		05/12/99  rrz  Update new field fcn_evals_per_rule within
				pi_status_t
		05/12/99  rrz  In pi_print_results(), add printing of rgn count
				and modify time-per printing
		05/25/99  rrz  Use new routine pi_sys_terminate()
		05/25/99  rrz  Remove pi_setopt_ms()
		07/14/99  rrz  Remove killheap/keepheap option
		07/19/99  rrz  If init region has zero volume, return a zero
				result immediately
		07/23/99  rrz  --- Code freeze for ParInt1.1 ---
		03/25/00  rrz  Move rparm_load_defaults() to here from
                                cmd-line-args.c
		03/25/00  rrz  Add pi_setotp_max_heap_size()				
*/

/* #define DBG_THIS_FILE */

/* Note: For information on the parameters of the user-called functions, see the documentation */

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

#include <pi-defs.h>
#include <parint.h>
#include <math-misc.h>
#include <math-intg.h>
#include <math-rgn.h>     /* for region_has_zero_volume() */
#include <math-adapt.h>     /* for clear_results() */
 
#include <messaging.h>
#include <measures.h>
#include <pi-vspec.h>
#include <cmd-line-args.h>

#define TPER_FACTOR 1000000

static bool pi_initialized = FALSE;
static bool mpi_initialized = FALSE;
static bool i_did_mpi_init;
static int nPE, id;
static pi_rparm_t rparm;
static MPI_Comm user_comm;

/* This value is, for now, set to be the number of PE's, minus 1 if the IC is not a worker.
 * Future versions of ParInt may have PE's that are not IW's or IC's, in which case
 * this will get more complicated: */
static int nworkers;

/* These values hold all the information needed to print out every conceivable result from
 * a multi-run run of parint: */
static double mrun_times[PI_MAX_NRUNS];
static int mrun_fcn_cnts[PI_MAX_NRUNS];

extern void rparm_load_defaults(pi_rparm_t *rparm_p)
{
    rparm_p->do_lb         = DEF_DO_LB;
    rparm_p->nruns         = DEF_NUM_RUNS;
    rparm_p->lb_help_ratio = DEF_LB_HELP_RATIO;
    rparm_p->ic_is_worker  = DEF_IC_IS_WORKER;
    rparm_p->simple_rgns   = DEF_SIMPLE_RGNS;
    rparm_p->ns            = DEF_NS;
    rparm_p->max_heap_size = DEF_MAX_HEAP_SIZE;
    rparm_p->vis_type      = DEF_VIS_TYPE;
    rparm_p->vis_port      = (int) 0;
    memset(rparm_p->pl_file, '\0', PI_MAX_PATH + 1);
} /* rparm_load_defaults() */

extern void multi_run_init(void)
{
    int i;
    for (i = 0; i < PI_MAX_NRUNS; ++i)
    {
	mrun_times[i] = 0.0;
	mrun_fcn_cnts[i] = 0;
    }
    ms_mrun_initialize();
} /* multi_run_init() */


/* run is 0-indexed */
extern void multi_run_update(pi_status_t *status_p, int run)
{
    mrun_times[run] = status_p->total_time;
    mrun_fcn_cnts[run] = status_p->fcn_eval_count;
} /* multi_run_update() */


/* This updates status_p with the correct values, given the number of runs performed
 * (e.g., avg time, avg fcn count, etc...)
 */
extern void multi_run_finish(pi_status_t *status_p, int nruns, int fcn_eval_limit)
{
    int tot_times_hit_flag = 0;
    double tot_time = 0.0;
    int tot_fcn_cnt = 0;
    int i;

    pi_assert(nruns > 0, "Bad rparm_p->nruns in multi_run_finish()");

    for (i = 0; i < nruns; ++i)
    {
	tot_time += mrun_times[i];
	tot_fcn_cnt += mrun_fcn_cnts[i];
	if (mrun_fcn_cnts[i] >= fcn_eval_limit)
	    ++tot_times_hit_flag;
    }

    status_p->total_time = tot_time / nruns;
    status_p->fcn_eval_count = tot_fcn_cnt / nruns;
    status_p->fcn_limit_flag = tot_times_hit_flag;

    /* This will print out every thing we ever wanted to know about this set of runs: */
    ms_mrun_print(mrun_times, mrun_fcn_cnts, nruns, fcn_eval_limit);
} /* multi_run_finish() */


/* This fn checks each of the given parameters and provides suitable error messages
 * if any are invalid.  A -1 is returned if anything is invalid, 0 otw.
 * Given the pi_hregion_t parameter, note that is only ever called when the user is using
 * the ParInt API; checks are completed in another fashion from the parint executable
 * (though maybe they should be combined?)
 */
static int check_parameters(pi_hregion_t *hregion_p, int nfcns, int intg_rule, int fcn_eval_limit,
			    pi_base_t eps_a, pi_base_t eps_r)
{
    int i;
    
    if (hregion_p->ndims > PI_MAX_NDIM)
    {
	printf("Invalid parm: In hregion, ndims value of %d is greater than max value of %d\n",
	       hregion_p->ndims, PI_MAX_NDIM);
	return -1;
    }
    else if (hregion_p->ndims < 1)
    {
	printf("Invalid parm: In hregion, ndims value is %d; it must be at least 1\n", 
	       hregion_p->ndims);
	return -1;
    }

    for (i = 0; i < hregion_p->ndims; ++i)
    {
	if (hregion_p->a[i] > hregion_p->b[i])
	{
	    printf("Invalid parm: In hregion, a[%d] is > than b[%d]\n", i, i);
	    return -1;
	}
    }

    if (nfcns > PI_MAX_NFCN)
    {
	printf("Invalid parm: nfcns value of %d is greater than max value of %d\n",
	       hregion_p->ndims, PI_MAX_NFCN);
	return -1;
    }
    else if (nfcns < 1)
    {
	printf("Invalid parm: nfcns value is %d; it must be at least 1\n", nfcns);
	return -1;
    }

    if (!valid_intg_rule(intg_rule))
    {
	printf("Invalid parm: Intg rule value of %d is not valid\n", intg_rule);
	return -1;
    }
    /* This prints out a message for us if anything is wrong: */
    if (!good_intg_rule_parameters(intg_rule, hregion_p->ndims, nfcns))
	return -1;

    if (fcn_eval_limit < 1)
    {
	printf("Invalid parm: fcn_eval_limit must be positive\n");
	return -1;
    }

    if (eps_a < 0.0)
    {
	printf("Invalid parm: eps_a must be at least 0.0\n");
	return -1;
    }
    if (eps_r < 0.0)
    {
	printf("Invalid parm: eps_r must be at least 0.0\n");
	return -1;
    }
    if (eps_a == 0.0 && eps_r < pow(10.0, -PI_DIG))
    {
	printf("Invalid parm: if eps_a==0.0, then eps_r must be at least %G\n",
	       pow(10.0, -PI_DIG));
	return -1;
    }

    return 0;
    
} /* check_parameters() */


/* This is called once, by the user, when they want to call MPI_Init().
 */
extern int pi_init(MPI_Comm comm)
{
    int i_mpi_initialized;

    if (pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi_init() called multiple times");

    MPI_Initialized(&i_mpi_initialized);
    if (!i_mpi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi_init() called before MPI_Init()");

    mpi_initialized = TRUE;
    i_did_mpi_init = FALSE;
    user_comm = comm;
    MPI_Comm_size(comm, &nPE);
    MPI_Comm_rank(comm, &id);

    D(if (id == 0) printf("[0]: Reporting start of parint from pi_init()\n"));
    
    rparm_load_defaults(&rparm);
    
    nworkers = nPE - (rparm.ic_is_worker ? 0 : 1);

    pi_initialized = TRUE;
    return 0;
} /* pi_init() */


extern pi_hregion_t *pi_allocate_hregion(int ndims)
{
    pi_hregion_t *hregion_p;
    int i;
    
    if (ndims > PI_MAX_NDIM)
    {
	pi_terminate(PI_ERR_PARMS, "ndims value in pi_allocate_hregion() is too large");
	return NULL;
    }
    else if (ndims < 1)
    {
	pi_terminate(PI_ERR_PARMS, "ndims value in pi_allocate_hregion() is < 1");
	return NULL;
    }

    /* I must allocate the memory for the arrays and the structure itself in 2 malloc()'s;
       I tried it in one and I got a bus error (!): */
    if (   !(hregion_p = (pi_hregion_t *) malloc(sizeof(pi_hregion_t)))
	|| !(hregion_p->a = (pi_base_t *) malloc(sizeof(pi_base_t) * 2 * ndims)))
    {
	pi_sys_terminate("In pi_allocate_hregion(); malloc() error");
	return NULL;
    }

    /* Set the b array to point to the appropiate offset from the beginning of the 2nd chunk 
       of memory allocated: */
    hregion_p->b = hregion_p->a + ndims;
    hregion_p->ndims = ndims;
    /* I could memset() the memory to all bits zero here, but I'm not sure that that
       is ok for long doubles: */
    for (i = 0; i < ndims; ++i)
    {
	hregion_p->a[i] = 0.0;
	hregion_p->b[i] = 0.0;
    }
    
    return hregion_p;
} /* pi_allocate_hregion() */


extern void pi_free_hregion(pi_hregion_t *hregion_p)
{
    if (!hregion_p)
    {
	pi_terminate(PI_ERR_PARMS, "NULL argument to pi_free_hregion()");
	return;
    }
    /* Note that this has to correspond to the pattern of malloc()'s: */
    free(hregion_p->a);
    free(hregion_p);
} /* pi_free_hregion() */


extern int pi_init_mpi(int *argc_ptr, char **argv_ptr[], int *rank, int *comm_size)
{
    int i_mpi_initialized;

    if (!argc_ptr || !argv_ptr || !rank || !comm_size)
	return pi_terminate(PI_ERR_PARMS, "NULL argument to pi_init_mpi()");

    if (pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi initialize fns called multiple times");
    
    MPI_Initialized(&i_mpi_initialized);
    if (i_mpi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi_init_mpi() called after MPI_Init()");

    rparm_load_defaults(&rparm);
    
    /* Set some global variables */
    MPI_Init(argc_ptr, argv_ptr);
    mpi_initialized = TRUE;
    i_did_mpi_init = TRUE;
    user_comm = MPI_COMM_WORLD;
    MPI_Comm_size(MPI_COMM_WORLD, &nPE);
    *comm_size = nPE;
    MPI_Comm_rank(MPI_COMM_WORLD, &id);
    *rank = id;
    
    D(if (id == 0) printf("[0]: Reporting start of parint from pi_init_mpi()\n"));

    nworkers = nPE - (rparm.ic_is_worker ? 0 : 1);

    pi_initialized = TRUE;
    return 0;
} /* pi_init_mpi() */


extern int pi_setopt(int flags)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");

    rparm.do_lb        = !(flags & PI_OPT_NOLOADBAL);
    rparm.ic_is_worker = !(flags & PI_OPT_NOICWORKS);
    rparm.simple_rgns  =  (flags & PI_OPT_SIMPRGNS);
    
    nworkers = nPE - (rparm.ic_is_worker ? 0 : 1);

    if (nworkers == 0)
	return pi_terminate(PI_ERR_NUM_PE,
			    "ERROR: IC must be a worker if # of processes is one\n");

    return 0;
} /* pi_setopt() */



extern int pi_setopt_help_ratio(double help_ratio)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");
    if (help_ratio < 1.0)
	return pi_terminate(PI_ERR_PARMS, "help_ratio too small in pi_setopt_help_ratio()");
    rparm.lb_help_ratio = help_ratio;
    return 0;
} /* pi_setopt_help_ratio() */


extern int pi_setopt_max_heap_size(int max_heap_size)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");
    if (max_heap_size < 0)
	return pi_terminate(PI_ERR_PARMS, "max_heap_size too small in pi_setopt_max_heap_size()");
    rparm.max_heap_size = max_heap_size;
    return 0;
} /* pi_setopt_max_heap_size() */


extern int pi_setopt_nruns(int nruns)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");
    if (nruns < 1)
	return pi_terminate(PI_ERR_PARMS, "nruns too small in pi_setopt_nruns()");
    rparm.nruns = nruns;
    return 0;
} /* pi_setopt_nruns() */


extern int pi_setopt_ns(int ns)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");
    if (ns < 1)
	return pi_terminate(PI_ERR_PARMS, "Value of ns too small in pi_setopt_ns()");
    rparm.ns = ns;
    return 0;
} /* pi_setopt_ns() */


extern int pi_finalize(void)
{
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");
    if (i_did_mpi_init)
	MPI_Finalize();
    return 0;
} /* pi_finalize() */


extern void pi_print_results(pi_base_t result[], pi_base_t estabs[], int nfcns, 
			     pi_status_t *status_p)
{
    if (!result || !estabs || !status_p)
    {
	pi_terminate(PI_ERR_PARMS, "NULL argument to pi_print_results()");
	return;
    }
    if (status_p->fcn_evals_per_rule <= 0)
    {
	pi_terminate(PI_ERR_PARMS, "Invalid value of fcn_evals_per_rule in pi_print_results()");
	return;
    }

    if (nfcns == 1)
    {
	printf("\nRESULT: ");
#ifdef USING_LONG_DOUBLES
	printf("%.33LG\n", result[0]);
#else
	printf("%.15G\n", result[0]);
#endif
	printf("ESTABS: ");
#ifdef USING_LONG_DOUBLES
	printf("%.33LG\n", estabs[0]);
#else
	printf("%.15G\n", estabs[0]);
#endif
    }
    else
    {
	int i;
#ifdef USING_LONG_DOUBLES
	printf("\nFn %38s %38s\n", "RESULT", "ESTABS");
#else
	printf("\nFcn# %21s %21s\n", "RESULT", "ESTABS");
#endif	
	for (i = 0; i < nfcns; i++)
	{
#ifdef USING_LONG_DOUBLES
	    printf("%2d %38.33LG %38.33LG\n", i, result[i], estabs[i]);
#else
	    printf("%3d  %21.15G %21.15G\n", i, result[i], estabs[i]);
#endif
	}
    }
    
    printf("STATUS: Fcn count: %d; Rgn count: %d; Fcn count flag: %d\n", 
	   status_p->fcn_eval_count, status_p->fcn_eval_count / status_p->fcn_evals_per_rule,
	   status_p->fcn_limit_flag);
    printf("        Time: %.3f; Time/1M: %.5f\n", 
	   status_p->total_time, 
	   (status_p->fcn_eval_count > 0
	    ? (status_p->total_time / status_p->fcn_eval_count) * TPER_FACTOR
	    : 0.0));
	   
} /* pi_print_results() */


extern int pi_integrate(pi_ifcn_t ifcn_p, int nfcns, int intg_rule, int fcn_eval_limit,
			pi_base_t eps_a, pi_base_t eps_r, void *region_p, int region_type,
			pi_base_t result[], pi_base_t estabs[], pi_status_t *status_p)
{
    bool result_will_be_zero;
    pi_hregion_t *hregion_p;
    int run;
    
    if (!pi_initialized)
	return pi_terminate(PI_ERR_FN_ORDER, "pi functions called out of order");

    if (!ifcn_p || !region_p || !result || !estabs || !status_p)
	return pi_terminate(PI_ERR_PARMS, "NULL argument to pi_integrate()");

    if (region_type == PI_RGN_HRECT)
	hregion_p = (pi_hregion_t *) region_p;
    else
	return pi_terminate(PI_ERR_PARMS, "Invalid parm: Bad region type");
    
    if (check_parameters(hregion_p, nfcns, intg_rule, fcn_eval_limit, eps_a, eps_r) < 0)
	return pi_terminate(PI_ERR_PARMS, "Invalid parameter");
    
    if (nworkers == 0)
	return pi_terminate(PI_ERR_NUM_PE, "IC must be a worker if # of processes is one");

    if (rparm.nruns > 1)
	multi_run_init();

    result_will_be_zero = region_has_zero_volume(hregion_p->a, hregion_p->b, hregion_p->ndims);

    for (run = 0; run < rparm.nruns; ++run)
    {
	MPI_Barrier(user_comm);
	ms_initialize_run();
	
	if (id == IC_ID)
	{
	    /* Use bool's for internal parint boolean types, even though we present
	       int flags to the user: */
	    bool fcn_limit_flag;
	    
	    if (result_will_be_zero)
	    {
		clear_results(result, estabs, nfcns);
		status_p->fcn_limit_flag = 0;
		status_p->total_time = 0.0;
		status_p->fcn_eval_count = 0;
		status_p->fcn_evals_per_rule = neval(hregion_p->ndims, intg_rule);
	    }
	    else
	    {
		integ_controller(ifcn_p, hregion_p->ndims, nfcns, intg_rule, fcn_eval_limit, eps_a,
				 eps_r, hregion_p->a, hregion_p->b, result, estabs, id, nworkers, 
				 user_comm, &rparm, &fcn_limit_flag, &(status_p->fcn_eval_count), 
				 &(status_p->total_time), NULL);
		status_p->fcn_limit_flag = fcn_limit_flag;
		status_p->fcn_evals_per_rule = neval(hregion_p->ndims, intg_rule);
	    }
	}
	else
	{
	    if (!result_will_be_zero)
		integ_worker(ifcn_p, hregion_p->ndims, nfcns, intg_rule, hregion_p->a,
			     hregion_p->b, nworkers, id, user_comm, &rparm);
	}

	ms_handle_results(neval(hregion_p->ndims, intg_rule), run);
	clean_message_buffers(user_comm);
	if (rparm.nruns > 1)
	    multi_run_update(status_p, run);
    }
    
    if (id == IC_ID && rparm.nruns > 1)
	multi_run_finish(status_p, rparm.nruns, fcn_eval_limit);

    return 0;

} /* pi_integrate() */

