/*        File: heaptest.c
     Copyright: (C) 1998,1999,2000 by Elise de Doncker, Ajay Gupta, Alan Genz & Rodger Zanny. 
                    All rights reserved. See file COPYRIGHT in top-level directory.
  ParInt Versn: All versions (but no MPI required)
   Description: Source for fully testing heaps and deaps.
           CVS: $Source: /home/other/parint/cvsroot/parint1.1/src/seq/heaptest.c,v $
                $Author: parint $
                $Revision: 1.1.1.1 $
                $Date: 2000/06/06 03:13:34 $

                $Log: heaptest.c,v $
                Revision 1.1.1.1  2000/06/06 03:13:34  parint
                Initial checkin


       Version: 02/26/00  rrz  Created for testing changes to heap/deap routines
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>

#include <pi-defs.h>
#include <parint.h>
#include <heapdeap.h>

#define MAX_ERR_LINE        256

/* CAREFUL: On my Linux box, RAND_MAX is used; on the Sparc, RAND_MAX is NOT
 * for use with random(), only rand()!
 * This fix may not be portable!
 */
#define RANDOM_MAX         INT_MAX
#define random_double()    (((double) random()) / RANDOM_MAX)


/* The functions _failed_assertion() and pi_terminate() are normally provided by messaging.c,
 * but we have no such source file in this test program
 */
extern void _failed_assertion(char *str, char *file, int line)
{
    fflush(stdout);
    printf("FAILED ASSERTION: %s\n"
	   " ASSERT LOCATION: File %s, Line %d\n", str, file, line);
    fflush(stdout);
    raise(2);
    exit(1);
} /* _failed_assertion() */


extern int pi_terminate(int pi_err_num, const char *err_str)
{
    printf("FATAL PARINT ERROR: %s\n", err_str);
    exit(pi_err_num);
} /* pi_terminate() */


static int pibase_compare_max(const void *vx1, const void *vx2)
{
    pi_base_t x1 = *(pi_base_t *)vx1;
    pi_base_t x2 = *(pi_base_t *)vx2;
    return ((x1 > x2)
	    ? -1
	    : ((x1 < x2) ? 1 : 0));
} /* pibase_compare_max() */


static int pibase_compare_min(const void *vx1, const void *vx2)
{
    pi_base_t x1 = *(pi_base_t *)vx1;
    pi_base_t x2 = *(pi_base_t *)vx2;
    return ((x1 > x2)
	    ? 1
	    : ((x1 < x2) ? -1 : 0));
} /* pibase_compare_max() */


/* Terminates a run of ParInt.  Call this in circumstances where some system call failed, and
 * where you want the standard message coressponding to errno to be printed out
 * Allows for formatted message strings! (fancy! ;-) )
 */
extern void pi_sys_terminate(const char *fmt, /* arg1, arg2, */ ...)
{
    int errno_save = errno;
    char buf[MAX_ERR_LINE];
    va_list ap;
    
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    fflush(stdout);
    fprintf(stdout, "FATAL PARINT SYSTEM ERROR:\n");
    fprintf(stdout, "%s: %s\n", buf, strerror(errno_save));
    fflush(stdout);

    va_end(ap);
    exit(PI_ERR_INTERNAL);
} /* pi_sys_terminate() */

enum { PRINT_NOTHING, PRINT_LITTLE, PRINT_LOTS };
enum { TEST_HEAP, TEST_MAX_DEAP, TEST_MIN_DEAP };

static void do_test(int test_type, int num_runs, int num_per_run, int print_level)
{
    bool testing_max_something = (test_type == TEST_HEAP || test_type == TEST_MAX_DEAP);
    bool testing_deap = (test_type == TEST_MAX_DEAP || test_type == TEST_MIN_DEAP);
    
    bool found_bug_this_loop;
    pi_base_t *error_array;
    int i, j, delete_num;
    heapdeap_t heap;
    region_t rgn;

    memset(&rgn, 0, sizeof(region_t));
    heap_create(&heap, testing_deap);
    
    if (!(error_array = malloc(sizeof(pi_base_t) * num_per_run)))
    {
	perror("Error trying to malloc() error_array");
	exit(1);
    }

    printf(" --- Starting set of tests; %d tests of %d iterations; deap? %d; max? %d\n",
	   num_runs, num_per_run, testing_deap, testing_max_something);
    for (i = 0; i < num_runs; ++i)
    {
	if (!(i % 2))
	    memset(error_array, 0, sizeof(pi_base_t) * num_per_run);
	if (print_level == PRINT_NOTHING)
	{
	    printf(".");
	    fflush(NULL);
	}
	else if (print_level >= PRINT_LITTLE)
	{
	    printf("Starting test %d of %d\n", i + 1, num_runs);
	    printf("Build the heap to a size of %d...\n", num_per_run);
	    if (print_level == PRINT_LOTS)
		printf("--------------------------------------------------\n");
	}
	for (j = 0; j < num_per_run; ++j)
	{
	    rgn.rgnerr = random_double();
	    if (!(i % 2))
		error_array[j] = rgn.rgnerr;
	    if (print_level == PRINT_LOTS)
		printf("Adding node with value %.10f\n", rgn.rgnerr);
	    heap_add_node(&heap, &rgn);
	    if (print_level == PRINT_LOTS)
	    {
		if (testing_deap)
		    deap_print(&heap, 0);
		else
		    heap_print(&heap, 0);
		printf("\n");
	    }
	}
	
	if (i % 2)
	{
	    if (print_level >= PRINT_LITTLE)
		printf("Odd iteration; freeing entire heap\n");
	    heap_free(&heap);
	    heap_create(&heap, testing_deap);
	}
	else
	{
	    if (print_level >= PRINT_LITTLE)
	    {
		printf("Now Deleting from the heap...\n");
		if (print_level == PRINT_LOTS)
		    printf("--------------------------------------------------\n");
	    }

	    /* Sort array of errors: */
	    if (testing_max_something)
		qsort(error_array, num_per_run, sizeof(pi_base_t), pibase_compare_max);
	    else
		qsort(error_array, num_per_run, sizeof(pi_base_t), pibase_compare_min);

	    if (print_level == PRINT_LOTS)
	    {
		int k;
		for (k= 0; k < num_per_run; ++k)
		    printf("sorted error_array[%5d] is %.10f\n", k, error_array[k]);
	    }
	    
	    delete_num = 0;
	    found_bug_this_loop = FALSE;
	    while (!heap_empty(&heap))
	    {
		{
		    pi_base_t retrieved_key = (testing_max_something
					       ? heap_max_key(&heap)
					       : deap_min_key(&heap));
		    pi_base_t matching_sorted_key = error_array[delete_num];
		    
		    if (matching_sorted_key != retrieved_key)
		    {
			printf("retrieved key is %.10f, not %.10f (delete# is %d); BUG!\n", 
			       retrieved_key, matching_sorted_key, delete_num);
#if 1
			/* For use with multiple tests: */
			heap_free(&heap);
			heap_create(&heap, testing_deap);
			break;
#else
			/* For use with debugger: */
			raise(2);
#endif
		    }
		}

		if (testing_max_something)
		    heap_delete_max(&heap, &rgn);
		else
		    deap_delete_min(&heap, &rgn);

		if (print_level == PRINT_LOTS)
		{
		    if (testing_max_something)
			printf("Result %5d; of heap_delete_max() is %.10f\n", 
			       delete_num, rgn.rgnerr);
		    else
			printf("Result %5d; of heap_delete_min() is %.10f\n", 
			       delete_num, rgn.rgnerr);
		}
		if (!found_bug_this_loop && error_array[delete_num] != rgn.rgnerr)
		{
		    printf("Rgnerr is %.10f, not %.10f (delete# is %d); BUG!!!!\n", 
			   rgn.rgnerr, error_array[delete_num], delete_num);
		    found_bug_this_loop = TRUE;
		    // exit(1);
		}
		++delete_num;
		if (print_level == PRINT_LOTS)
		{
		    if (testing_deap)
			deap_print(&heap, 0);
		    else
			heap_print(&heap, 0);
		    printf("\n");
		}
	    }
	    if (print_level == PRINT_LOTS)
	    {
		printf("\n");
		printf("\n");
	    }
	}
    }
    
    heap_free(&heap);
    free(error_array);
    
} /* do_test() */


int main(int argc, char *argv[])
{
    /* Init random number seed: */
#if 0
    srandom((unsigned) ((time(NULL) & 0xFFFF)));
#else
    srandom((unsigned) 42); // 42!
#endif
    
    printf("******************** STARTING ALL TEST CASES ********************\n");

#if 0
    printf("******** TEST of small heap\n");
    do_test(TEST_HEAP, 1, 10, PRINT_LOTS);

    printf("******** TEST of large heap\n");
    do_test(TEST_HEAP, 100, 10000, PRINT_NOTHING);

    printf("******** TEST of small min deap\n");
    do_test(TEST_MIN_DEAP, 1, 10, PRINT_LOTS);

    printf("******** TEST of small max deap\n");
    do_test(TEST_MAX_DEAP, 1, 10, PRINT_LOTS);

    printf("******** TEST of large max deap\n");
    do_test(TEST_MAX_DEAP, 100, 10000, PRINT_NOTHING);
#endif

    printf("******** TEST of large min deap\n");
    do_test(TEST_MIN_DEAP, 200, 10000, PRINT_NOTHING);

    printf("\n******************** DONE WITH ALL TEST CASES ********************\n");
    
    exit(0);
} /* main() */








